/*
axis - generates a positioned and annotated axis
*/

#ident "$Id: axis.c,v 1.4 2007/05/22 15:33:04 ccrawford Exp $"

#include <math.h> 
#include <string.h> 
#include "crcplot.h"

#define TICMARK .07		/* length of tic/hash mark */
#define ADX -.1			/* x displacement for annotation */
#define ADY1 .15		/* y displacement for annotation */
#define ADY0 -.05		/* y displacement for annotation */
#define NUMHGT .105		/* height of numbers for annotation */
#define LHGT .14		/* height of labels */
#define LDY1 .4			/* y displacement for labels */

void axis(x,y,label,nchar,len,angle,fval,dv)
double x,y;		/* coordinates of start of exis */
char *label;		/* axis label */
int nchar;		/* abs, length of label, -, below, +, above */
double len;		/* (abs) length of axis, -len compressed annotation */
double angle;		/* angle of axis (degrees), 0=x 90=y */
double fval;		/* first annotation value */
double dv;		/* bump between annotation values */
{
	int side;		/* side to annotation +/- 1 */
	int ndec;		/* # decimal digits in call to number */
	int tens;		/* powers of ten to shift annotation */
	int ntic;		/* number of tic marks */
	int i;			/* index for for loops */
	float ticd;		/* distance between tic marks */
	float sina,cosa;	/* sin,cos of angle */
	float dx,dy;		/* fractional displacements */
	float xp,yp;		/* point to plot something */
	float adv;		/* abs(dv) */
	float ll;		/* length of label */
	float scfac;		/* present scale factor from where() */
	int ctic;		/* compressed ticmark annotation */
	int nta;		/* number of characters in tic annotation */
/* 
	initialization
*/
	plots_open_check("axis");

	if(nchar < 0){
		side = -1;
		nchar *= -1;
	}else{
		side = 1;
	}
	if(label && nchar == 999) nchar = strlen(label);
	else nchar = 0;

	cosa = cos(angle * M_PI / 180.0);
	sina = sin(angle * M_PI / 180.0);

	if(len < 0){
		len *= -1;
		ctic = 1;
	}else{
		ctic = 0;
	}
	if(len < 1.0) return;

	ntic = 1.0 + len;
	ticd = len / (ntic - 1);
/*
	draw the axis is and tic marks
*/
	xp = x;
	yp = y;
	dx = -TICMARK * side * sina;
	dy = TICMARK * side * cosa;

	plot(xp,yp,3);
	for(i=0;i<ntic;i++){
		plot(xp,yp,2);
		plot(xp+dx,yp+dy,2);
		plot(xp,yp,3);
		xp += cosa * ticd;
		yp += sina * ticd;
	}
/*
	put dv into (.01,99) range
*/
	tens = 0;
	adv = fabs(dv);

	if(dv != 0.0){
		while(adv >= 99.0){
			adv /= 10.0;
			tens++;
		}
		while(adv < 0.01){
			adv *= 10;
			tens--;
		}
	}
/*
	fix decimal places
*/
	if(tens == 1){ 
		ndec = 1;
		tens = 0;
	}else if(tens == 2){
		ndec = -1;
		tens = 0;
	}else {
		ndec = 2;
	}
/*
	normalize annotations
*/
	fval *= pow(10.0,(float) -tens);
	adv = dv * pow(10.0,(float) -tens);
/*
	plot annotation next to tic marks
*/
	if(!ctic){
				/* normal annotaton */
		dx = ADX;
		dy = ADY1 * side + ADY0;
		xp = x + dx*cosa - dy*sina;
		yp = y + dx*sina + dy*cosa;

		for(i=0;i<ntic;i++){
			number(xp,yp,NUMHGT,fval,angle,ndec);
			xp += cosa * ticd;
			yp += sina * ticd;
			fval += adv;
		}
	}else{
				/* compressed annotaton */
		for(i=0;i<ntic;i++){
			nta = ndec + 2;
			if(fval < 0.0)nta++;
			if(floor(fabs(fval)) >= 10.0) nta++;
			if(floor(fabs(fval)) >= 100.0) nta++;
			if(floor(fabs(fval)) >= 1000.0) nta++;
			if(floor(fabs(fval)) >= 10000.0) nta++;

			dx = -NUMHGT * nta/2.0 * 2.0 *
			   ((i-(ntic-1.0)/2.0)/(ntic-1.0) + .5);
			dx += i * ticd;
			dy = ADY1 * side + ADY0;
			xp = x + dx*cosa - dy*sina;
			yp = y + dx*sina + dy*cosa;

			number(xp,yp,NUMHGT,fval,angle,ndec);
			fval += adv;
		}
	}

/*
	plot the label itself
*/
	ll = nchar;
	if(tens) ll += 6;

	dx = (-LHGT*ll + len) / 2.0;
	dy = LDY1 * side;
	xp = x + dx*cosa - dy*sina;
	yp = y + dy*cosa + dx*sina;
	symbol(xp,yp,LHGT,label? label : " ",angle,999);
/*
	plot the exponent
*/
	if(tens){
		symbol(999.,999.,LHGT,"  *10",angle,5);
		plotwhere_(&xp,&yp,&scfac);
		xp = xp / scfac + LHGT/2.0*cosa;
		yp = yp / scfac + LHGT/2.0*sina;
		number(xp,yp,LHGT/2.0,(float)tens,angle,-1);
	}
}
