/*
cplot - generate contour plot of 2d array
*/

#ident "$Id: cplot.c,v 1.5 2007/05/22 15:33:04 ccrawford Exp $"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include "crcplot.h"

#define HEIGHT 0.14	/* character heights */

static float *d;	/* 2d array of interpolated cube */

static void contour(),err();

/*
	err - print error message and exit
*/
static void err(s)
char *s;
{
	fputs("cplot: ",stderr);
	fputs(s,stderr);
	fputs("\n",stderr);
	exit(1);
}

/*
	contour - plot contours passing through square
*/
static void contour(v0,v1,v2,v3,xp,yp,sx,sy,t,nix,niy)
float v0,v1,v2,v3,xp,yp,sx,sy,t;
int nix,niy;
{
	float e,x=-1,y=-1;
	int ipen,test,i,j;
	float xo,yo,*dp,dx,dy;

			/* case table to determine where lines are */
	static int *ap,a[16*5] = {
		-1,-1,-1,-1,-1,
		0,3,-1,-1,-1,
		0,1,-1,-1,-1,
		3,1,-1,-1,-1,
		1,2,-1,-1,-1,
		0,3,1,2,-1,
		0,2,-1,-1,-1,
		2,3,-1,-1,-1,
		2,3,-1,-1,-1,
		0,2,-1,-1,-1,
		0,1,2,3,-1,
		1,2,-1,-1,-1,
		1,3,-1,-1,-1,
		0,1,-1,-1,-1,
		0,3,-1,-1,-1,
		-1,-1,-1,-1,-1
	};


	test = (v0>t) + ((v1>t)<<1) + ((v2>t)<<2) + ((v3>t)<<3);
	/*printf("test: %d\n",test);*/
	if(test != 0 && test != 15){
		/*printf("subdivide:\n");*/

			/* interpolate top row */
		dx = (v1 - v0) / (nix + 1);
		dp = d;
		*dp++ = v0;
		for(i=1;i<=nix;i++) *dp++ = v0 + i*dx;
		*dp = v1;

			/* interpolate bottom row */
		dx = (v2 - v3) / (nix + 1);
		dp = d + (nix+2) * (niy + 1);
		*dp++ = v3;
		for(i=1;i<=nix;i++) *dp++ = v3 + i*dx;
		*dp = v2;

			/* interpolate columns */
		for(j=0;j<nix+2;j++){
			dy = (d[j + (niy+1)*(nix+2)] - d[j]) / ( niy + 1);
			dp = d + j + nix + 2;
			for(i=1;i<=niy;i++){
				*dp = d[j] + i * dy;
				dp += (nix + 2);
			}
		}

		/*
		for(dp=d,i=0;i<(nix+2)*(niy+2);i++){
			if(i%(nix+2) == 0) putchar('\n');
			printf("%f ",*dp++);
		}
		putchar('\n');
		*/

		dx = 1.0 / (nix + 1);
		dy = 1.0 / (niy + 1);
		yo = 1.0;
		for(j=0;j<=niy;j++){
			yo -= dy;
			dp = d + (nix+2) * j;
			for(i=0;i<=nix;i++){
				xo = i * dx;
				v0 = *dp;
				v1 = *(dp+1);
				v2 = *(dp + 1 + nix + 2);
				v3 = *(dp + nix + 2);
				dp++;
				test = (v0>t) + ((v1>t)<<1) + ((v2>t)<<2) + 
					((v3>t)<<3);
				/*
				printf("v0,v1,v2,v3,test: %f %f %f %f %d\n",
					v0,v1,v2,v3,test);
				*/

				ap = a + test * 5;
				ipen = 3;
				while(*ap != -1){
					/* printf("%d %d\n",*ap,*(ap+1));*/
					switch(*ap++){

					case 0:
						e = (t - v0) / (v1 - v0);
						x = e;
						y = 1;
						break;
					case 1:
						e = (t - v1) / (v2 - v1);
						x = 1;
						y = 1-e;
						break;
					case 2:
						e = (t - v2) / (v3 - v2);
						x = 1 - e;
						y = 0;
						break;
					case 3:
						e = (t - v3) / (v0 - v3);
						x = 0;
						y = e;
						break;
					}

					/*
					printf("x,y,i: %f %f %d\n",x,y,ipen);*/
					x = xp + (xo+x*dx)*sx;
					y = yp + (yo+y*dy)*sy;
					plot(x,y,ipen);
					ipen  ^= 1;
				}
			}
		}
	}
}


void cplot(xd,yd,zd,nx,ny,xmin,xmax,ymin,ymax,xlen,ylen,pxaxis,pyaxis,
	tv,nc,nix,niy,tl,bl,xl,yl)

float *xd,*yd;
float *zd;
int nx,ny;
double xmin,xmax,ymin,ymax;
double xlen,ylen;
int pxaxis,pyaxis;
float *tv;
int nc;
int nix,niy;
char *tl,*bl,*xl,*yl;
{
	float q,dx,dy,xp,yp,sx,sy;
	int i,j,l;
/*
	check parameters and data
*/
	plots_open_check("cplot");

	if(nx < 2 || ny < 2)
		err("nx and ny must be greater than one");

	for(i=0;i<nx;i++) if(xd[i] > xmax || xd[i] < xmin)
		err("x data out of range");

	for(i=0;i<ny;i++) if(yd[i] > ymax || yd[i] < ymin)
		err("y data out of range");

	if(xlen <= 0 || ylen <= 0) err("improper axis length");

	if(xmin == xmax) err("no x range");
	if(ymin == ymax) err("no y range");

	if(nc < 1) err("no contours");
	if(nix < 0 || niy < 0 ) err("illegal interpolation factors");
/*
	flip x and y vectors when monotonically decreasing
*/
	i = monotonicity(xd,nx);
	if(i == 2 || i == 4){
		q = xmax;
		xmax = xmin;
		xmin = q;
	}

	i = monotonicity(yd,ny);
	if(i == 2 || i == 4){
		q = ymax;
		ymax = ymin;
		ymin = q;
	}
/*
	get malloc'ed space for internal vectors
*/
	if(NULL == (d = (float *) malloc((nix+2)*(niy+2)*sizeof(float)))){
		err("malloc failure\n");
	}
/*
	set up scaling factors
*/
	sx = xlen / (xmax - xmin);
	sy = ylen / (ymax - ymin);
/*
	plot axes and annotation
*/
	if(pxaxis){
		axis(0.0,0.0,xl,-999,xlen,0.0,xmin,1.0/sx);
	}

	if(pyaxis){
		axis(0.0,0.0,yl,999,ylen,90.0,ymin,1.0/sy);
	}

	if(bl && (i = strlen(bl))){
		symbol(xlen*0.5-(3./7.*i*HEIGHT),-.6,HEIGHT,bl,0.0,i);
	}

	if(tl && (i = strlen(tl))){
		symbol(xlen*0.5-(3./7.*i*HEIGHT),ylen+HEIGHT,HEIGHT,tl,0.0,i);
	}
/*
	loop through "squares" in image for a dividing squares approach
*/

	for(i=0;i<ny-1;i++){
		dy = (yd[i+1] - yd[i]) * sy;
		yp = (yd[i] - ymin) * sy;
		for(j=0;j<nx-1;j++){
			dx = (xd[j+1] - xd[j]) * sx;
			xp = (xd[j]-xmin) * sx;
			for(l=0;l<nc;l++){
				contour(
					zd[(i+1)*nx+j],
					zd[(i+1)*nx+j+1],
					zd[(i)*nx+j+1],
					zd[(i)*nx+j],
					xp,yp,dx,dy,tv[l],nix,niy);
			}
		}
	}
/*
	clean up and return
*/
	free((char *)d);
	return;
}

