/*
hatch - draw hatch patterns in polygons
*/

#ident "$Id: hatch.c,v 1.4 2007/05/22 15:33:04 ccrawford Exp $"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "crcplot.h"

#define MAXN 50		/* maximum number of points */

#define MAX(x, y)	((x) > (y) ? (x) : (y))
#define MIN(x, y)	((x) < (y) ? (x) : (y))

static int
intersect(y,x1,y1,x2,y2,xint)
float y,x1,y1,x2,y2,*xint;
{
	float	t, a;

	a = y2 - y1;
	if (fabs(a) >= 0.00001) {
		t = (y - y1) / a;
		if (t >= 0.0 && t <= 1.0) {
			*xint = x1 + t*(x2 - x1);
			return (1);
		}
	}
	return (0);
}

void hatch(xa,ya,nn,ang,pitch)
float *xa,*ya;		/* vertices of polygon */
			/* line assumed from last to first vertices */
double ang;		/* angle of hatch marks in degrees */
double pitch;		/* distance between lines, in plot units */
int nn;			/* abs = number of vertices */
			/* >0 draw boarder, <0 don't draw */
{
	float x[MAXN+1],y[MAXN+1];
	float yl, xint, y1, y2, x1, x2, c, s, tmp; 
	float ymax = -HUGE_VAL, ymin = HUGE_VAL;
	float	xi[MAXN+1];
	int	n,np, i, j, sorted;

	/*
	 * Check arguments
	 */
	plots_open_check("hatch");

	n = abs(nn);
	np = n + 1;

	if(pitch <= 0.0){
		fputs("hatch: pitch must be positive\n",stderr);
		exit(1);
	}
	if(n < 3){
		fputs("hatch: must have at least three nodes\n",stderr);
		exit(1);
	}
	if(n > MAXN){
		fprintf(stderr,"hatch: can't have more than %d nodes\n",MAXN);
		exit(1);
	}

	/*
	 * Rotate by the x-hatch angle...
	 */
	ang *= -0.0174532;
	c = cos((double)ang);
	s = sin((double)ang);
	
	for (i = 0; i < np; i++) {
		j = i % n;
		x[i] = xa[j] * c - ya[j] * s;
		y[i] = xa[j] * s + ya[j] *c;
		ymax = MAX(ymax, y[i]);
		ymin = MIN(ymin, y[i]);
	}

	/*
	 * For each y value, get a list of X intersections...
	 */
	yl = ymax - pitch;
	while (yl > ymin) {
		j = 0;
		for (i = 0; i < np-1; i++) 
			if (intersect(yl, x[i], y[i], x[i+1], y[i+1], &xint))
				xi[j++] = xint;
		/*
   		 * Sort the X intersections...
		 */
		sorted = 0;
		while (!sorted) {
			sorted = 1;
			for (i = 0; i < j-1; i++) 
				if (xi[i] > xi[i+1]) {
					tmp = xi[i];
					xi[i] = xi[i+1];
					xi[i+1] = tmp;
					sorted = 0;
				}
		}
		/*
		 *  Draw the lines (Rotated back)...
		 */
		for (i = 0; i < j-1; i += 2) {
			y1 = yl * c - xi[i] * s;
			y2 = yl * c - xi[i + 1] * s;
			x1 = xi[i] * c + yl * s;
			x2 = xi[i + 1] *c + yl * s;
			plot(x1, y1,3);
			plot(x2, y2,2);
		}
		yl -= pitch;
	}

	/*
	 * Draw the outline...
	 */
	if(nn > 0){
		plot(xa[0], ya[0],3);
		for (i = 1; i < n; i++) plot(xa[i], ya[i],2);
		plot(xa[0], ya[0],2);
	}
}


#ifdef HATCH_TEST
main()
{
	static float x[] = { 1,3,2,5,8,5,1};
	static float y[] = { 1,4,2,1,8,2,5};
	int n=sizeof(x)/sizeof(float);

	plots();
	clear();
	hatch(x, y, n, 45., .2);
	plot(0.,0.,999);
}
#endif
