/*
plex - plot sources of examples and their output; plot symbols
*/

static const char rcsid[] = "$Id: plex.c,v 1.6 2004/05/17 17:22:17 ccrawford Exp $";

#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <time.h>
#include <crcplot.h>
#include <crcplot_config.h>
#include <parse.h>

time_t time();

#define ML 70				/* max number of lines in source file */
#define PH 10.0				/* usable paper height in inches */
#define PW (11.0/8.5*PH)		/* usable paper width in inches */
#define PFILE "/var/tmp/plex.plot"	/* plotting output */
#define TFILE "/var/tmp/plex.out"	/* stderr plotting output */
#define SFILE "table.as"		/* list of symbols */

static int hard;	/* 1=send outout to lpr */
static int sym;		/* 1=plot symbols */
static char *ifn;	/* input filename */
static int pn;		/* page number */
static int debug;	/* 1 = debug mode */
static char *ofn = NULL;/* output file for hard copy in PS */
static char cmd[512];	/* for system calls */

struct PARSE_STRUCT parse_table[] = {
	{"=",PARSE_CHAR,(PARSE_ARG)&ifn,"input file"},
	{"page",PARSE_INT,(PARSE_ARG)&pn,"page number"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,"output file for PS"},
	{"of",PARSE_SET,(PARSE_ARG)&hard},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0,0}
	};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'d',&debug,PARSE_FLAG_SET,"debug mode"},
	{'h',&hard,PARSE_FLAG_SET,"hard copy"},
	{'s',&sym,PARSE_FLAG_SET,"plot symbols"},
	{'?',(int *)parse_help_basic,PARSE_PROC},
	{0,0,0,0}
};

static void
page()
{
	char d[256];
	time_t clock;
	struct tm *tt;
	float x,y,h;

	if(!parse_opexist("page")) return;

	h = .15;
	x = h/2;

	y = (pn%2)? 3*h : PH;
	number(x,y,h,(float)pn,-90.0,-1);

	time(&clock);
	tt = localtime(&clock);
	strftime(d,256,"%d %B %Y",tt);
	y = (PH + strlen(d)*h)/2;
	symbol(x,y,h,d,-90.0,999);

	/*
	sr = "Sun Release 4.1";
	y = (pn%2)? PH : strlen(sr)*h;
	symbol(x,y,h,sr,-90.0,999);
	*/
}

static void
pfile(fn,col)
char *fn;
int col;
{
	char a[PARSE_MAX_LINE],b[PARSE_MAX_LINE];
	float x,y,h;
	static char cr[2] = {0x15,0};
	FILE *fd;
	int i,j,e;

	h = PH / ( ML * (1.0 + 5.0/7.0));

	fd = parse_fopen(fn,"r");

	if(col){
		x = PH * .8;
	}else{
		x = 1.0;
	}
	y = PH - .1;

	while(!parse_urline(fd,b)){
		i= j = 0;
		while(b[i]){
			if(b[i] != '\t'){
				a[j++] = b[i];
			}else{
				e = 8 - (j%8);
				while(e-- > 0){
					a[j++] = ' ';
				}
			}
			i++;
		}
		a[j] = 0;

		symbol(x,y,h,a,0.0,999);
		symbol(999.0,999.0,h,cr,0.0,999);
		x = 999.0;
		y = 999.0;
	}
}


static void
plex()
{
	char iifn[100];
	FILE *fd;

	if(!ifn){
		parse_help_basic();
		exit(1);
	}

	strcpy(iifn,ifn);
	parse_remove_suffix(iifn);
	sprintf(cmd,"%s;%s;%s;./%s > %s 2> %s",
		"SUNPLOTHOST=stdout",
		"XPLOTHOST=stdout",
		"export SUNPLOTHOST XPLOTHOST",
		iifn,PFILE,TFILE);
	system(cmd);

	fd = parse_fopen(PFILE,"a+");
	plots_fd(fd);
	parse_suffix(iifn,"c");
	pfile(iifn,1);
	pfile(TFILE,0);
	page();

	plot(0.0,0.0,999);
}

static void
symbols()
{
	FILE *fd;
	char s[PARSE_MAX_LINE],*s1,*s2;		
	int i,n=0;
	int nc=8,nr;
	float x=0,y=0,h;
	static char cr[2] = {0x15,0};

	fd = parse_fopen(SFILE,"r");

	while(!parse_urline(fd,s)){
		if(*s && *s != '#' && *s != ' ' && *s != '\t'){
			n++;
		}else if(strncmp(s,"$Id:",4)){
			n++;
		}
	}
	parse_fseek(fd,0L,"b");
	nr = ceil(((float) n) / nc);
	h = PH / ( (nr + 4) * (1.0 + 5.0/7.0));

	plots_fd(parse_fopen(PFILE,"w"));
	clear();

	i = 0;
	while(!parse_urline(fd,s)){
		if(!*s || *s == '#' || *s == ' ' || *s == '\t')continue;
		if(!strncmp(s,"$Id:",4)) continue;

		if(!(i%nr)){
			x = .75 + (PW-.75)/nc*(i/nr);
			y = PH - 4*h;
		}
		number(x,y,h*.5,(float)i,0.0,-1);
		symbol(999.0,999.0,h," ",0.0,999);
		if((s1 = strchr(s,'<')) == NULL || (s2 = strchr(s,'>')) == NULL
		 || s1 > s2){
			symbol(999.0,999.0,h,(char *)i,0.0,0);
		}else{
			*s2 = 0;
			symbol(999.0,999.0,h,s1+1,0.0,999);
		}
		symbol(999.0,999.0,h,cr,0.0,999);
		x = 999.0;
		y = 999.0;
		i++;
	}
	s1 = "Symbol Routine Character Definitions";
	x = (PW - h*strlen(s1))/2;
	y = PH - 2.0*h;
	symbol(x,y,h,s1,0.0,999);
	page();
	plot(0.0,0.0,999);
}

int
main(argc,argv)
int argc;
char **argv;
{
	int n;
	FILE *ofd,*pfd,*ifd;
	char s[256];

			/* vars for plot_tops call */
	int scale;
	int eps;
	int rot;
	float plot_ho,plot_vo;

	parse_basic(argc,argv);

	unlink(PFILE);
	unlink(TFILE);

	if(sym){
		symbols();
	}else{
		plex();
	}

	ifd = parse_fopen(PFILE,"r");
	if(hard){
		if(ofn == NULL){
			ofd = parse_popen("lpr","w");
		}else if(!strcmp(ofn,"stdout")){
			ofd = stdout;
		}else{
			ofd = parse_fopen(ofn,"w");
		}
		scale = 80;
		eps = 0;
		rot = 1;
		plot_ho = HOR;
		plot_vo = VER;

		plot_tops(ifd,ofd,scale,eps,rot,plot_ho,plot_vo);
	}else{
		plots();
		pfd = plot_fd();
		while((n = fread(s,sizeof(char),256,ifd)) > 0){
			fwrite(s,sizeof(char),n,pfd);
		}
	}
	parse_fclose(ifd);

	if(!debug){
		unlink(PFILE);
		unlink(TFILE);
	}
	exit(0);
}
