/*
plotsubs - plot(3x) subroutines to generate plot(5) format graphics
*/

#ident "$Id: plotsubs.c,v 1.5 2007/05/22 15:33:04 ccrawford Exp $"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <sys/utsname.h>
#include "crcplot.h"
#include "crcplot_config.h"

#define MAXHOSTNAME 65
static char plot_host_name[MAXHOSTNAME+1];
static int port_num;
static FILE *_pipe_fd;
static int sock;

void exit();

#ifdef DEBUG
#define MESSAGE(x) puts(x)
#else
#define MESSAGE(x)
#endif	

static int plot_open=0;	/* 1 = have open plot connection */

static void 
plot_close_check(s)
char *s;
{
	if(plot_open){
		fprintf(stderr,"plotcrc: %s called with active connection\n",s);
		exit(1);
	}
}

static void 
plot_open_check(s)
char *s;
{
	if(!plot_open){
		fprintf(stderr,"plotcrc: call to %s before openpl\n",s);
		exit(1);
	}
}


static void
plot_put_int(a)
int a;
{
	plot_open_check("plot_put_int");
	putc((char)a,_pipe_fd);
	putc((char)(a>>8),_pipe_fd);
}


static void
plot_text(c,s)
char c;
char *s;
{
	plot_open_check("plot_text");
	putc(c,_pipe_fd);
	while(*s && *s != '\n'){
		putc(*s,_pipe_fd);
		s++;
	}
	putc('\n',_pipe_fd);
}

static void 
plot_close()
{
	plot_open_check("plot_close");
	plot_open = 0;
}

void 
plot_openpl_fd(fd)
FILE *fd;
{
	plot_close_check("plot_openpl_fd");
	_pipe_fd = fd;
	sock = -1;
	plot_open = 1;
}

char *
plot_host()
{
	return(plot_host_name);
}

int 
plot_port()
{
	return(port_num);
}

FILE *
plot_fd()
{
	plot_open_check("plot_fd");
	return(_pipe_fd);
}
	
int 
plot_openpl()
{
	char *sp,*getenv();

	if((sp = getenv("XPLOTHOST")) != NULL){
		(void) strncpy(plot_host_name,sp,MAXHOSTNAME);
	}else{
		plot_host_name[0] = 0;
	}
	return(plot_openpl_host(plot_host_name));
}

int 
plot_openpl_host(host)
char *host;
{
	char *c;
	struct sockaddr_in sa;
	struct hostent *hp;
	int addr_type;
	long addr;
	static struct utsname buf;

	MESSAGE("plot_openpl_host");

	plot_close_check("plot_openpl_host");
	port_num = PORT;

	if(host == NULL){
		plot_host_name[0] = 0;
	}else{
		strcpy(plot_host_name,host);
		if(NULL != (c = strchr(plot_host_name,':'))){
			*c = 0;
			port_num = atoi(c+1);
		}
	}

	if(!strcmp(plot_host_name,"stdout")){
		_pipe_fd = stdout;
		sock = -1;
		port_num = -1;
		plot_open = 1;
		return(0);
	}
		
	if(!strlen(plot_host_name)){
		uname(&buf);
		strncpy(plot_host_name,&buf.nodename[0],MAXHOSTNAME);
	}

	memset(&sa,0,sizeof(sa));

	if((hp= gethostbyname(plot_host_name)) == NULL){ 
		if(!isdigit(*plot_host_name)) return(-1);
		if((addr = inet_addr(plot_host_name)) == -1) return(-1);
		addr_type = AF_INET;
		memcpy( (char *)&sa.sin_addr, (char *)&addr, sizeof(addr));
	}else{
		memcpy( (char *)&sa.sin_addr, hp->h_addr, hp->h_length);
		addr_type = hp->h_addrtype;
	}

	sa.sin_port= htons((u_short)port_num);
	sa.sin_family = addr_type;

	if ((sock= socket(addr_type,SOCK_STREAM,0)) < 0) return(-1);

	if (connect(sock,(struct sockaddr *)&sa,sizeof sa) < 0) return(-1);

	MESSAGE("plot_openpl_host - have socket");
	if((_pipe_fd = fdopen(sock,"w")) == NULL){
		fprintf(stderr,"can't fdopen socket\n");
		exit(1);
	}
	plot_open = 1;
	return(0);
}

void 
plot_cont(x,y)
int x,y;
{
	plot_open_check("plot_cont");
	putc('n',_pipe_fd);
	plot_put_int(x);
	plot_put_int(y);
}

void 
plot_hardcopy()
{
	plot_open_check("plot_hardcopy");
	putc('h',_pipe_fd);
}

void 
plot_erase()
{
	plot_open_check("plot_erase");
	putc('e',_pipe_fd);
}

void 
plot_move(x,y)
int x,y;
{
	plot_open_check("plot_move");
	putc('m',_pipe_fd);
	plot_put_int(x);
	plot_put_int(y);
}

void plot_point(x,y)
int x,y;
{
	plot_open_check("plot_point");
	putc('p',_pipe_fd);
	plot_put_int(x);
	plot_put_int(y);
}

void 
plot_arc(x,y,x0,y0,x1,y1)
int x,y,x0,y0,x1,y1;
{
	plot_open_check("plot_arc");
	putc('a',_pipe_fd);
	plot_put_int(x);
	plot_put_int(y);
	plot_put_int(x0);
	plot_put_int(y0);
	plot_put_int(x1);
	plot_put_int(y1);
}

void 
plot_circle(x,y,r)
int x,y,r;
{
	plot_open_check("plot_circle");
	putc('c',_pipe_fd);
	plot_put_int(x);
	plot_put_int(y);
	plot_put_int(r);
}

void 
plot_line(x1,y1,x2,y2)
int x1,y1,x2,y2;
{
	plot_open_check("plot_line");
	putc('l',_pipe_fd);
	plot_put_int(x1);
	plot_put_int(y1);
	plot_put_int(x2);
	plot_put_int(y2);
}

void 
plot_linemod(s)
	char *s;
{
	plot_open_check("plot_linemod");
	if(strcmp(s,"dotted") && strcmp(s,"solid") &&
	 strcmp(s,"longdashed") && strcmp(s,"shortdashed") &&
	 strcmp(s,"dotdashed")){
		fprintf(stderr,"plot_linemod (warning): illegal mode: %s\n",s);
	}else{
		plot_text('f',s);
	}
}

void 
plot_label(s)
char *s;
{
	plot_open_check("plot_label");
	plot_text('t',s);
}

void plot_space(x0,y0,x1,y1)
int x0,y0,x1,y1;
{
	MESSAGE("pspace");

	plot_open_check("plot_space");

	putc('s',_pipe_fd);
	plot_put_int(x0);
	plot_put_int(y0);
	plot_put_int(x1);
	plot_put_int(y1);
}

void 
plot_closepl()
{
	char h[2];

	MESSAGE("pend");

	fflush(_pipe_fd);
	plot_open_check("plot_closepl");
	plot_close();
	if(sock >= 0){
		shutdown(sock,1);

		if(2 != read(sock,&h[0],2) || h[0] != HANDSHAKE1
		 || h[1] != HANDSHAKE2){
			fprintf(stderr,"crcplot: handshake error\n");
			exit(1);
		}

		MESSAGE("pend: have handshake");
		fclose(_pipe_fd);
		shutdown(sock,2);
	}
}

void 
plot_mouse(x,y)
int *x,*y;
{
	int h[2];
	char h1,h2;

	plot_open_check("plot_mouse");

	if(sock < 0){
		fprintf(stderr,"crcplot: plot_mouse not connected to socket\n");
		exit(1);
	}
	plot_text('X',"mouse");
	fflush(_pipe_fd);

	if(sizeof(h1) != read(sock,&h1,sizeof(h1)) || h1 != CMD_MOUSE1
	  || sizeof(h2) != read(sock,&h2,sizeof(h2)) || h2 != CMD_MOUSE2){
		fprintf(stderr,"crcplot: handshake (mouse) error\n");
		exit(1);
	}

	if(sizeof(h) != read(sock,h,sizeof(h))){
		fprintf(stderr,"crcplot: handshake (mouse) error\n");
		exit(1);
	}
	*x = h[0];
	*y = h[1];
}

void plot_flush(void)
{
	plot_text('X',"flush");
	fflush(_pipe_fd);
}
