/*
scale - scale data point values
*/

#ident "$Id: scale.c,v 1.4 2007/05/22 15:33:04 ccrawford Exp $"

#include <math.h>
#include <stdlib.h>
#include <crcplot.h>

static float 
crc_aint(x)
float x;
{
	if(x > 0)
		return floor(x);
	else
		return -floor(-x);
}

void
scale(array,axlen,npts,inc)
float *array;	/* points to scale */
double axlen;	/* length of axis */
int npts;	/* number of points */
int inc;	/* stride in array, <0 supply negative increment */
{
	static float tst[7] = {1.0,2.0,4.0,5.0,8.0,10.0,20.0};

	int i;			/* counter for for loops */
	int n;			/* location of adjusted min */
	float y0,yn;		/* min/max values of array */
	int k;			/* abs(inc) */
	float firstv;		/* adjusted min */
	float deltav;		/* adjusted delta */
	float p;		/* power of 10 of delta */
	float t;		/* temp var */
	float fad;		/* adjustment factor */
	float ys;		/* intermediate value of array */
/*
	some initialization
*/
	fad = 0.01;
	k = abs(inc);
	n = npts * k;
	yn = y0  = array[0];
/*
	find min and max
*/
	for(i=0;i<npts;i++){
		ys = array[i*inc];
		if(ys > yn) yn = ys;
		if(ys < y0) y0 = ys;
	}
/*
	put increment in limited range
*/
	firstv = y0;
	if (y0 < 0.0) fad -= 1.0;
	deltav = (yn-firstv)/ axlen;
	if (deltav <= 0.0){
		deltav = 1.0;
		firstv -= 0.5;
	}
	i = log10(deltav)+1000.0;
	p = pow(10.0,i-1000.);
	deltav = deltav/p - 0.01;
/*
	quantize dv 
*/
	for(i=0;i<6;i++){
		if(tst[i] >= deltav) break;
	}
/*
	adjust delta to span min and max and adjust first values
*/
	while(1){
		deltav = tst[i]*p;
		firstv = deltav* crc_aint(y0/deltav+fad);
		t = firstv + (axlen+0.01)*deltav;
		if(t >= yn) break;
		++i;
	}
/*
	adjust first and last values
*/
	firstv = firstv-crc_aint((axlen+(firstv-yn)/deltav)/2.0) * deltav;
	if((y0*firstv) <= 0.0) firstv = 0.0;
	if(inc < 0){
		firstv = firstv + axlen*deltav;
		deltav = -deltav;
	}
/*
	put values into array
*/
	array[n] = firstv;
	array[n+k] = deltav;
}
