/*
symbol - plot symbols and characters
*/

#ident "$Id: symbol.c,v 1.6 2007/05/22 15:33:04 ccrawford Exp $"

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "symbol.h"		/* contains nxny and charp arrays */
#include "crcplot.h"

#define REDUCE 0.7;		/* height reduction for sup- & subscript */
#define SHIFT_DY 4;		/* shift units in sup- & subscript mode */

static float sdxy = -1;		/* saved scale from font units to screen */
static float sangle;		/* saved plotting angle */
static float dxycos,dxysin;	/* dxy in rotated space */
static float xc,yc;		/* location of present character */
static float xcall,ycall;	/* location of xc&yc at beginning of call */
static int supsub;		/* super/subscript mode; 0=no 1=sup -1=sun */

static void
fontscale(dxy,angle)
	float dxy,angle;
{
	sdxy = dxy;
	sangle = angle;
	angle *= M_PI / 180.0;
	dxycos = cos(angle) * dxy;
	dxysin = sin(angle) * dxy;
}

void 
symbol(x,y,height,itext,angle,nc)
double x,y;		/* location of text, 999=use last position */
double height;		/* height of text; <0 use previous height */
char *itext;		/* text (pointer or actual character, see nc */
double angle;		/* angle of text in degrees */
int nc;			/* +nc=# characters, 0=single char in itext, 
			   -nc=itext is an integer symbol number
			   -1=pen up, <-1 pen down */
{
	int c=-1;		/* character or symbol to plot */
	float hs;		/* scaling factor for height */
	int udm;		/* up/down mode for calls to plot() */
	float dxy;		/* scale from font units to screen */
	int index;		/* index into nxny table, see table.c */
	int count;		/* # nodes for a characyer, see table.c */
	int nx=-1,ny;		/* node values */
	float xt,yt;		/* temporary location for call to plot() */
	int pp=-1;		/* pre- post- move flag in sub/superscript */
/*
	check for open plotter
*/
	plots_open_check("symbol");
/*
	get special characters and set up scale factors 
*/
	udm = 3;
	hs = 7;

	if(nc <= 0){
		c = (int) itext;
		if(nc < 0){
			if(c < 14) hs = 4;
			if(nc < -1) udm = 2;
		}
		if(c > MCHAR - 1){
			fputs("symbol: illegal character/symbol\n",stderr);
			exit(1);
		}
	}else if(nc == 999){
		if(!itext) return;
		nc = strlen(itext);
		if(nc == 0)return;
	}
/*
	set up dx and dy
*/
	if(height > 0){
		dxy = height / hs;
		supsub = 0;
	}else{
		dxy = sdxy;
	}

	if(dxy != sdxy || angle != sangle) fontscale(dxy,angle);
/*
	set up initial plotting locations
*/
	if(x != 999.0) xc = xcall = x;
	if(y != 999.0) yc = ycall = y;
/*
	do the plotting
*/
	while(1){
		if(nc > 0) c = *itext++;
		if(nc > MCHAR - 1) c = ' ';	/* ***** ADD FULL MASK */

		count = charc[c];
		index = charp[c];

		while(count--){
			nx = (nxny[index] >> 4) & 017;
			ny = nxny[index++] & 017;
			if(nx < 15){
				nx -= 2;
				ny -= 2;
				xt = xc + nx*dxycos - ny*dxysin;
				yt = yc + nx*dxysin + ny*dxycos;
				plot(xt,yt,udm);
				udm = 2;
			}else if(ny == 0){	/* move or blank */
				udm = 3;
				nx = 0;		/* bypass bypass move */
			}else if(ny == 1 || ny == 2){	/* sup or sub */

					/* state: sup req: sup res: ignore */
				if(ny == 1 && supsub == 1){
					pp = 0;

					/* state: sup req: sub res: normal */
				}else if(ny == 2 && supsub == 1){
					supsub = 0;
					dxy /= REDUCE;
					pp = 1;
					ny = -SHIFT_DY;

					/* state: norm req: sup res: sup */
				}else if(ny == 1 && supsub == 0){
					supsub = 1;
					dxy *= REDUCE;
					pp = 2;
					ny = SHIFT_DY;

					/* state: norm req: sub res: sub */
				}else if(ny == 2 && supsub == 0){
					supsub = -1;
					dxy *= REDUCE;
					pp = 2;
					ny = -SHIFT_DY;

					/* state: sub req: sup res: normal */
				}else if(ny == 1 && supsub == -1){
					supsub = 0;
					dxy /= REDUCE;
					pp = 1;
					ny = SHIFT_DY;

					/* state: sub req: sub res: ignore */
				}else if(ny == 2 && supsub == -1){
					pp = 0;
				}
				if(pp == 1){
					xc += -ny*dxysin;
					yc += ny*dxycos;
					fontscale(dxy,angle);
				}else if(pp == 2){
					fontscale(dxy,angle);
					xc += -ny*dxysin;
					yc += ny*dxycos;
				}
			}else if(ny == 3){	/* carriage return */
				udm = 3;
				xc = xcall += 12*dxysin;
				yc = ycall += -12*dxycos;
			}else if(ny == 4){	/* back space */
				xc += -7*dxycos;
				yc += -7*dxysin;
			}
		}
		if(nx < 15){
			xc += 7*dxycos;
			yc += 7*dxysin;
		}
		udm = 3;
		if(--nc <= 0)break;
	}
}
