/*
plot_tops - convert plot commands to postscript
*/

#ident "$Id: tops.c,v 1.9 2007/05/22 15:33:04 ccrawford Exp $"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pwd.h>
#include <time.h>
#include <unistd.h>
#include <sys/utsname.h>
#include "crcplot.h"

#define PSINCH 72
#define EPSS 100.0		/* eps scaling factor */

static FILE *pfd;
static int start_up;
static int gscale;
static int eps;
static int rot;
static int hor,ver;
static float bbx,bby;		/* bounding box scale factors */

#define DOTTED 0
#define DOT_DASHED 1
#define SHORT_DASHED 2
#define LONG_DASHED 3
#define SOLID 4

static char *dash_list[] = {
	"/dotted",
	"/dotdashed",
	"/shortdashed",
	"/longdashed",
	"/solid"
};

static char *pro1[] = { 
"% RCS: $Id: tops.c,v 1.9 2007/05/22 15:33:04 ccrawford Exp $\n",
"save 50 dict begin /psplot exch def\n",
"/StartPSPlot\n",
"   {newpath 0 0 moveto .48 setlinewidth 0 setgray 1 setlinecap\n",
"    /imtx matrix currentmatrix def /dmtx matrix defaultmatrix def\n",
"    /fnt /Courier findfont def /smtx matrix def fnt 8 scalefont setfont}def\n",
/* "/EndPSPlot {clear psplot end restore}def\n", drop clear 6/15/94 */
"/EndPSPlot {psplot end restore}def\n",
"/solid {{}0}def\n",
"/dotted	{[2 nail 10 nail ] 0}def\n",
"/longdashed {[10 nail] 0}def\n",
"/shortdashed {[6 nail] 0}def\n",
"/dotdashed {[2 nail 6 nail 10 nail 6 nail] 0}def\n",
"/disconnected {{}0}def\n",
"/min {2 copy lt{pop}{exch pop}ifelse}def\n",
"/max {2 copy lt{exch pop}{pop}ifelse}def\n",
"/len {dup mul exch dup mul add sqrt}def\n",
"/nail {0 imtx dtransform len 0 idtransform len}def\n",
"\n",
"/m {newpath moveto}def\n",
"/n {lineto currentpoint stroke moveto}def\n",
"/p {newpath moveto gsave 1 setlinecap solid setdash\n",
"    dmtx setmatrix .4 nail setlinewidth\n",
"    .05 0 idtransform rlineto stroke grestore}def\n",
"/l {moveto lineto currentpoint stroke moveto}def\n",
"/t {smtx currentmatrix pop imtx setmatrix show smtx setmatrix}def\n",
"/a {gsave newpath /y2 exch def /x2 exch def \n",
"    /y1 exch def /x1 exch def /yc exch def /xc exch def\n",
"    /r x1 xc sub dup mul y1 yc sub dup mul add sqrt\n",
"       x2 xc sub dup mul y2 yc sub dup mul add sqrt add 2 div def\n",
"    /ang1 y1 yc sub x1 xc sub atan def\n",
"    /ang2 y2 yc sub x2 xc sub atan def\n",
"    xc yc r ang1 ang2 arc stroke grestore}def\n",
"/c {gsave newpath 0 360 arc stroke grestore}def\n",
"/f {load exec setdash}def\n",
};
static char *pro3[] = { 
"/e {gsave showpage grestore newpath 0 0 moveto}def\n"
};
static char *pro4[] = { 
"/e {gsave grestore newpath }def\n"
};

#define MAXHOSTNAME 31

static void
intro(llx,lly,urx,ury)
int llx,lly,urx,ury;
{
	time_t clock;
	char *ctime();
	char *host;
	struct passwd *pw;
	int uid;
	static struct utsname buf;

	if(eps){
		fputs("%!PS-Adobe-2.0 EPSF-2.0\n",pfd);
		fprintf(pfd,"%%%%BoundingBox: %d %d %d %d\n",
			(int) (llx / EPSS + 0.5),
			(int) (lly / EPSS + 0.5),
			(int) (bbx * urx / EPSS + 0.5),
			(int) (bby * ury / EPSS + 0.5));
	}else{
		fputs("%!\n",pfd);
	}

	uid = geteuid();
	if(NULL == (pw = getpwuid(uid))){
		fprintf(stderr,"plotps: don't know uid: %d\n",uid);
		exit(1);
	}
	uname(&buf);
	host = &buf.nodename[0];

	fprintf(pfd,"%%%%Creator: %s:%s(%s)\n",host,pw->pw_name,pw->pw_gecos);

	fputs("%%Title: Unix plot file\n",pfd);
	time(&clock);
	fprintf(pfd,"%%%%CreationDate: %s",ctime(&clock));
	fputs("%%DocumentFonts: Courier\n",pfd);
	fputs("%%EndComments\n",pfd);
}

static int
linemod(mode)
char *mode;
{
	int gcn;

	if(!strcmp(mode,"dotted"))
		gcn = DOTTED;
	else if(!strcmp(mode,"longdashed"))
		gcn =  LONG_DASHED;
	else if(!strcmp(mode,"shortdashed"))
		gcn =  SHORT_DASHED;
	else if(!strcmp(mode,"dotdashed"))
		gcn =  DOT_DASHED;
	else
		gcn = SOLID;

	return(gcn);
}

static void
ramp_up()
{
	fputs("StartPSPlot\n",pfd);
	fputs("%%EndProlog\n",pfd);
	fputs("%%Page: 1 1\n",pfd);
}

static void
ramp_down()
{
	fputs("e\n",pfd);
	fputs("%%Trailer\n",pfd);
	fputs("EndPSPlot\n",pfd);
}

static void
check_start()
{
	if(start_up) return;
	fputs("plotps: plot command before scale (s)\n",stderr);
	exit(1);
}

static void
prolog(scale)
int scale;
{
	int i,n;
	float s;

	n = sizeof(pro1) / sizeof(char *);
	for(i=0;i<n;i++) fputs(pro1[i],pfd);

	fputs("/s {/ury exch def /urx exch def \n",pfd);
	fputs("/lly exch def /llx exch def\n",pfd);
	fputs("imtx setmatrix newpath clippath pathbbox newpath\n",pfd);
	fputs("/dury exch def /durx exch def \n",pfd);
	if(eps){
		fputs("/dlly exch def /dllx exch def\n",pfd);
	}else{
		fprintf(pfd,"/dlly exch %d add def\n",ver);
		fprintf(pfd,"/dllx exch %d add def\n",hor);
	}
	fputs("/md durx dllx sub dury dlly sub min def\n",pfd);
	fputs("/Mu urx llx sub ury lly sub max def\n",pfd);
	fputs("dllx dlly translate\n",pfd);
	if(rot && !eps){
		fprintf(pfd,"md %d sub 0 translate\n",hor);
		fputs("90 rotate\n",pfd);
	}
	fprintf(pfd,"md Mu div dup scale\n");
	s = (float) scale / 100;
	fprintf(pfd,"%f %f\n",s,s);
	fputs(" 	scale llx neg lly neg translate}def\n",pfd);

	if(eps){
		n = sizeof(pro4) / sizeof(char *);
		for(i=0;i<n;i++) fputs(pro4[i],pfd);
	}else{
		n = sizeof(pro3) / sizeof(char *);
		for(i=0;i<n;i++) fputs(pro3[i],pfd);
	}
}

#define MASK 0377
static char s[520],*sp;
static int cmd,next,a[260];
static int have_erase;

static void
convert(cs,n)
char *cs;
int n;
{
	int c,i,j;

	for(j=0;j<n;j++){
		c = *cs++;
		if(next > 0){
			*sp++ = c;
			if(sp-s == 2*next){
				for(i=0,sp=s;i<next;i++){
					a[i] = ((sp[1]&MASK) << 8) 
						| (MASK&sp[0]);
					sp += 2;
				}
				next = -2;
			}
		}else if(next == 0){
			switch(c) {
			case 'm':
			case 'p':
			case 'n':
				next = 2;
				break;
			case 'l':
			case 's':
				next = 4;
				break;
			case 't':
			case 'f':
				next = -1;
				break;
			case 'e':
				next = -2;
				break;
			case 'a':
				next = 6;
				break;
			case 'c':
				next = 3;
				break;
			default:
				fprintf(stderr,
					"plotps: unknown command '%c (%x)'\n",
					c,c);
				exit(1);
			}
			sp = s;
			cmd = c;
		}else if(next == -1){
			if(c == '\n'){
				next = -2;
				*sp = '\0';
			}else
				*sp++ = c;
		}
		if(next == -2){
			switch(cmd) {
			case 'm':
			case 'n':
			case 'p':
				check_start();
				if(!eps){
					fprintf(pfd,"%d %d %c\n",a[0],a[1],cmd);
				}else{
					fprintf(pfd,"%f %f %c\n",
						a[0]/EPSS,
						a[1]/EPSS,
						cmd);
				}
				break;
			case 'l':
				if(!eps){
					fprintf(pfd,"%d %d %d %d %c\n",
						a[0],a[1],a[2],a[3],cmd);
				}else{
					fprintf(pfd,"%f %f %f %f %c\n",
						a[0]/EPSS,
						a[1]/EPSS,
						a[2]/EPSS,
						a[3]/EPSS,
						cmd);
				}
				break;
			case 's':
				if(!start_up){
					start_up = 1;
					intro(a[0],a[1],a[2],a[3]);
					prolog(gscale);
					ramp_up();
					if(!eps){
						fprintf(pfd,"%d %d %d %d %c\n",
							a[0],a[1],a[2],a[3],cmd);
					}else if(rot){
						fprintf(pfd,"%f 0 translate\n",
							bbx*a[2]/EPSS);
						fputs("90 rotate\n",pfd);
					}
				}
				break;
			case 't':
				check_start();
				fprintf(pfd,"(%s)%c\n",s,cmd);
				break;
			case 'a':
				check_start();
				if(!eps){
					fprintf(pfd,"%d %d %d %d %d %d %c\n",
						a[0],a[1],a[2],a[3],a[4],a[5],
						cmd);
				}else{
					fprintf(pfd,"%f %f %f %f %f %f %c\n",
						a[0]/EPSS,
						a[1]/EPSS,
						a[2]/EPSS,
						a[3]/EPSS,
						a[4]/EPSS,
						a[5]/EPSS,
						cmd);
				}
				break;
			case 'c':
				check_start();
				if(!eps){
					fprintf(pfd,"%d %d %d %c\n",
						a[0],a[1],a[2],cmd);
				}else{
					fprintf(pfd,"%f %f %f %c\n",
						a[0]/EPSS,
						a[1]/EPSS,
						a[2]/EPSS,
						cmd);
				}
				break;
			case 'f':
				check_start();
				fprintf(pfd,"%s %c\n",
					dash_list[linemod(s)],cmd);
				break;
			case 'e':
				if(have_erase){
					check_start();
					fprintf(pfd,"%c\n",cmd);
				}else{
					have_erase = 1;
				}
				break;
			}
			next = 0;
		}
	}
}

static void
tops_start(ofd,scale,ieps,irot,ihor,iver)
FILE *ofd;
int scale;
int ieps;
int irot;
float ihor,iver;
{
	pfd = ofd;
	gscale = scale;
	eps = ieps;
	rot = irot;
	if(!eps){
		if(rot){
			ver = ihor * PSINCH;
			hor = iver * PSINCH;
		}else{
			hor = ihor * PSINCH;
			ver = iver * PSINCH;
		}
	}else{
		if(rot){
			bby = ihor;
			bbx = iver;
		}else{
			bbx = ihor;
			bby = iver;
		}
	}
	start_up = 0;
}
	
static void
tops_end()
{
	ramp_down();
}

void
plot_tops_buf(buf,n,ofd,scale,ieps,irot,ihor,iver)
char *buf;
int n;
FILE *ofd;
int scale;
int ieps;
int irot;
float ihor,iver;
{
	tops_start(ofd,scale,ieps,irot,ihor,iver);
	convert(buf,n);
	tops_end();
}

void 
plot_tops(ifd,ofd,scale,ieps,irot,ihor,iver)
FILE *ifd,*ofd;
int scale;
int ieps;
int irot;
double ihor,iver;
{
	char buf[512];
	int n;

	tops_start(ofd,scale,ieps,irot,ihor,iver);

	while((n = fread(buf,1,512,ifd)) > 0) convert(buf,n);
	tops_end();
}
