/*
can - convert to and from canonical form
*/

#ident "$Id: can.c,v 1.9 2007/05/22 15:51:51 ccrawford Exp $"

#include "parse.h"

static int can_not_checked = 1;		/* 0 = if can_check run */
static int long_not_checked = 1;	/* 0 = if long's checked */
static int can_native_big;		/* 1 = native big endian */
static int can_short_ok;		/* size of short ok */
static int can_long_ok;			/* size of long ok */
static int can_float_ok;		/* size of float ok */
static int can_int_ok;			/* size of int ok */

#define DO_CHECK	if(can_not_checked) can_check()
#define CHECK_LONG	if(long_not_checked) check_long()

static 
void check_long()
{
	if(!can_long_ok) parse_er("parse: long not four bytes");
	long_not_checked = 0;		
}

static 
void can_check()
{
	union{
                short int v;
		char c[2];
	} u;

	can_short_ok = sizeof(short) == 2;
	can_long_ok = sizeof(long) == 4;
	can_float_ok = sizeof(float) == 4;
	can_int_ok = sizeof(int) == 4;

	if(!can_short_ok) parse_er("parse: short not two bytes");
	if(!can_int_ok) parse_er("parse: int not four bytes");
	if(!can_float_ok) parse_er("parse: float not four bytes");

	u.v = 0x581A;

	if(u.c[0] == 0x58 && u.c[1] == 0x1A){
		can_native_big = 1; 
	}else if(u.c[1] == 0x58 && u.c[0] == 0x1A){
		can_native_big = 0; 
	}else{
		parse_er("parse: can't determine endianess of computer");
	}

	can_not_checked = 0;		
}


long parse_from_can_long(long n)
{
	union{
		long v;
		char c[4];
	} in,out;

	DO_CHECK;
	CHECK_LONG;

	if(can_native_big){		
		return(n);
	}else{
		in.v = n;

		out.c[0] = in.c[3];
		out.c[1] = in.c[2];
		out.c[2] = in.c[1];
		out.c[3] = in.c[0];

		return(out.v);
	}
}	

short parse_from_can_short(short n)
{
	union{
		short v;
		char c[2];
	} in,out;

	DO_CHECK;

	if(can_native_big){		
		return(n);
	}else{
		in.v = n;

		out.c[0] = in.c[1];
		out.c[1] = in.c[0];

		return(out.v);
	}
}	

unsigned short parse_from_can_ushort(unsigned short n)
{
	union{
		unsigned short v;
		char c[2];
	} in,out;

	DO_CHECK;

	if(can_native_big){		
		return(n);
	}else{
		in.v = n;

		out.c[0] = in.c[1];
		out.c[1] = in.c[0];

		return(out.v);
	}
}	

float parse_from_can_float(int n)
{
	union{
		int l;
		float v;
		char c[4];
	} in,out;

	DO_CHECK;

	if(can_native_big){		
		in.l = n;
		return(in.v);
	}else{
		in.l = n;

		out.c[0] = in.c[3];
		out.c[1] = in.c[2];
		out.c[2] = in.c[1];
		out.c[3] = in.c[0];

		return(out.v);
	}
}	

long parse_to_can_long(long n)
{
	union{
		long v;
		char c[4];
	} in,out;

	DO_CHECK;
	CHECK_LONG;

	if(can_native_big){		
		return(n);
	}else{
		in.v = n;

		out.c[0] = in.c[3];
		out.c[1] = in.c[2];
		out.c[2] = in.c[1];
		out.c[3] = in.c[0];

		return(out.v);
	}
}	

short parse_to_can_short(short n)
{
	union{
		short v;
		char c[2];
	} in,out;

	DO_CHECK;

	if(can_native_big){		
		return(n);
	}else{
		in.v = n;

		out.c[0] = in.c[1];
		out.c[1] = in.c[0];

		return(out.v);
	}
}	

unsigned short parse_to_can_ushort(unsigned short n)
{
	union{
		unsigned short v;
		char c[2];
	} in,out;

	DO_CHECK;

	if(can_native_big){		
		return(n);
	}else{
		in.v = n;

		out.c[0] = in.c[1];
		out.c[1] = in.c[0];

		return(out.v);
	}
}	

int parse_to_can_float(float n)
{
	union{
		float v;
		int l;
		char c[4];
	} in,out;

	DO_CHECK;

	if(can_native_big){		
		in.v = n;
		return(in.l);
	}else{
		in.v = n;

		out.c[0] = in.c[3];
		out.c[1] = in.c[2];
		out.c[2] = in.c[1];
		out.c[3] = in.c[0];

		return(out.l);
	}
}	

void parse_from_can_longv(long *v,int n)
{
	int i;

	DO_CHECK;
	if(can_native_big) return;

	for(i=0;i<n;i++){
		v[i] = parse_from_can_long(v[i]);
	}
}

void parse_from_can_shortv(short *v,int n)
{
	int i;

	DO_CHECK;
	if(can_native_big) return;

	for(i=0;i<n;i++){
		v[i] = parse_from_can_short(v[i]);
	}
}

void parse_from_can_ushortv(unsigned short *v,int n)
{
	int i;

	DO_CHECK;
	if(can_native_big) return;

	for(i=0;i<n;i++){
		v[i] = parse_from_can_ushort(v[i]);
	}
}

void parse_from_can_floatv(void *v,int n)
{
	int i;
	int *l=(int *) v;
	float *f=(float *) v;

	DO_CHECK;
	if(can_native_big) return;

	for(i=0;i<n;i++){
		f[i] = parse_from_can_float(l[i]);
	}

}

void parse_to_can_longv(long *v,int n)
{
	int i;

	DO_CHECK;
	CHECK_LONG;
	if(can_native_big) return;

	for(i=0;i<n;i++){
		v[i] = parse_to_can_long(v[i]);
	}
}

void parse_to_can_shortv(short *v,int n)
{
	int i;

	DO_CHECK;
	if(can_native_big) return;

	for(i=0;i<n;i++){
		v[i] = parse_to_can_short(v[i]);
	}
}

void parse_to_can_ushortv(unsigned short *v,int n)
{
	int i;

	DO_CHECK;
	if(can_native_big) return;

	for(i=0;i<n;i++){
		v[i] = parse_to_can_ushort(v[i]);
	}
}

void parse_to_can_floatv(void *v,int n)
{
	int i;
	int *l=(int *)v;
	float *f=(float *)v;

	DO_CHECK;
	if(can_native_big) return;

	for(i=0;i<n;i++){
		l[i] = parse_to_can_float(f[i]);
	}
}

long parse_from_ican_long(long n)
{
	union{
		long v;
		char c[4];
	} in,out;

	DO_CHECK;
	CHECK_LONG;

	if(!can_native_big){		
		return(n);
	}else{
		in.v = n;

		out.c[0] = in.c[3];
		out.c[1] = in.c[2];
		out.c[2] = in.c[1];
		out.c[3] = in.c[0];

		return(out.v);
	}
}	

short parse_from_ican_short(short n)
{
	union{
		short v;
		char c[2];
	} in,out;

	DO_CHECK;

	if(!can_native_big){		
		return(n);
	}else{
		in.v = n;

		out.c[0] = in.c[1];
		out.c[1] = in.c[0];

		return(out.v);
	}
}	

float parse_from_ican_float(int n)
{
	union{
		float v;
		int l;
		char c[4];
	} in,out;

	DO_CHECK;

	if(!can_native_big){		
		in.l = n;
		return(in.v);
	}else{
		in.l = n;

		out.c[0] = in.c[3];
		out.c[1] = in.c[2];
		out.c[2] = in.c[1];
		out.c[3] = in.c[0];

		return(out.v);
	}
}	

long parse_to_ican_long(long n)
{
	union{
		long v;
		char c[4];
	} in,out;

	DO_CHECK;

	if(!can_native_big){		
		return(n);
	}else{
		in.v = n;

		out.c[0] = in.c[3];
		out.c[1] = in.c[2];
		out.c[2] = in.c[1];
		out.c[3] = in.c[0];

		return(out.v);
	}
}	

short parse_to_ican_short(short n)
{
	union{
		short v;
		char c[2];
	} in,out;

	DO_CHECK;

	if(!can_native_big){		
		return(n);
	}else{
		in.v = n;

		out.c[0] = in.c[1];
		out.c[1] = in.c[0];

		return(out.v);
	}
}	

int parse_to_ican_float(float n)
{
	union{
		float v;
		int l;
		char c[4];
	} in,out;

	DO_CHECK;

	if(!can_native_big){		
		in.v = n;
		return(out.l);
	}else{
		in.v = n;

		out.c[0] = in.c[3];
		out.c[1] = in.c[2];
		out.c[2] = in.c[1];
		out.c[3] = in.c[0];

		return(out.l);
	}
}	

void parse_from_ican_longv(long *v,int n)
{
	int i;

	DO_CHECK;
	if(!can_native_big) return;

	for(i=0;i<n;i++){
		v[i] = parse_from_ican_long(v[i]);
	}
}

void parse_from_ican_shortv(short *v,int n)
{
	int i;

	DO_CHECK;
	if(!can_native_big) return;

	for(i=0;i<n;i++){
		v[i] = parse_from_ican_short(v[i]);
	}
}

void parse_from_ican_floatv(void *v,int n)
{
	int i;
	float *f=(float *)v;
	int *l=(int *)v;

	DO_CHECK;
	if(!can_native_big) return;

	for(i=0;i<n;i++){
		f[i] = parse_from_ican_float(l[i]);
	}

}

void parse_to_ican_longv(long *v,int n)
{
	int i;

	DO_CHECK;
	if(!can_native_big) return;

	for(i=0;i<n;i++){
		v[i] = parse_to_ican_long(v[i]);
	}
}

void parse_to_ican_shortv(short *v,int n)
{
	int i;

	DO_CHECK;
	if(!can_native_big) return;

	for(i=0;i<n;i++){
		v[i] = parse_to_ican_short(v[i]);
	}
}

void parse_to_ican_floatv(void *v,int n)
{
	int i;
	float *f=(float *)v;
	int *l=(int *)v;

	DO_CHECK;
	if(!can_native_big) return;

	for(i=0;i<n;i++){
		l[i] = parse_to_ican_float(f[i]);
	}
}


//#define TEST_CAN
#ifdef TEST_CAN

int
main(int argc,char **argv)
{
	FILE *fd;
	int i;
	float a[10];
	float b[10];

	for(i=0;i<10;i++){
		//a[i] = i + i*.1;
		//a[i] = i * 10;
		a[i] = 9.166869 + i*5;;
	}
	parse_to_ican_floatv(a,10);
	fd = parse_fopen("y","w");
	parse_fwrite(a,4,10,fd);
	parse_fclose(fd);

	fd = parse_fopen("y","r");
	parse_fread(b,4,10,fd);
	parse_from_ican_floatv(b,10);
	for(i=0;i<10;i++){
		if(a[i] != b[i])
			printf("%d:, a,b %f %f\n",i,a[i],b[i]);
	}

	return(0);
}
#endif
