/*
parse_fft - calculate one-dimensional fast fourier transforms 
parse_fft2d - calculate two-dimensional fast fourier transforms 
*/

#ident "$Id: fft.c,v 1.5 2007/05/22 15:51:51 ccrawford Exp $"

#include <math.h>
#include "parse.h"

void 
parse_fft (
    float *a,	/* real part of data */
    float *b,	/* imaginary part of data */
    int m,	/* size of data = 2**m */
    int iopt	/* 0=dft, 1=idft */
)
{
	int	nv2,nm1,n,le,le1,ip;
	float	ua,ub,wa,wb,ta,tb,*ap,*bp;
	float   pi,pile1,tmp;
	register int i,j,l;

	n = 1<<m;
	if(iopt){
		for(i=0,ap=a,bp=b;i<n;i++){
			*ap++ /= n;
			*bp++ /= -n;
		}
	}
	nv2 = n>>1;
	nm1 = n - 1;
	j = 0;
	for(i=0;i<nm1;i++){
		if(i<j){
			ta = a[j];	tb = b[j];
			a[j] = a[i];	b[j] = b[i];
			a[i] = ta;	b[i] = tb;
		}
		l = nv2;
		while(l < (j+1) ){
			j -= l;
			l = l >> 1;
		}
		j += l;
	}
	pi = 3.1415926535;
	for(l=1;l<=m;l++){
		le = 1<<l;
		le1 = le>>1;
		ua = 1.0;	ub = 0.0;
		pile1 = pi / le1;
		wa = cos(pile1);	wb = -sin(pile1);
		for(j=0;j<le1;j++){
			for(i=j;i<n;i += le){
				ip = i + le1;
				ta = a[ip] * ua - b[ip] * ub;
				tb = a[ip] * ub + b[ip] * ua;
				a[ip] = a[i] - ta;
				b[ip] = b[i] - tb;
				a[i] += ta;
				b[i] += tb;
			}
			ua = (tmp = ua) * wa - ub * wb;
			ub = tmp * wb + ub * wa;
		}
	}
	if(iopt){
		for(i=0;i<n;i++)b[i] = -b[i];
	}
}

void 
parse_fft2d (float *a, float *b, int mx, int my, int iopt)
{
	register int i,j;
	int nx,ny;
	float *da,*db;
/* 
	initialization
*/
	nx = 1<<mx;
	ny = 1<<my;
/* 
	get space for temp arrays 
*/
	da = parse_malloc_float(2*ny);
	db = da + ny;
/* 
	take fft's of rows 
*/
	for(i=0;i<ny;i++){
		parse_fft(a+i*nx,b+i*nx,mx,iopt);
	}
/* 
	take fft's of columns 
*/
	for(i=0;i<nx;i++){
		for(j=0;j<ny;j++){
			da[j] = a[i+j*nx];
			db[j] = b[i+j*nx];
		}
		parse_fft(da,db,my,iopt);
		for(j=0;j<ny;j++){
			a[i+j*nx] = da[j];
			b[i+j*nx] = db[j];
		}
	}
/*
	clean up 
*/
	parse_free((char *)da);
}
