/*
flags - manipulate PARSE_FLAG_TABLE tables

parse_fflags - install one flag
parse_flag_help - print syntax of flag table
*/

#ident "$Id: flags.c,v 1.6 2007/05/22 15:51:51 ccrawford Exp $"

#include <stdlib.h>
#include "parse.h"

static int 
first_flag (			/* see if first occurrence of flag */
    struct PARSE_FLAG_TABLE *pt,
    int i
)
{
	int m=0;
	while(pt[m].flag){
		if(pt[m].flag == pt[(int)i].flag) return( i == m );
		m++;
	}
	parse_er("parse_flags (first_flag): should not be here");
	exit(1);
	return(-1);		/* make compiler happier */
}

void 
parse_flag_help (struct PARSE_FLAG_TABLE table[], FILE *fd)
{
	int i=0;

	if(!table[i].flag){
		fputs("No flags.\n",fd);
		return;
	}

	while(table[i].flag){
		if(table[i].flag != '?' && first_flag(table,i)){
			fprintf(fd,"-%c: %s\n",table[i].flag,
				(table[i].help != NULL)? 
					table[i].help : "(no help)");
		}
		i++;
	}
}

int 
parse_fflags (struct PARSE_FLAG_TABLE table[], char *s, FILE *fd)
{
	int type,i,found,negate=0;
	int mask = ((1 << PARSE_FLAG_SHIFT) - 1);

	if(*s != '-'){
		fprintf(fd,"flag string missing hypen, %s\n",s);
		return(-1);
	}
	while(*(++s)){
		if(*s == ' '){
			;
		}else if(*s == '-'){
			negate = 1;
		}else if(*s == '+'){
			negate = 0;
		}else if(*s == '?'){
			for(i=0;;i++){
				if(!table[i].flag){
					parse_flag_help(table,fd);
					return(-1);
				}else if(*s == table[i].flag){
						/* see 'C traps and pitfalls' */
					(*(void(*)())(table[i].var))();
					return(-1);
				}
			}
			/*NOTREACHED*/
		}else{
			found = i = 0;
			while(table[i].flag){
			  if(table[i].flag == *s){
				found = 1;

				type = mask & table[i].type;
				if(negate && type != PARSE_FLAG_SET
				  && type != PARSE_FLAG_CLR){
					fprintf(fd,"can't negate: -%c\n",*s);
					return(-1);
				}

				switch(type){
				case PARSE_FLAG_SET:
					*(table[i].var) = 1 - negate;
					break;
				case PARSE_FLAG_CLR:
					*(table[i].var) = negate;
					break;
				case PARSE_FLAG_INC:
					(*(table[i].var))++; 
					break;
				case PARSE_FLAG_DCR:
					(*(table[i].var))-- ;
					break;
				case PARSE_FLAG_VAL:
					(*(table[i].var)) = 
					  table[i].type >> PARSE_FLAG_SHIFT;
					break;
				case PARSE_FLAG_PROC:
						/* see 'C traps and pitfalls' */
				    (*(void(*)(char))(table[i].var))(table[i].flag);
					break;
				default:
					fputs("parse_fflags: bad type\n",fd);
					return(-1);
				}
			  }
			  i++;
			}
			if(!found){
				fprintf(fd,"bad flag: -%c\n",*s);
				return(-1);
			}
		}
	}
	return(0);	
}
