/*
mi - handle mi image files
*/

#ident "$Id: mi.c,v 1.13 2011/09/07 09:38:19 carl Exp $"

#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>

#include "parse.h"

static int checklimit=-1; 	// check all images if less than this number
				// <0: check all files

void
parse_omi_checklimit(int limit)
{
	checklimit = limit; 	
}

static int 
mi_fread(void *ptr,int size,int n,FILE *unit,int err)
{
	if(n != (int) fread(ptr,size,n,unit)){
		if(err){
			parse_fn_error("mi_fread: read error",unit);
		}else{
			return(-1);
		}
	}
	return(0);
}

static int 
mi_fseek(FILE *unit,long byte,char *type,int err)
{
	int m;

	if(!strcmp(type,"b")){
		m = 0;
	}else if(!strcmp(type,"c")){
		m = 1;
	}else if(!strcmp(type,"e")){
		m = 2;
	}else{
		fprintf(stderr,"mi_fseek: illegal mode: %s file: %s",
			type,parse_fn_name(unit));
		exit(1);
	}

	if(-1 == fseek(unit,byte,m)){
		if(err){
			parse_fn_error("mi_fseek: seek error",unit);
		}else{
			return(-1);
		}
	}
	return(0);
}

static int 
mi_rblock(FILE *unit,int start,void *data,int count,int err)
{
	if(-1 == mi_fseek(unit,(long) (start * 512L),"b",err)){
		return(-1);
	}
	if(-1 == mi_fread(data,1,count*512,unit,err)){
		return(-1);
	}
	return(0);
}

void 
parse_wmi_l (char *file, unsigned short *image, int ncol, int nrow, char *label)
{
	FILE *ou=NULL;
	unsigned short head[256];
	short ihead[256];
	char *sfn,*filep;
	int oblk=-1;
	int nblk;
	int nmi=-1,n,np,newf;
	struct stat sbuf;
	int msg;
	PARSE_MIHEAD *mihead;
	int first=-1;

	sfn = filep = parse_suffix_malloc(file,"mi");
	newf = 1;
	if(ncol < 0){
		msg = 0;
		ncol = -ncol;
	}else{
		msg = 1;
	}

	if((int) strlen(filep) >= 2 && filep[0] == '_' && filep[1] == '_'){
		filep += 2;
		if(!stat(filep,&sbuf)){
			ou = parse_fopen(filep,"rb+");
			mihead = parse_omi(ou,1);

			if(mihead->type != PARSE_MI_65K)
				parse_er("parse_wmi: can only append to 65K mi-files");
			nmi = mihead->n + 1;
			first = mihead->first;
			oblk = mihead->tblk;
			parse_frmi(mihead);
			newf = 0;
		}else{
			nmi = 1;
		}
		if(msg) fprintf(stdout,"parse_wmi: image %d added to: %s\n",
			nmi,filep);
	}

	if(newf){
		ou = parse_fopen(filep,"wb");
		nmi = 1; 
		oblk = 1;
		first = 1;
	}

	if(ou == NULL) parse_er("parse_wmi: ou == NULL");
	if(first == -1) parse_er("parse_wmi: first == -1");
	if(oblk == -1) parse_er("parse_wmi: oblk == -1");

	n = ncol*nrow;
	nblk = PARSE_BLOCK_SHORT(n);

	parse_zbuf((short *)head,256);
	head[PARSE_MI_LOC_MAGIC] = PARSE_MI_MAGIC_65K;
	head[PARSE_MI_LOC_N] = nmi;
	head[PARSE_MI_LOC_FIRST] = first;

	parse_to_can_ushortv(head,256);
	parse_wblock(ou,0,(char *)head,1);

	parse_zbuf(ihead,256);
	ihead[PARSE_MI_LOC_NCOL] = parse_to_can_short(ncol);
	ihead[PARSE_MI_LOC_NROW] = parse_to_can_short(nrow);
	strncpy((char *) &ihead[PARSE_MI_LOC_LABEL],label,PARSE_MI_LEN_LABEL);
	parse_wblock(ou,oblk,(char *)ihead,1);
	np = nblk * 512 / sizeof(short);
	parse_to_can_ushortv(image,n);
	parse_fwrite((char *)image,sizeof(short),n,ou);
	parse_fwrite((char *)head,sizeof(short),np-n,ou);
	parse_fclose(ou);
	parse_free(sfn);
	parse_from_can_ushortv(image,n);
}

void 
parse_wmi (char *file, unsigned short *image, int ncol, int nrow)
{
	parse_wmi_l(file,image,ncol,nrow,"");
}

unsigned short int *
parse_rmi (char *file, unsigned short *image, int *rncol, int *rnrow)
{
	int nrow,ncol;
	FILE *iu;
	char *filep;
	int n;
	unsigned short *ip;
	int i;
	int suffix;		/* 0=none, 1=first, 2=last, 3=number */
	int inum=-1;		/* image number when suffix=3 */
	char *ifn;		/* copy of input filename */
	char *cp;		/* pointer to ifn */
	PARSE_MIHEAD *mihead;

	ifn = parse_malloc_char(strlen(file)+1);
	strcpy(ifn,file);
	if((cp = strchr(ifn,':')) != NULL){
		*cp++ = 0;
		if(!*cp)parse_err("parse_rmi: bad suffix ",file);
		if(*cp == 'f' || *cp == 'F'){
			suffix = 1;
		}else if(*cp == 'l' || *cp == 'L'){
			suffix = 2;
		}else{
			suffix = 3;
			inum = (int) parse_atof(cp);
		}
	}else{
		suffix = 0;
	}

	filep = parse_suffix_malloc(ifn,"mi");
	parse_free(ifn);
	iu = parse_fopen(filep,"rb");
	mihead = parse_omi(iu,0);
	i = mihead->n;

	if(i == 0) parse_err("parse_rmi: file empty: ",filep);

	if(suffix == 1){
		inum = 0;
	}else if(suffix == 2){
		inum = i - 1;
	}else if(suffix == 3){
		if(inum == -1) parse_er("parse_rmi: inum == -1");
		parse_pcheck(inum,1,i,"parse_rmi: image number out of range");
		inum--;
	}else{
		if(i > 1)parse_wtext("parse_rmi (warning): using last image");
		inum = i - 1;
	}
		
	*rncol = ncol = mihead->ncol[inum];
	*rnrow = nrow = mihead->nrow[inum];
	n = ncol * nrow;

	if(image == NULL){
		ip = (unsigned short *) parse_malloc(n * sizeof(unsigned short));
	}else{
		ip = image;
	}

	parse_fread_block((char *)ip,sizeof(short),n,iu,mihead->bptr[inum]+1);
	parse_from_can_ushortv(ip,n);
	parse_fclose(iu);
	parse_free(filep);
	parse_frmi(mihead);
	return(ip);
}

void
parse_frmi(PARSE_MIHEAD *mihead)
{
	int i;
	int n;

	if(mihead == NULL) parse_er("parse_frmi: can't free NULL pointer");
	n = mihead->n;
	if(mihead->ncol != NULL) free((void *) mihead->ncol);
	if(mihead->nrow != NULL) free((void *) mihead->nrow);
	if(mihead->bptr != NULL) free((void *) mihead->bptr);
	if(mihead->nblk != NULL) free((void *) mihead->nblk);
	if(mihead->label != NULL){
		for(i=0;i<n;i++){
			if(mihead->label[i] != NULL) free((void *) mihead->label[i]); 
		}
		free((void *) mihead->label); 
	}
	free((void *) mihead); 
}

PARSE_MIHEAD *
parse_omi_fe(FILE *iu,int writeonly,FILE *ou,int err)
{
	PARSE_MIHEAD *mihead=NULL;	/* return header */
	int i;				/* looper */
	int n;				/* number of images in file */
	unsigned short head[1024];	/* mi-file header */
	unsigned short ihead[1024];	/* mi-image header */
	int contig=-1;			/* 1=images contiguous */
	int same=-1;			/* 1=images same size */
	int order=-1;			/* 1=images in order in file */
	int ncol=-1;			/* # col in specific image */
	int nrow=-1;			/* # rows in specific image */
	int min_ncol=-1,max_ncol=-1;	/* min/max ncol for all images */
	int min_nrow=-1,max_nrow=-1;	/* min/max ncol for all images */
	int nblk;			/* nblks for specific file */
	int nblkp=-1;			/* nblks for previous file */
	char lbl[PARSE_MI_LEN_LABEL+1];	/* temp array for manipulating labels */
	int tblk=-1;			/* total blocks in file */
	int first=-1;			/* first image block */
	int iblk;			/* block pointer */
	int type=-1;			/* type of mi-file - see parse.h */
	struct stat sbuf;		/* for stat'ing file */
	int needcheck;			// 1=need to check all images

				/* open file and get initial header */
	if(-1 == mi_rblock(iu,0,(char *)head,1,err)){
		fputs("Can't read mi-file header\n",ou);
		return(NULL);
	}
	parse_from_can_shortv((short *)head,256);


				/* get rest of header and get header information */
	if(head[PARSE_MI_LOC_MAGIC] == PARSE_MI_MAGIC_65K){
		type = PARSE_MI_65K;
		n = head[PARSE_MI_LOC_N];
		first = head[PARSE_MI_LOC_FIRST];
	}else{
		if(head[255]){
			if(-1 == mi_rblock(iu,1,head+256,3,err)){
				fputs("Can't read mi-file header\n",ou);
				return(NULL);
			}
			parse_from_can_shortv((short *)head+256,768);
			if(head[255] == 1) head[255] = 0;
			type = PARSE_MI_1024;
			first = 4;
		}else{
			type = PARSE_MI_256;
			for(i=256;i<1024;i++) head[i] = 0;
			first = 1;
		}

		for(i=0;i<1024;i++){
			if(head[i] == 0) break;
		}
		n = i;
	}

			/* start filling up header */
	mihead = (PARSE_MIHEAD *) parse_malloc(sizeof(PARSE_MIHEAD));
	mihead->first = first;
	mihead->type = type;
	mihead->n = n;

			/* make sure file is block aligned */
	if(-1 == fstat(fileno(iu),&sbuf)){
		fputs("parse_omi: can't stat mi-file\n",ou);
		if(err){ 
			exit(1);
		}else{
			free(mihead);
			return(NULL);
		}
	}
	tblk = sbuf.st_size / 512;
	if(tblk * 512 != sbuf.st_size){
		if(type == PARSE_MI_65K){
			fputs("parse_omi: mi-file is not block aligned\n",ou);
			if(err){ 
				exit(1);
			}else{
				free(mihead);
				return(NULL);
			}
		}
	}

	mihead->tblk = tblk;

			/* allocate arrays if necessary */
	if(n == 0 || writeonly == 1){
		mihead->ncol = NULL;
		mihead->nrow = NULL;
		mihead->bptr = NULL;
		mihead->nblk = NULL;
		mihead->label = NULL;
		return(mihead);
	}else{
		mihead->ncol = (int *) parse_malloc(sizeof(int) * n);
		mihead->nrow = (int *) parse_malloc(sizeof(int) * n);
		mihead->bptr = (long unsigned int *) parse_malloc(sizeof(long) * n);
		mihead->nblk = (long unsigned int *) parse_malloc(sizeof(long) * n);
		mihead->label = (char **) parse_malloc(sizeof(char *) * n);
		for(i=0;i<n;i++){
			mihead->label[i] = NULL; 
		}
	}

					// editor tag: peep
	if(type == PARSE_MI_65K){
		iblk = first;
		contig = 1;
		order = 1;

		needcheck = 1;
		if(checklimit > 0 && n > checklimit){
			if(-1 == mi_rblock(iu,iblk,ihead,1,err)){
				fputs("Can't read mi-file image headers\n",ou);
				parse_frmi(mihead);
				return(NULL);
			}
			ncol = parse_from_can_short(ihead[PARSE_MI_LOC_NCOL]);
			nrow = parse_from_can_short(ihead[PARSE_MI_LOC_NROW]);
			nblk = PARSE_BLOCK_SHORT(ncol*nrow);
			if(n*(nblk+1)+1 == tblk){
				needcheck = 0;
				max_ncol = min_ncol = ncol;
				max_nrow = min_nrow = nrow;
				for(i=0;i<n;i++){
					mihead->nrow[i] = nrow;
					mihead->ncol[i] = ncol;
					mihead->bptr[i] = iblk;
					mihead->nblk[i] = nblk;

					iblk += 1 + nblk;
				}
			}
		}

		if(needcheck){
			for(i=0;i<n;i++){
				if(-1 == mi_rblock(iu,iblk,ihead,1,err)){
					fputs("Can't read mi-file image headers\n",ou);
					parse_frmi(mihead);
					return(NULL);
				}
				ncol = parse_from_can_short(ihead[PARSE_MI_LOC_NCOL]);
				nrow = parse_from_can_short(ihead[PARSE_MI_LOC_NROW]);
				nblk = PARSE_BLOCK_SHORT(ncol*nrow);

				if(i==0){
					max_ncol = min_ncol = ncol;
					max_nrow = min_nrow = nrow;
				}else{
					if(ncol > max_ncol) max_ncol = ncol;
					if(ncol < min_ncol) min_ncol = ncol;
					if(nrow > max_nrow) max_nrow = nrow;
					if(nrow < min_nrow) min_nrow = nrow;
				}
				mihead->nrow[i] = nrow;
				mihead->ncol[i] = ncol;
				mihead->bptr[i] = iblk;
				mihead->nblk[i] = nblk;

				memcpy(lbl,(char *)&ihead[PARSE_MI_LOC_LABEL],
				  PARSE_MI_LEN_LABEL);
				lbl[PARSE_MI_LEN_LABEL] = 0;
				mihead->label[i] = parse_malloc_char(strlen(lbl)+1);
				strcpy(mihead->label[i],lbl);

				iblk += 1 + nblk;
			}
		}
	}else{
		for(i=0;i<n;i++){
			if(-1 == mi_rblock(iu,head[i],ihead,1,err)){
				fputs("Can't read mi-file image headers\n",ou);
				parse_frmi(mihead);
				return(NULL);
			}
			ncol = parse_from_can_short(ihead[PARSE_MI_LOC_NCOL]);
			nrow = parse_from_can_short(ihead[PARSE_MI_LOC_NROW]);
			nblk = PARSE_BLOCK_SHORT(ncol*nrow);

			if(i==0){
				contig = 1;
				order = 1;
				max_ncol = min_ncol = ncol;
				max_nrow = min_nrow = nrow;
			}else{
				if(ncol > max_ncol) max_ncol = ncol;
				if(ncol < min_ncol) min_ncol = ncol;
				if(nrow > max_nrow) max_nrow = nrow;
				if(nrow < min_nrow) min_nrow = nrow;
				if(head[i] <= head[i-1]) order = 0;
				if(head[i-1]+nblkp+1 != head[i]) contig = 0;
			}
			mihead->nrow[i] = nrow;
			mihead->ncol[i] = ncol;
			mihead->bptr[i] = head[i];
			mihead->nblk[i] = nblk;

			memcpy(lbl,(char *)&ihead[PARSE_MI_LOC_LABEL],
			  PARSE_MI_LEN_LABEL);
			lbl[PARSE_MI_LEN_LABEL] = 0;
			mihead->label[i] = parse_malloc_char(strlen(lbl)+1);
			strcpy(mihead->label[i],lbl);

			nblkp  = nblk;
		}
	}

	same = 1;
	nrow = mihead->nrow[0];
	ncol = mihead->ncol[0];
	for(i=1;i<n;i++){
		if(nrow != mihead->nrow[i]){
			same = 0;
			break;
		}
		if(ncol != mihead->ncol[i]){
			same = 0;
			break;
		}
	}

	mihead->same = same; 
	mihead->contig = contig; 
	mihead->order = order; 
	mihead->max_ncol = max_ncol;
	mihead->min_ncol = min_ncol;
	mihead->max_nrow = max_nrow;
	mihead->min_nrow = min_nrow;

	return(mihead);
}

PARSE_MIHEAD *
parse_omi(FILE *iu,int writeonly)
{
	return(parse_omi_fe(iu,writeonly,stderr,1));
}
