/*
options - handle options defined in PARSE_STRUCT

parse_pparse - install one option defined in string
parse_pparse_aa - install one option from argc/argv context
parse_opexist - check existence of option
parse_option_help - print syntax of option table
*/

#ident "$Id: options.c,v 1.7 2009/11/19 13:07:49 carl Exp $"

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "parse.h"

static int *opexist = NULL;		/* existence array */
static int opexist_size;		/* size of existence array */
static struct PARSE_STRUCT *table_save;	/* save for opexist */

static int 
install(struct PARSE_STRUCT *parse_table, int m, char *arg)
{
	double res;
	int val;
	struct PARSE_STRUCT *hp = parse_table;
	int mask = ((1 << PARSE_SHIFT) - 1);

	while(hp->label){
	  if(!strcmp((parse_table+m)->label,hp->label)){

		switch(mask & hp->type){

		case PARSE_CHAR:
			*((char **)hp->pointer) = arg;
			break;
		case PARSE_STRING:
			strcpy((char *)hp->pointer,arg);
			break;
		case PARSE_INT:
			res = parse_atof(arg);
			if(parse_atof_error()) return(-1);
			*((int *)hp->pointer) = (int) res;
			break;
		case PARSE_FLOAT:
			res = parse_atof(arg);
			if(parse_atof_error()) return(-1);
			*((float *)hp->pointer) = res;
			break;
		case PARSE_LONG:
			res = parse_atof(arg);
			if(parse_atof_error()) return(-1);
			*((long *)hp->pointer) = (long) res;
			break;
		case PARSE_TOGGLE:
			if(!strcmp(arg,"yes") || 
			   !strcmp(arg,"y") || 
			   !strcmp(arg,"YES") || 
			   !strcmp(arg,"Y") || 
			   !strcmp(arg,"1")){
				val = 1;
			}else if(!strcmp(arg,"no") || 
			   !strcmp(arg,"n") || 
			   !strcmp(arg,"NO") || 
			   !strcmp(arg,"N") || 
			   !strcmp(arg,"0")){
				val = 0;
			}else{
				fprintf(stderr,"%s bad argument: %s\n",
					hp->label,arg);
				exit(1);
			}
			*((int *)hp->pointer) = val;
			break;
		case PARSE_PROC:
				/* see 'C traps and pitfalls' */
			(*(void(*)(char *,char *))(hp->pointer))(hp->label,arg);
			break;
		case PARSE_SET:
			*((int *)hp->pointer) = 1;
			break;
		case PARSE_CLR:
			*((int *)hp->pointer) = 0;
			break;
		case PARSE_INC:
			(*((int *)hp->pointer))++;
			break;
		case PARSE_DCR:
			(*((int *)hp->pointer))--;
			break;
		case PARSE_VAL:
			*((int *)hp->pointer) = hp->type >> PARSE_SHIFT;
			break;
		default:		
			fprintf(stderr,"parse_options: label: %s bad type: %d\n",
				hp->label,mask & hp->type);
			exit(1);
		}
	  }
	  hp++;
	}
	return(m+1);
}

static int 
first_opt (		/* see if first occurrence of label */
    struct PARSE_STRUCT *pt,
    int i
)
{
	int m=0;
	while(pt[m].label){
		if(!strcmp(pt[m].label,pt[i].label)) return( i == m );
		m++;
	}
	parse_er("parse (first_opt): should not be here");
	return(-1);
}

static int 
lookup (char *s, int sl, struct PARSE_STRUCT *hp, FILE *fd)
{
	struct PARSE_STRUCT *hpp;
	int m = -1;
	int i;

	for(hpp=hp,i=0;hpp->label;hpp++,i++) {
		if(!strncmp(s,hp[i].label,sl)){
			if((int) strlen(hp[i].label) == sl){
				return(i);
			}
			if(m == -1){
				m = i;
			}else if(m >= 0){
				if(strcmp(hp[m].label,hp[i].label)) m = -2;
			}
		}
	}
	if(m == -1){
		fprintf(fd,"Illegal option: %.*s\n",sl,s);
		return(-1);
	}else if(m == -2){
		fprintf(fd,"Abbreviation, %.*s, matches:",sl,s);
		for(hpp=hp,i=0;hpp->label;hpp++,i++){
			if(!strncmp(s,hp[i].label,sl) && first_opt(hp,i)){
				fprintf(fd," %s",hp[i].label);
			}
		}
		fputc('\n',fd);
		return(-1);
	}else{
		return(m);
	}
}

int 
parse_opexist (char *op)
{
	int i;

	if(opexist == NULL){
		return(0);
	}else if(op == NULL){
		if(opexist != NULL) parse_free((char *)opexist);
		opexist = NULL;
		return(0);
	}else{
		i = 0;
		while(table_save[i].label != NULL){
		  	if(!strcmp(table_save[i].label,op)) return(opexist[i]);
			i++;
		}
		parse_prs("*parse_opexist: unknown flag:",op);
	}
	parse_er("parse (parse_opexist): should not be here");
	return(-1);
}

int 
parse_pparse_aa (struct PARSE_STRUCT parse_table[], FILE *fd, int *argc, char ***argv)
{
	int cmdl=-1;
	int need_lookup,m;
	char *cmd,*arg;
	struct PARSE_STRUCT *hp;
/*
	check for noop
*/
	if(*argc < 1) return(0);
/*
	check for help
*/
	cmd = **argv;
	if(*cmd == '?'){
		hp = parse_table;
		while(1){
			if(!hp->label){
				parse_option_help(parse_table,fd);
				return(-1);
			}else if(!strcmp("?",hp->label)){
						/* see 'C traps and pitfalls' */
				(*(void(*)())(hp->pointer))();
				return(-1);
			}else{
				hp++;
			}
		}
	}
/*
	set up error paths
*/
	if(fd != stderr && fd != stdout) parse_atof_args(fd,1);
/*
	create opexist array of necessary
*/
	m = 0;		
	while(1){
		if(!parse_table[m].label) break;
		m++;
	}

	if(opexist == NULL || opexist_size != m){
		if(opexist != NULL) parse_free((char *)opexist);
		table_save = parse_table;
		opexist = (int *)parse_malloc(m*sizeof(int));
		opexist_size = m;
		while(m--) opexist[m] = 0;
	}
/*
	check for command split from = or just an equals sign
*/
	need_lookup = 1;
	if(NULL == (arg = strchr(cmd,'='))){
		m = 0;
		while(1){
			if(!parse_table[m].label){
				fprintf(fd,"missing equals sign: %s\n",cmd);
				return(-1);
		  	}else if(!strcmp(parse_table[m].label,"=")){
				need_lookup = 0;
				arg = cmd;
				break;
			}
			m++;
		}
	}else{
		cmdl = arg - cmd;
		arg++;
		if(!*arg){
			if(*argc == 1){
				fprintf(fd,
					"empty string or no next arg: %.*s\n",
					cmdl,cmd);
				return(-1);
			}else{
				(*argc)--;
				(*argv)++;
				arg = **argv;
			}
		}
	}
/*
	do it!
*/
	if(need_lookup){
		if(cmdl == -1) parse_er("parse_pparse_aa: cmdl == -1");
		m = lookup(cmd,cmdl,parse_table,fd);
		if(m == -1) return(m);
	}
	opexist[m] = 1;
	return(install(parse_table,m,arg));
}

int 
parse_pparse (struct PARSE_STRUCT parse_table[], char *s, FILE *fd)
{
	int argc;
	char **argv,*args[1];

	argc = 1;
	args[0] = s;
	argv = args;
	return(parse_pparse_aa(parse_table,fd,&argc,&argv));
}


void 
parse_option_help (struct PARSE_STRUCT parse_table[], FILE *fd)
{
	int i;
	int type;
	int mask = ((1 << PARSE_SHIFT) - 1);

	if(!parse_table[0].label){
		fputs("No options.\n",fd);
		return;
	}

	for(i=0;parse_table[i].label;i++){
	  if(!strcmp(parse_table[i].label,"=") 
	    && first_opt(parse_table,i)
	    && parse_table[i].help){
		fprintf(fd,"(empty) equivalent to: %s\n",parse_table[i].help);
	  }else if(strcmp(parse_table[i].label,"?") 
	    && strcmp(parse_table[i].label,"=") 
	    && *parse_table[i].label
	    && first_opt(parse_table,i)){
		fprintf(fd,"%s=",parse_table[i].label);
		type = mask & parse_table[i].type;
		switch(type){

		case PARSE_CHAR:
			if(!*((char **)parse_table[i].pointer)){
				fprintf(fd,"%s","(null)");
			}else{
				fprintf(fd,"%s",*((char **)parse_table[i].pointer));
			}
			break;
		case PARSE_STRING:
			fprintf(fd,"%s",(char *)parse_table[i].pointer);
			break;
		case PARSE_INT:
			fprintf(fd,"%d",*((int *)parse_table[i].pointer));
			break;
		case PARSE_FLOAT:
			fprintf(fd,"%f",
				*((float *)parse_table[i].pointer));
			break;
		case PARSE_LONG:
			fprintf(fd,"%ld",
				*((long *)parse_table[i].pointer));
			break;
		case PARSE_TOGGLE:
			fprintf(fd,"%s",
				(*((long *)parse_table[i].pointer))?"yes":"no");
			break;
		case PARSE_PROC:
			fprintf(fd,"%s","(calls procedure)");
			break;
		case PARSE_SET:
			fprintf(fd,"%d %s",
				*((int *)parse_table[i].pointer),
				"(sets flag)");
			break;
		case PARSE_CLR:
			fprintf(fd,"%d %s",
				*((int *)parse_table[i].pointer),
				"(clears flag)");
			break;
		case PARSE_INC:
			fprintf(fd,"%d %s",
				*((int *)parse_table[i].pointer),
				"(increments flag)");
			break;
		case PARSE_DCR:
			fprintf(fd,"%d %s",
				*((int *)parse_table[i].pointer),
				"(decrements flag)");
			break;
		case PARSE_VAL:
			fprintf(fd,"%d (sets flag to %d)",
				*((int *)parse_table[i].pointer),
				parse_table[i].type >> PARSE_SHIFT);
			break;
		default:		
			fprintf(stderr,"parse_options: label: %s bad type: %d\n",
				parse_table[i].label,type);
			exit(1);
		}

		if(parse_table[i].help){
			fputs(": ",fd);
			fputs(parse_table[i].help,fd);
		}
		fputc('\n',fd);
	  }
	}
}

