/*
string - manipulate strings
*/

#ident "$Id: string.c,v 1.7 2007/05/22 15:51:51 ccrawford Exp $"

#include "parse.h"
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#ifndef NT
#include <pwd.h>
#include <unistd.h>
#include <sys/utsname.h>
#endif

#define LPATH 256		/* path storage for parse_root */


void 
parse_toupper (char *s)
{
	while(*s){
		if(islower(*s)) *s = toupper(*s);
		s++;
	}
}

void 
parse_tolower (char *s)
{
	while(*s){
		if(isupper(*s)) *s = tolower(*s);
		s++;
	}
}

void 
parse_words (
    char *s,	/*input string*/
    int *argc,	/*word count*/
    char **argv,	/*word pointers*/
    int mword	/*maximum number of words*/
)
{
	char *l;

        *argc = 0;                
	l = s;
	while(1){
		while(*l == ' ' || *l == '\t') l++;
		if(!*l) return;
		if(*argc > mword) parse_err("parse_words: ","too many words");
		argv[(*argc)++] = l;
		while(*l && *l != ' ' && *l != '\t') l++;
		if(!*l) return;
		*l++ = '\0';
	}
}

char *
parse_itoa (int i, char *s)
{
	char *is;

	is = (s == NULL)? ((char *) parse_malloc(PARSE_ITOA)) : s;

	sprintf(is,"%d",i);
	return(is);
}

char *
parse_suffix_malloc (char *n, char *s)
{
	char *p;

	if((p=strrchr(n,'/')) == NULL) p = n;
	if(strchr(p,'.') == NULL){
		p = parse_malloc_char(strlen(s)+strlen(n)+2);
		strcpy(p,n);
		strcat(p,".");
		strcat(p,s);
	}else{
		p = parse_malloc_char(strlen(n)+1);
		strcpy(p,n);
	}
	return(p);
}

void 
parse_suffix (char *n, char *s)
{
	char *p;
	if((p=strrchr(n,'/')) == NULL) p = n;
	if(strchr(p,'.') == NULL){
		strcat(n,".");
		strcat(n,s);
	}
}

void 
parse_remove_suffix (char *n)
{
	char *p;
	if((p=strrchr(n,'/')) == NULL) p = n;
	if((p = strchr(p,'.')) != NULL) *p = 0;
}

char *
parse_extract_suffix(char *n)
{
	char *p;
	char *s;

	if((p=strrchr(n,'/')) == NULL) p = n;
	if((p = strchr(p,'.')) != NULL){
		p++;
		if(*p){
			s = parse_malloc_char(strlen(p)+1);
			strcpy(s,p);
			return(s);
		}else{
			return(NULL);
		}
	}else{
		return(NULL);
	}
}

#ifdef TEST_EXTRACT_SUFFIX
int 
main (void)
{
	char fn[256];
	char *s;
	fn[0] = 0;

	while(1){
		parse_acs("filename",fn,NULL,fn);
		s = parse_extract_suffix(fn);
		if(s == NULL){
			parse_wtext("no suffix!");
		}else{
			parse_prs("suffix:",s);
		}
	}
}
#endif

#ifndef NT
char *
parse_root (char *s)
{
	static char path[LPATH];
	struct passwd *pw;
	char *sp;
	int uid;

	if(*s != '~') return(strncpy(path,s,LPATH));
	if(isalpha(s[1])){
		sp = ++s;
		while(isalnum(*s)) s++;
		strncpy(path,sp,s-sp);
		path[s-sp] = 0;
		if(NULL == (pw = getpwnam(path))){
			parse_err("parse_root: don't know user: ",path);
		}
	}else{
		uid = geteuid();
		if(NULL == (pw = getpwuid(uid))){
			fprintf(stderr,"parse_root: don't know uid: %d\n",uid);
			exit(1);
		}
		s++;
	}
	strcpy(path,pw->pw_dir);
	strncat(path,s,LPATH-strlen(path));
	return(path);
}

char *
parse_host(void)
{
	static struct utsname buf;

	uname(&buf);
	return(&buf.nodename[0]);
}
#endif

