/*
plot3d - program to plot three dimensional surfaces
*/

#ident "$Id: plot3d.c,v 1.11 2011/09/14 15:41:12 carl Exp $"

#include <math.h>
#include <string.h>
#include <parse.h>
#include <crcplot.h>

/* 
	file types:
		0: signed char
		1: unsigned char
		2: short
		3: int
		4: long
		5: float
		6: double
		7: ascii/text
		8: da-file
		9: mi-file
*/

static int pxaxis=1;		/* 1=plot x axis */
static int pyaxis=1;		/* 1=plot y axis */
static int pzaxis=1;		/* 1=plot the z-axis*/
static int use[2];		/* 1=use x or y file */
static int reverse;		/* 1=reverse z direction in plot */
static float scfac=1.0;		/* scale factor in plot */
static char *xyf[2]={"x","y"};	/* default x and y input file */
static char *zf="z";		/* default z input file */
static float xxs,yys;		/* starting locations for the graph */
static char *axslbl[3];		/* axes labels */
static char *tl;		/* top label */
static char *bl;		/* bottom label */
static float maxmin[6];		/* command string max,min values */
static int umxmn[6];		/* 1=corresponding maxmin value entered */
static int byte[3]={8,8,8};	/* default type for x, y and z vectors */
static int begin[2];		/* starting x and y coord of interest */
static int skip[2];		/* points in x and y to skip */
static int xysize[2]={64,64};	/* # of points in matrix for x and y indices */
static int count[3];		/* vector points in x and y direction to use*/
static float len[3]={6,6,4};	/* length of x, y and z axes */
static int frame;		/* 1=frame plot with border */
static int erase=1;		/* 1=earse screen */
static int ydir=1;		/* 1=print y lines (crosshatch) */
static int xdir;		/* 1=print x lines (crosshatch) */
static float phi1=45.0;		/* orientation of x-y plane */
static float phi2=30.0;		/* tilt of plane */
static float resol=1.0;		/* determines # of steps for line hiding */ 
static char *dirn="y";		/* default direction for line drawings */
static char *axisn="xyz";	/* default for axes drawings */
static int use_stdout;		/* 1=send output to stdout */
static float *zd,*xd,*yd;	/* arrays of x, y and z values */
static float zmax,zmin;		/* from matrix input */
static char *host;		/* host to connect to */

static void fn_proc(char *option,char *arg);

struct PARSE_STRUCT parse_table[] = {
	{"resolution",PARSE_FLOAT,(PARSE_ARG)&resol,"surface resolution"},
	{"phi1",PARSE_FLOAT,(PARSE_ARG)&phi1,"z-axis rotation"},
	{"rz",PARSE_FLOAT,(PARSE_ARG)&phi1,"z-axis rotation"},
	{"phi2",PARSE_FLOAT,(PARSE_ARG)&phi2,"horizontal rotation"},
	{"rx",PARSE_FLOAT,(PARSE_ARG)&phi2,"horizontal rotation"},
	{"direction",PARSE_CHAR,(PARSE_ARG)&dirn,"line orientation [x,y]"},
	{"axis",PARSE_CHAR,(PARSE_ARG)&axisn,"axes to draw [x,y,z]"},
	{"x",PARSE_CHAR,(PARSE_ARG)&xyf[0],"name of x-vector"},
	{"x",PARSE_PROC,(PARSE_ARG)fn_proc},
	{"x",PARSE_SET,(PARSE_ARG)(use+0)},
	{"y",PARSE_CHAR,(PARSE_ARG)&xyf[1],"name of y-vector"},
	{"y",PARSE_PROC,(PARSE_ARG)fn_proc},
	{"y",PARSE_SET,(PARSE_ARG)(use+1)},
	{"z",PARSE_CHAR,(PARSE_ARG)&zf,"name of z-matrix"},
	{"z",PARSE_PROC,(PARSE_ARG)fn_proc},
	{"=",PARSE_PROC,(PARSE_ARG)fn_proc,"z"},
	{"xp",PARSE_FLOAT,(PARSE_ARG)&xxs,"x-position of plot"},
	{"yp",PARSE_FLOAT,(PARSE_ARG)&yys,"y-position of plot"},
	{"xmin",PARSE_FLOAT,(PARSE_ARG)&maxmin[0],"minimum of x-vector"},
	{"xmin",PARSE_SET,(PARSE_ARG)(umxmn+0)},
	{"xmax",PARSE_FLOAT,(PARSE_ARG)&maxmin[1],"maximum of x-vector"},
	{"xmax",PARSE_SET,(PARSE_ARG)(umxmn+1)},
	{"ymin",PARSE_FLOAT,(PARSE_ARG)&maxmin[2],"minimum of y-vector"},
	{"ymin",PARSE_SET,(PARSE_ARG)(umxmn+2)},
	{"ymax",PARSE_FLOAT,(PARSE_ARG)&maxmin[3],"maximum of y-vector"},
	{"ymax",PARSE_SET,(PARSE_ARG)(umxmn+3)},
	{"zmin",PARSE_FLOAT,(PARSE_ARG)&maxmin[4],"minimum of z-vector"},
	{"zmin",PARSE_SET,(PARSE_ARG)(umxmn+4)},
	{"zmax",PARSE_FLOAT,(PARSE_ARG)&maxmin[5],"maximum of z-vector"},
	{"zmax",PARSE_SET,(PARSE_ARG)(umxmn+5)},
	{"xcount",PARSE_INT,(PARSE_ARG)&count[0],"points to plot in x-vector"},
	{"ycount",PARSE_INT,(PARSE_ARG)&count[1],"points to plot in y-vector"},
	{"scale",PARSE_FLOAT,(PARSE_ARG)&scfac,"scale factor"},
	{"scfac",PARSE_FLOAT,(PARSE_ARG)&scfac,"scale factor"},
	{"xlen",PARSE_FLOAT,(PARSE_ARG)&len[0],"length of x-axis"},
	{"ylen",PARSE_FLOAT,(PARSE_ARG)&len[1],"length of y-axis"},
	{"zlen",PARSE_FLOAT,(PARSE_ARG)&len[2],"length of z-axis"},
	{"len",PARSE_FLOAT,(PARSE_ARG)&len[0],"axes length"},
	{"len",PARSE_FLOAT,(PARSE_ARG)&len[1]},
	{"len",PARSE_FLOAT,(PARSE_ARG)&len[2]},
	{"n",PARSE_INT,(PARSE_ARG)&xysize[0],"set x and y size"},
	{"n",PARSE_INT,(PARSE_ARG)&xysize[1]},
	{"count",PARSE_INT,(PARSE_ARG)&count[0],"set x and y counts"},
	{"count",PARSE_INT,(PARSE_ARG)&count[1]},
	{"xbegin",PARSE_INT,(PARSE_ARG)&begin[0],"first x point"},
	{"ybegin",PARSE_INT,(PARSE_ARG)&begin[1],"first y point"},
	{"begin",PARSE_INT,(PARSE_ARG)&begin[0],"first point in x,y files"},
	{"begin",PARSE_INT,(PARSE_ARG)&begin[1]},
	{"skip",PARSE_INT,(PARSE_ARG)&skip[0],"x and y skip values"},
	{"skip",PARSE_INT,(PARSE_ARG)&skip[1]},
	{"xskip",PARSE_INT,(PARSE_ARG)&skip[0],"x skip value"},
	{"yskip",PARSE_INT,(PARSE_ARG)&skip[1],"y skip value"},
	{"size",PARSE_INT,(PARSE_ARG)&xysize[0],"set and y sizes"},
	{"size",PARSE_INT,(PARSE_ARG)&xysize[1]},
	{"xsize",PARSE_INT,(PARSE_ARG)&xysize[0],"set x size"},
	{"ysize",PARSE_INT,(PARSE_ARG)&xysize[1],"set y size"},
	{"xl",PARSE_CHAR,(PARSE_ARG)&axslbl[0],"x axis label"},
	{"yl",PARSE_CHAR,(PARSE_ARG)&axslbl[1],"y axis label"},
	{"zl",PARSE_CHAR,(PARSE_ARG)&axslbl[2],"z axis label"},
	{"tl",PARSE_CHAR,(PARSE_ARG)&tl,"top label"},
	{"bl",PARSE_CHAR,(PARSE_ARG)&bl,"bottom label"},
	{"host",PARSE_CHAR,(PARSE_ARG)&host,"host[:port]"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0,0}
	};

int     size[] = {
		sizeof(char),   /* signed char */
		sizeof(char),   /* unsigned char */
		sizeof(short),  /* short integer */
		sizeof(int),    /* integer */
		sizeof(long),   /* long integer */
		sizeof(float),  /* float */
		sizeof(double), /* double */
	};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'a',&pxaxis,PARSE_FLAG_CLR,"don't plot axes"},
	{'a',&pyaxis,PARSE_FLAG_CLR,"don't plot axes"},
	{'a',&pzaxis,PARSE_FLAG_CLR,"don't plot axes"},
	{'e',&erase,PARSE_FLAG_CLR,"don't erase screen"},
	{'f',&frame,PARSE_FLAG_SET,"frame plot"},
	{'o',&use_stdout,PARSE_FLAG_SET,"send plot output to stdout"},
	{'r',&reverse,PARSE_FLAG_SET,"reverse plot"},
	{'x',use+0,PARSE_FLAG_SET,"use x-vector"},
	{'y',use+1,PARSE_FLAG_SET,"use y-vector"},
	{'?',(int *)parse_help_basic,PARSE_PROC},
	{0,0,0,0}
};

static int
matrixinput()
{
	char *name = zf;	
	int n = byte[2];
	int i,ns,ix,iy,j;
	int k,nblk,iblk=-1,itype;
	FILE *fd;
	char buf_ii[sizeof(double)];
	float *vec;
	long *da;
	short head[256];
	int zminmaxs=0;		/* =1 if zmin/max set */
	void *buf;

	buf = (void *) buf_ii;

	if(n == 7)name = parse_suffix_malloc(name,"as");
	if(n == 8)name = parse_suffix_malloc(name,"da");
	if(n == 9)name = parse_suffix_malloc(name,"mi");
	fd = parse_fopen(name,"r");
	if(n == 7 || n == 8 || n == 9)parse_free(name);

	if(n == 8){
		parse_rblock(fd,0,(char *)head,1);
		xysize[0] = parse_from_can_short(head[0]);
		xysize[1] = parse_from_can_short(head[1]);
		itype = parse_from_can_short(head[2]);
		if(itype)parse_er("plot3d: can't handle short da-files");
	}

	if(n == 9){
		parse_rblock(fd,0,(char *)head,1);
		iblk = parse_from_can_short(head[0]);
		parse_rblock(fd,iblk,(char *)head,1);
		xysize[0] = parse_from_can_short(head[40]);
		xysize[1] = parse_from_can_short(head[41]);
	}

	for(k=0;k<2;k++){
		    j = (xysize[k]-begin[k])/(double)(skip[k]+1)+.999999;
		if(count[k] == 0){
			count[k] = j;
		}else if(count[k] > j){
			count[k] = j;
		}
	}

	zd = parse_malloc_float(count[0]*count[1]);

	vec = zd;
	if(n == 8){
		nblk = (parse_from_can_short(head[0]) + 127 ) / 128;
		da = parse_malloc_long(nblk*512/sizeof(float));
		iblk = begin[1] * nblk + 1;
		for(k=0;k<count[1];k++){
			parse_rblock(fd,iblk,(char *)da,nblk);
			for(i=0,j=begin[0];i<count[0];i++){
				*vec = parse_from_can_float(da[j]);
				if(reverse) *vec *= -1.0;
				if(zminmaxs){
					if(*vec < zmin) zmin = *vec;
					if(*vec > zmax) zmax = *vec;
				}else{
					zminmaxs = 1;
					zmin = *vec;
					zmax = *vec;
				}
				vec++;
				j += skip[0] + 1;
			}
			iblk += (skip[1] + 1) * nblk;
		}
		i = count[0] * count[1];
		parse_free((char *)da);
	}else if(n == 7){
		int	SeekTo, numcount;
		float	xtmp;

		numcount = i = 0;
		for (iy=0;iy<count[1] && !feof(fd);iy++){
			SeekTo = ((int)iy*(skip[1]+1)+begin[1])*xysize[0]
								+ begin[0];
			while(numcount < SeekTo){
				if(EOF == fscanf(fd,"%f",&xtmp)) perror("plot3d: fscaf error");
				numcount++;
			}
			for (ix=0;ix<count[0] && !feof(fd);ix++){
				if(EOF == fscanf(fd,"%f",&xtmp)) perror("plot3d: fscanf error");
				*vec = xtmp;
				if(reverse) *vec *= -1.0;
				if(*vec < zmin) zmin = *vec;
				if(*vec > zmax) zmax = *vec;
				vec++;
				i++;
				numcount++;
				if( ix < count[0]-1 )
					for (j=0;j<skip[0]&&fscanf(fd,"%f"
							,&xtmp)>0;j++)
						numcount++;
			}
		}
	}
	else {
		if(n == 9)
			ns = sizeof(short);
		else
			ns = size[n];
		i = 0;

		for (iy=0;iy<count[1];iy++){
			fseek(fd,(((long)iy*(skip[1]+1)+begin[1])*xysize[0]
							+ begin[0])*ns,0);
			if(n == 9){
				fseek(fd,(long)iblk*512+512,1);
				vec = zd + (count[1]-iy-1)*count[0];
			}
			for (ix=0;ix<count[0];ix++){
				if(0 == fread(buf,ns,1,fd)) break;
				i++;
				switch(n){

				case 0: /* signed char */
					*vec = *((char *)buf);
					break;
				case 1: /* unsigned char */
					*vec = *((char *)buf) & 0377;
					break;
				case 2: /* short */
				case 9: /* mi file */
					*vec = parse_from_can_short(
						*((short *)buf));
					break;
				case 3: /* int */
					*vec = *((int *)buf);
					break;
				case 4: /* long */
					*vec = *((long *)buf);
					break;
				case 5: /* float */
					*vec = *((float *)buf);
					break;
				case 6: /* double */
					*vec = *((double *)buf);
					break;
				}
				if(reverse) *vec *= -1.0;
				if(*vec < zmin) zmin = *vec;
				if(*vec > zmax) zmax = *vec;
				vec++;
				fseek(fd,(long)(skip[0]*ns),1);
			}
		}
	}

	if(!i) parse_er("no points read from input file");
	parse_fclose(fd);
	if(zmax == zmin){
		printf("plot3d: Entire matrix equals %g.\n",zmax);
		exit(2);
	}
	return(i);
}

static int
input(name,n,vec,count,begin,skip)
char *name;
int n;
int count,begin,skip;
float *vec;
{
	long *da;
	int j,i,ns;
	FILE *fd;
	char buf_ii[sizeof(double)];
	short head[256];
	int itype,iblk,nblk;
	void *buf;

	buf = (void *) buf_ii;

	if(n == 7)name = parse_suffix_malloc(name,"as");
	if(n == 8)name = parse_suffix_malloc(name,"da");
	fd = parse_fopen(name,"r");
	if(n == 7 || n == 8)parse_free(name);

	if(n == 8){		/* da-files */
		parse_rblock(fd,0,(char *)head,1);
		nblk = (parse_from_can_short(head[0]) + 127 ) / 128;
		itype = parse_from_can_short(head[2]);
		if(itype)parse_er("plot3d: can't handle short da-files");
		da = parse_malloc_long(nblk*512/sizeof(float));
		iblk = 1;
		parse_rblock(fd,iblk,(char *)da,nblk);
		for(i=0,j=begin;i<count;i++){
			*vec = parse_from_can_float(da[j]);
			vec++;
			j += skip + 1;
		}
		i = count;
		parse_free((char *)da);
	}else if(n == 7){
		int	j;
		float	tmp;

		i = begin;
		while(i-- > 0 && fscanf(fd,"%f",&tmp) > 0);
		i = 0;
		while(i < count && fscanf(fd,"%f",&tmp) > 0){
			*vec++ = tmp;
			i++;
			for (j=0;j<skip && fscanf(fd,"%f",&tmp) > 0;j++);
		}
	}
	else {
		ns = size[n];
		i = 0;
		fseek(fd,(long)(begin * ns),0);
		while(i<count && fread(buf,ns,1,fd)){
			switch(n){

			case 0: /* signed char */
				*vec++ = *((char *)buf);
				break;
			case 1: /* unsigned char */
				*vec++ = *((char *)buf) & 0377;
				break;
			case 2: /* short */
				*vec++ = *((short *)buf);
				break;
			case 3: /* int */
				*vec++ = *((int *)buf);
				break;
			case 4: /* long */
				*vec++ = *((long *)buf);
				break;
			case 5: /* float */
				*vec++ = *((float *)buf);
				break;
			case 6: /* double */
				*vec++ = *((double *)buf);
				break;
			}
			i++;
			fseek(fd,(long)(skip*ns),1);
		}
	}
	fclose(fd);
	if(i == 0){
		parse_er("no points read from input file");
	}
	return(i);
}

static void
nbyte(s,n)
char *s;
int *n;
{
	while(*s)if(*s == ',' && *(s+1)){
		*s++ = 0;
		switch(*s){

		case 'b':       /* char data */
		case 'c':
		case '1':
			if(*++s == 's')*n = 0;
			else *n = 1;
			return;
		case 's':       /* short */
		case '2':
			*n = 2;
			return;
		case 'i':       /* integer */
			*n = 3;
			return;
		case 'l':       /* long */
			*n = 4;
			return;
		case 'f':       /* float */
		case '4':
			*n = 5;
			return;
		case 'p':       /* double precision */
		case '8':
			*n = 6;
			return;
		case 'a':	/* ASCII */
			*n = 7;
			return;
		case 'd':	/* da-file */
			*n = 8;
			return;
		case 'm':	/* mi-file */
			*n = 9;
			return;
		default:
			parse_err("byte declaration error","");
		}
	}else{
		s++;
	}
}

static void
pscale(a,n,mm)
float *a;
int n;
float *mm;
{
	float ai,fmax,fmin;
	int i;

	fmax = fmin = *a;
	for(i=0;i<n;i++){
		ai=a[i];
		if(ai > fmax) fmax = ai;
		if(ai < fmin) fmin = ai;
	}
	mm[0] = fmin;
	mm[1] = fmax;
}

static void 
fn_proc(option,arg)
char *option,*arg;
{
	char **fn,*dfn;
	int b;

	if(!strcmp(option,"x")){
		fn = &xyf[0];
		dfn = "x";
		b = 0;
	}else if(!strcmp(option,"y")){
		fn = &xyf[1];
		dfn = "y";
		b = 1;
	}else{
		fn = &zf;
		dfn = "z";
		b = 2;
	}

	if(*arg == ','){
		nbyte(arg,byte+b);
		*fn = strcpy(parse_malloc_char(strlen(dfn)+1),dfn);
	}else{
		*fn = arg;
		nbyte(*fn,byte+b);
	}
}

int
main(argc,argv)
int argc;
char **argv;
{
	int i,j,k;
	float *mm,mmt[2],*xyd,dxy;
/*
	parse command line
*/
	parse_all_basic(".plot3d","PLOT3DARGS",argc,argv);
/*
	set arguments based on strings obtained from command line
*/
	if(strcmp(axisn,"xyz")){
		pxaxis = 0;
		pyaxis = 0;
		pzaxis = 0;
		for( i=0 ; *(axisn+i) != 0 ; i++ ){
			if( *(axisn+i) == 'x' ){
				pxaxis = 1;
			}else if( *(axisn+i) == 'y' ){
				pyaxis = 1;
			}else if( *(axisn+i) == 'z' ){
				pzaxis = 1;
			}else{
				parse_er("bad argument in axis=");
			}
		}
	}
		
	if(strcmp(dirn,"x") == 0){
		xdir = 1;
		ydir = 0;
	}else if(strcmp(dirn,"y") == 0){
		xdir = 0;
		ydir = 1;
	}else if( (strcmp(dirn,"xy") == 0) || (strcmp(dirn,"yx") == 0) ){
		xdir = 1;
		ydir = 1;
	}else{
		parse_er("bad argument in dir=");
	}
/*
	check and perhaps set ranges of data
*/
	if((use[0] && byte[0] == 9) || (use[1] && byte[1] == 9)){
		fprintf(stderr,"Can't use mi-format for x- or y-files\n");
		exit(2);
	}
/*
	get z matrix and set limits
*/
	count[2] = matrixinput();

	if(count[2] != count[0]*count[1]){
		fprintf(stderr,"Only read %d of %d points in z file.\n",
			count[2], count[0]*count[1]);
		exit(2);
	}

	if(umxmn[4]){
		zmin = maxmin[4];
		for(i=0;i<count[2];i++) if((zd[i] < zmin)) zd[i] = zmin;
	}
	if(umxmn[5]){
		zmax = maxmin[5];
		for(i=0;i<count[2];i++) if((zd[i] > zmax)) zd[i] = zmax;
	}
/*
	get or generate x and y vectors
*/
	xd = parse_malloc_float(count[0]);
	yd = parse_malloc_float(count[1]);

	for(k=0;k<2;k++){
		xyd = (k)? yd : xd;
		mm = (k)? maxmin+2 : maxmin;
					/* read data if necessary */
		if(use[k]){
			i = input(xyf[k],byte[k],xyd,
				  count[k],begin[k],skip[k]);
			if(i != count[k]){
			  fprintf(stderr,"Only read %d of %d points in %s\n"
							,i,count[k],xyf[k]);
			  exit(1);
			}
			pscale(xyd,count[k],mmt);

					/* override max and min values */
			for(i=0;i<2;i++)
				if(!umxmn[k*2+i]) mm[i] = mmt[i];
			if(mm[0] > mmt[0] || mm[1] < mmt[1])
				parse_er("plot extends beyond the axes");
		}

					/* generate vectors if necessary */
		if(!use[k]){

					/* min and max both set */
			if(umxmn[k*2] && umxmn[k*2+1]){
				dxy = (maxmin[k*2+1]-maxmin[k*2])/(count[k]-1);
				xyd[0] = maxmin[k*2];
				for(i=1;i<count[k];i++) xyd[i] = xyd[i-1] + dxy;

					/* only min set */
			}else if(umxmn[k*2]){
				for(i=0,j=maxmin[k*2];i<count[k];i++){
					xyd[i] = j;
					j += skip[k]+1;
				}
				mm[1] = xyd[count[k]-1];

					/* neither min nor max set */
			}else{
				for(i=0,j=begin[k];i<count[k];i++){
					xyd[i] = j;
					j += skip[k]+1;
				}
				mm[0] = xyd[0];
				mm[1] = xyd[count[k]-1];
			}
		}
	}
/*
	preparatory calls to graphics primitives
*/
	if(host){
		plots_host(host);
	}else if(use_stdout){
		plots_host("stdout");
	}else{
		plots();
	}

	if(erase) clear();
	plot(xxs,yys,-3);
	factor(scfac);
	plot(.75,1.0,-3);

	if(frame){
		plot(0.05,0.05,3);
		plot(0.05,9.95,2);
		plot(9.95,9.95,2);
		plot(9.95,0.05,2);
		plot(0.05,0.05,2);
	}
/*	
	call function that does real surface plot
*/
	plot3d(
		xd,yd,zd,
		count[0],count[1],
		maxmin[0],maxmin[1],maxmin[2],maxmin[3],zmin,zmax,
		len[0],len[1],len[2],
		phi1,phi2,resol,
		pxaxis,pyaxis,pzaxis,
		xdir,ydir,
		tl,bl,axslbl[0],axslbl[1],axslbl[2]
	);
/*
	end plotting, clean up, and exit
*/
	plot(0.0,0.0,999);
	exit(0);
}
