/*
armi - archive mi-files
*/

static char rcsid[] = "$Id: armi.c,v 1.10 2011/09/14 14:09:03 carl Exp $";

#include "parse.h"
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#define TMPFILE "armi.0x3x5x.mi"		/* temp file for delete */
#define LLEN PARSE_MI_LEN_LABEL 		/* length of label in bytes */
#define LLOC PARSE_MI_LOC_LABEL 		/* location of label */

static enum {NONE,CREATE,WRITE,LIST,DELETE,EXTRACT} mode=NONE;
static int verbose=0;
static int numspec;

static void
syntax()
{
	parse_err(
	  "syntax: armi [-]{cdrtx}[nv] archive[.mi] [num] image[.mi]...\n"
	  ,"");
}

static void
list(afn)
char *afn;
{
	int i,n,nx,ny;
	FILE *afd;
	char *label;
	PARSE_MIHEAD *mihead;

	afd = parse_fopen(afn,"r");
	mihead = parse_omi(afd,0);
	n = mihead->n;

	printf("%s: image(s): %d\n",afn,n);
	if(verbose){
		switch(mihead->type){
		case PARSE_MI_256:
			parse_wtext("type: 256"); break;
		case PARSE_MI_1024:
			parse_wtext("type: 1024"); break;
		case PARSE_MI_65K:
			parse_wtext("type: 65K"); break;
		default:
			parse_pri("*smi: illegal type:",mihead->type);
		}
		parse_prt("same size",mihead->same);
		parse_prt("contiguous",mihead->contig);
		parse_prt("in order",mihead->order);
		parse_pri("first block:",mihead->first);
		parse_pri("total blocks:",mihead->tblk);
		parse_pri("max rows:",mihead->max_nrow);
		parse_pri("min rows:",mihead->min_nrow);
		parse_pri("max columns:",mihead->max_ncol);
		parse_pri("min columns:",mihead->min_ncol);
	}

	for(i=0;i<n;i++){
		label = mihead->label[i];
		nx = mihead->ncol[i];
		ny = mihead->nrow[i];
		printf("Image: %d (%d,%d)",i+1,nx,ny);
		if(*label) printf(" \"%s\"",label);
		putchar('\n');
	}
}

static void
add(exist,afn,argc,argv)
int exist;
char *afn;
int argc;
char **argv;
{
	int j,ip=1,hp=0,opn,i,nx,ny,blk,m,ne=0;
	unsigned short head[1024];
	short aihead[256];
	FILE *afd=NULL,*ifd;
	char *ifn,*label,lspace[LLEN+1];
	char block[512];
	PARSE_MIHEAD *ihead,*ohead;

	if(exist){
		opn = 0;
	}else{
		if((afd = fopen(afn,"r+")) == NULL){
			opn = 0;
			exist = 1;
		}else{
			opn = 1;

			ohead = parse_omi(afd,1);
			if(ohead->type != PARSE_MI_65K)
				parse_er("armi: can only append to 65K mi-files");
			ne = hp = ohead->n;
			ip = ohead->tblk;
		}
	}

	while(++argv,--argc){
		ifn = parse_suffix_malloc(*argv,"mi");
		ifd = parse_fopen(ifn,"r");
		ihead = parse_omi(ifd,0);
		m = ihead->n;

		if(m) (void) printf("adding: %s - image(s): %d\n",ifn,m);
		for(i=0;i<m;i++){
			parse_rblock(ifd,ihead->bptr[i],aihead,1);

			label = ihead->label[i];
			nx = ihead->ncol[i];
			ny = ihead->nrow[i];
			blk = ihead->nblk[i];

			printf("\tImage: %d (%d,%d)",i+1,nx,ny);
			if(*label) (void) printf(" \"%s\"",label);
			putchar('\n');

			if(!*label){
				label = lspace;
				label[LLEN] = 0;
				strncpy(label,ifn,LLEN);
				strncat(label," - image: ",LLEN-strlen(label));
				strncat(label,parse_itoa(i+1,(char *)NULL),LLEN-strlen(label));
				strcpy((char *)&aihead[LLOC],label);
			}

			if(!opn){
				afd = parse_fopen(afn,"w");
				opn = 1;
			}

			hp++;
			parse_wblock(afd,ip++,(char *)aihead,1);
			for(j=0;j<blk;j++){
				parse_rblock(ifd,ihead->bptr[i]+j+1,(char *)block,1);
				parse_wblock(afd,ip++,(char *)block,1);
			}
		}
		parse_fclose(ifd);
	}
	if((exist && hp) || (!exist && hp>ne)){
		parse_zbuf((short *)head,256);
		head[PARSE_MI_LOC_MAGIC] = PARSE_MI_MAGIC_65K;
		head[PARSE_MI_LOC_N] = hp;
		head[PARSE_MI_LOC_FIRST] = 1;
		parse_to_can_shortv((short *) head,256);
		parse_wblock(afd,0,(char *)head,1);
		if(exist){
			printf("created: %s - total image(s) %d\n",afn,hp);
		}else{
			printf("appended: %s - appended/total image(s) %d/%d\n",afn,hp-ne,hp);
		}
	}
}

static void
extract(afn,argc,argv)
char *afn;
int argc;
char **argv;
{
	int num,i,n,nx,ny;
	FILE *afd;
	char *ofn;
	PARSE_MIHEAD *ihead;
	unsigned short *a;
	char *label;
	int blk;

	if(numspec){
		++argv;
		--argc;
		if(argc == 0 || !isdigit(*argv[0])) syntax();
		num = parse_atof(*argv);
	}else{
		num = 1;
	}
	if(argc == 1) return;

	afd = parse_fopen(afn,"r");
	ihead = parse_omi(afd,0);
	n = ihead->n;

	while(++argv,--argc){
		if(num < 1 || num > n){
			fprintf(stderr,"armi: extract: no image: %d\n",num);
			exit(1);
		}
		i = num - 1;
		ofn = parse_suffix_malloc(*argv,"mi");
		label = ihead->label[i];
		nx = ihead->ncol[i];
		ny = ihead->nrow[i];
		blk = ihead->nblk[i];
		a = (unsigned short *) parse_malloc_short(blk*256);
		parse_rblock(afd,ihead->bptr[i]+1,a,blk);
		parse_from_can_shortv((short *)a,nx*ny);
		parse_wmi_l(ofn,a,nx,ny,label);
		printf("extract: %s image: %d (%d,%d)",ofn,num,nx,ny);
		if(*label) (void) printf(" \"%s\"",label);
		putchar('\n');
		num++;
		parse_free(a);
	}
}

/* editor tag: peep */
static void
delete(afn,argc,argv)
char *afn;
int argc;
char **argv;
{
	int m,op=1,haveto=0,j,num,i,n,nx,ny,blk;
	unsigned short new[256];
	short image[256];
	int *del;
	FILE *afd,*ofd;
	char *ofn,*label;
	PARSE_MIHEAD *ihead;

	afd = parse_fopen(afn,"r");
	ihead = parse_omi(afd,0);
	n = ihead->n;
	if(!n)parse_er("armi: empty input file");
	del = parse_malloc_int(n);

	for(i=0;i<n;i++) del[i] = 0;	/* 1 = delete image */
	printf("delete: input images: %d\n",n);

	while(++argv,--argc){
		num = parse_atof(*argv);
		if(num < 1 || num > n){
			(void )fprintf(stderr,
			  "armi: delete: no image: %d\n",num);
			exit(1);
		}
		del[num-1] = 1;
		haveto = 1;
	}
	if(!haveto) parse_er("armi: no images specified to delete");

	for(m=i=0;i<n;i++)m += del[i] == 0;

	if(m == 0){
		fputs("armi: delete: no images left; archive deleted\n",
		  stderr);
		parse_fclose(afd);
		if(-1 == unlink(afn))
			parse_err("armi: delete: can't unlink: ",afn);
		return;
	}

	ofn = TMPFILE;
	ofd = parse_fopen(ofn,"w");

	for(i=0;i<n;i++){
		label = ihead->label[i];
		nx = ihead->ncol[i];
		ny = ihead->nrow[i];
		blk = ihead->nblk[i];

		if(del[i] == 0){
			parse_rblock(afd,ihead->bptr[i],image,1);
			parse_wblock(ofd,op++,(char *)image,1);
			for(j=0;j<blk;j++){
				parse_rblock(afd,ihead->bptr[i]+j+1,image,1);
				parse_wblock(ofd,op++,image,1);
			}
		}else{
			printf("delete: deleted image: %d (%d,%d)",
			  i+1,nx,ny);
			if(*label) (void) printf(" \"%s\"",label);
			putchar('\n');
		}
	}
	printf("delete: output images: %d\n",m);
	parse_zbuf((short *) new,256);
	new[PARSE_MI_LOC_MAGIC] = PARSE_MI_MAGIC_65K;
	new[PARSE_MI_LOC_N] = m;
	new[PARSE_MI_LOC_FIRST] = 1;
	parse_to_can_shortv((short *)new,256);
	parse_wblock(ofd,0,new,1);
	parse_fclose(ofd);
	parse_fclose(afd);

	if(-1 == unlink(afn)){
		fprintf(stderr,"armi: delete: can't unlink: %s\n",afn);
		unlink(ofn);
		exit(1);
	}

	if(-1 == (rename(ofn,afn))){
		(void ) fprintf(stderr,"armi: delete: can't rename: %s\n",afn);
		exit(1);
	}
}

int
main(argc,argv)
int argc;
char **argv;
{
	char *afn;

	if(++argv,--argc < 2) syntax();

	while(**argv){
		switch(**argv){

		case '-':
			break;
		case 'c':
			mode = CREATE;
			break;
		case 'r':
			mode = WRITE;
			break;
		case 't':
			mode = LIST;
			break;
		case 'd':
			mode = DELETE;
			break;
		case 'x':
			mode = EXTRACT;
			break;
		case 'n':
			numspec = 1;
			break;
		case 'v':
			verbose = 1;
			break;
		default:
			fprintf(stderr,"illegal option: %c\n",**argv);
			exit(1);
		}
		(*argv)++;
	}

	if(mode == NONE){
		parse_err("armi: one of [cdrtx] must be specified","");
	}

	argc--; argv++;
	afn = parse_suffix_malloc(*argv,"mi");

	parse_wtext("****************** armi *********************");
	parse_pdt("---> start time");
	parse_rcsid_print(rcsid);

	if(mode == LIST){
		list(afn);
	}else if(mode == CREATE){
		add(1,afn,argc,argv);
	}else if(mode == WRITE){
		add(0,afn,argc,argv);
	}else if(mode == DELETE){
		delete(afn,argc,argv);
	}else if(mode == EXTRACT){
		extract(afn,argc,argv);
	}

	parse_pdt("---> stop time");
	exit(0);
}
