/*
cda - combine data files through weighted combinations
*/

static char rcsid[] = "$Id: cda.c,v 1.8 2004/05/16 04:06:18 ccrawford Exp $";

#include <parse.h>

static char *ofn="y.da";	/*output filename*/
static char *ifn1="y1.da";	/*input filename*/
static char *ifn2="y2.da";	/*input filename*/
static int prompt;		/*1=prompt for parameters*/
static int print=1;		/*flag for printing program info*/
static float w1=1.0;		/*weight*/
static float w2=1.0;		/*weight*/
static int cmode=1;		/*0=add 1=subtract 2=* 3=/ */
static int ncf=0;		/*# files on command line w/o =*/

static void eq_proc(char *option,char *arg);

struct PARSE_OPTION_TABLE parse_table[] = {
	{"i1",PARSE_CHAR,(PARSE_ARG)&ifn1,"first input file"},
	{"i2",PARSE_CHAR,(PARSE_ARG)&ifn2,"second input file"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,"output file"},
	{"w1",PARSE_FLOAT,(PARSE_ARG)&w1,"weight for first vector"},
	{"w2",PARSE_FLOAT,(PARSE_ARG)&w2,"weight for second vector"},
	{"=",PARSE_PROC,(PARSE_ARG)eq_proc,"set files"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0,0}
};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'p',&print,PARSE_FLAG_CLR,"don't print program information"},
	{'P',&prompt,PARSE_FLAG_SET,"prompt for parameters"},
	{'a',&cmode,PARSE_FLAG_VALUE(0),"add vectors"},
	{'s',&cmode,PARSE_FLAG_VALUE(1),"subtract vectors (default)"},
	{'m',&cmode,PARSE_FLAG_VALUE(2),"multiply vectors"},
	{'d',&cmode,PARSE_FLAG_VALUE(3),"divide vectors"},
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

int
main(argc,argv)
	int argc;
	char **argv;
{
	float *ivr1,*ivr2,*ov;
	short int head[256];
	FILE *ifd1,*ifd2,*ofd;
	int np,mp,typep,n,m,j,i,iblki,nblki,type,iblko,nblko;

/*
	parse command line and check for errors
*/
	parse_basic(argc,argv);
	if(ncf == 1)parse_err("cda: only one file set","");
/*
	prompt for parameters
*/
	if(prompt){
		ifn1 = parse_acs("first input file",ifn1,"da",(char *)NULL);
		ifn2 = parse_acs("second input file",ifn2,"da",(char *)NULL);
		ofn = parse_acs("output file",ofn,"da",(char *)NULL);
		parse_acm("mode: + - * /",cmode,"+-*/",&cmode);
		parse_acf("weight for #1",w1,&w1);
		parse_acf("weight for #2",w2,&w2);
	}
/*
	open files
*/
        ifn1 = parse_suffix_malloc(ifn1,"da");
        ifn2 = parse_suffix_malloc(ifn2,"da");
        ofn = parse_suffix_malloc(ofn,"da");

        ifd1 = parse_fopen(ifn1,"r");
        ifd2 = parse_fopen(ifn2,"r");
        ofd = parse_fopen(ofn,"w");
/*
	header information
*/
        parse_rblock(ifd1,0,(char *)head,1);
        parse_from_can_shortv(head,256);
        n = head[0];
        m = head[1];
        type = head[2];
        if(type)parse_err("cda: can't handle integer da files:",ifn1);

        parse_rblock(ifd2,0,(char *)head,1);
        parse_from_can_shortv(head,256);
        np = head[0];
        mp = head[1];
        typep = head[2];
        if(typep)parse_err("cda: can't handle integer da files:",ifn2);

	if(np != n || mp != m || typep != type)
		parse_err("cda: input files don't match","");
/*
	get internal vectors
*/
	ivr1 = parse_malloc_float(BLOCK_FLOAT(n)*128);
	ivr2 = parse_malloc_float(BLOCK_FLOAT(n)*128);
	ov = parse_malloc_float(BLOCK_FLOAT(n)*128);
/*
	print program status
*/
	if(print){
		parse_wtext("****************** cda *********************");
		parse_pdt("---> start time");
		parse_rcsid_print(rcsid);
		parse_prs("first input file:",ifn1);
		parse_prs("second input file:",ifn2);
		parse_prs("output file:",ofn);
		parse_pri("record size =",n);
		parse_pri("records =",m);
		parse_prf("w1 =",w1);
		parse_prf("w2 =",w2);
		if(cmode == 0)parse_wtext("mode: add");
		if(cmode == 1)parse_wtext("mode: subtract");
		if(cmode == 2)parse_wtext("mode: multiply");
		if(cmode == 3)parse_wtext("mode: divide");
	}
/*
	compute record information
*/
	nblki = BLOCK_FLOAT(n);
	nblko = BLOCK_FLOAT(n);
        iblko = iblki = 1;
/*
	write header to the output file
*/
        parse_to_can_shortv(head,256);
        parse_wblock(ofd,0,(char *)head,1);
/*
	read input records and send them to the output file
*/
	for(i=0;i<m;i++){
		parse_rblock(ifd1,iblki,(char *)ivr1,nblki);
		parse_rblock(ifd2,iblki,(char *)ivr2,nblki);
		parse_from_can_floatv(ivr1,n);
		parse_from_can_floatv(ivr2,n);
		iblki += nblki;

		if(cmode == 0)
			for(j=0;j<n;j++)ov[j]=w1*ivr1[j] + w2*ivr2[j];
		else if(cmode == 1)
			for(j=0;j<n;j++)ov[j]=w1*ivr1[j] - w2*ivr2[j];
		else if(cmode == 2)
			for(j=0;j<n;j++)ov[j]=w1*ivr1[j] * w2*ivr2[j];
		else
			for(j=0;j<n;j++)ov[j]=w1*ivr1[j] / (w2*ivr2[j]);

		parse_to_can_floatv(ov,n);
		parse_wblock(ofd,iblko,(char *)ov,nblko);
		iblko += nblko;
	}
/*
	clean up and exit
*/
        if(print)parse_pdt("---> stop time");
	exit(0);
}

static void eq_proc(option,arg)
char *option,*arg;
{
	switch(ncf++){

	case 0:
		ifn1 = arg;
		break;
	case 1:
		ifn2 = arg;
		break;
	default:
		parse_err("cda: too many files","");
	}
}
