/*
dami - convert da files to mi files
*/

static char rcsid[] = "$Id: dami.c,v 1.6 2011/09/14 14:14:49 carl Exp $";

#include <parse.h>
#include <math.h>
#include <string.h>

static int print=1;		/* 1=print program status */
static char *ifn="y.da";	/* input file name*/
static char *ofn;		/* output file name*/
static float zero=512;		/* number where zero will land */
static float scale=1023;	/* scale (range) of output */
static float cl_min,cl_max;	/* command line min/max */

struct PARSE_OPTION_TABLE parse_table[] = {
	{"if",PARSE_CHAR,(PARSE_ARG)&ifn,"name of da-file"},
	{"=",PARSE_CHAR,(PARSE_ARG)&ifn,"if"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,
		"output filename (default built from input filename)"},
	{"zero",PARSE_FLOAT,(PARSE_ARG)&zero,"zero output location"},
	{"scale",PARSE_FLOAT,(PARSE_ARG)&scale,"scale (range) of image"},
	{"min",PARSE_FLOAT,(PARSE_ARG)&cl_min,"data minimum for scaling"},
	{"max",PARSE_FLOAT,(PARSE_ARG)&cl_max,"data maximum for scaling"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0}
};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'p',&print,PARSE_FLAG_CLR,"don't print program information"},
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

int
main(argc,argv)
	int argc;
	char **argv;
{
	int i;
	int nxy,nx,ny;
	short int *mip;
	float *dap;
	float dmin,dmax;
/*
	parse command line 
*/
	parse_all_basic(".dami","DAMI",argc,argv);
/*
	fix up parameters
*/
	ifn = parse_suffix_malloc(ifn,"da");
	if(!parse_opexist("of")){
		ofn = parse_malloc_char(strlen(ifn)+1);
		strcpy(ofn,ifn);
		parse_remove_suffix(ofn);
	}
	ofn = parse_suffix_malloc(ofn,"mi");
/*
	open and da file 
*/
	dap = parse_rda(NULL,&nx,&ny,ifn);
	nxy = nx * ny;
/*
	print program information
*/
	if(print){
		parse_wtext("***************** dami *****************");
		parse_pdt("---start time");
		parse_rcsid_print(rcsid);
		parse_prs("input file:",ifn);
		parse_prs("output file:",ofn);
		parse_prf("scale =",scale);
		parse_prf("zero =",zero);
		parse_pri("nx =",nx);
		parse_pri("ny =",ny);
		if(parse_opexist("min"))parse_prf("min =",cl_min);
		if(parse_opexist("max"))parse_prf("max =",cl_max);
	}
/*
	get space
*/
	mip = (short int *) parse_malloc(nxy * sizeof(short int));
/*
	find min and max 
*/
	dmin = dmax = dap[0];
	for(i=0;i<nxy;i++){
		if(dap[i] > dmax)dmax = dap[i];
		if(dap[i] < dmin)dmin = dap[i];
	}

	if(parse_opexist("min")) dmin = cl_min;
	if(parse_opexist("max")) dmax = cl_max;


	if(dmax == dmin){
		parse_prf("dami (warning): min = max = ",dmax);
		scale = 1;
	}else{
		dmax = fabs(dmax);
		dmin = fabs(dmin);
		dmax = (dmax > dmin)? dmax : dmin;
		dmin = -dmax;
		zero = zero - scale/2;
		scale = scale / (dmax - dmin);
	}
/*
	scale data
*/
	for(i=0;i<nxy;i++){
		mip[i] = scale * (dap[i] - dmin) + zero;
		if(mip[i] < 0) mip[i] = 0;
	}
/*
	output mi file
*/
	parse_wmi(ofn,(unsigned short *) mip,nx,ny);
/*
	clean and exit
*/
	if(print)parse_pdt("---stop time");
	exit(0);
}
