/*
exda - extract a record from a data file
*/

static char rcsid[] = "$Id: exda.c,v 1.8 2005/03/29 21:01:11 ccrawford Exp $";

#include <stdio.h>
#include "parse.h"

static int n;			/*data size - from header*/
static int m;			/*y size of data - from header*/
static int print=1;		/*flag for printing program info*/
static char *ofn="y.da";	/*output filename*/
static char *ifn="y1.da";	/*input filename*/
static int r;			/*record number, origin=0 */
static int c=1;            	/*record count*/
static int cs;			/*1= c set */
static int rs;           	/*start of record*/
static int rss;			/*1=rs set*/
static int re;			/*end of record*/
static int res;			/*1=re set*/
static int col;			/*1 = col mode */
static int prompt;		/*1=prompt for parameters*/

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'P',&prompt,PARSE_FLAG_SET,"prompt for parameters"}, 
	{'p',&print,PARSE_FLAG_CLR,"don't print parameters"}, 
	{'c',&col,PARSE_FLAG_SET,"get column of data"}, 
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

struct PARSE_OPTION_TABLE parse_table[] = {
	{"if",PARSE_CHAR,(PARSE_ARG)&ifn,"input file"},
	{"=",PARSE_CHAR,(PARSE_ARG)&ifn,"if"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,"output file"},
	{"start",PARSE_INT,(PARSE_ARG)&rs,"start within record"},
	{"column",PARSE_INT,(PARSE_ARG)&rs,"column; sets -c"},
	{"column",PARSE_SET,(PARSE_ARG)&col},
	{"end",PARSE_INT,(PARSE_ARG)&re,"end within record"},
	{"count",PARSE_INT,(PARSE_ARG)&c,"number records (default all)"},
	{"n",PARSE_INT,(PARSE_ARG)&c,"synonym for count"},
	{"first",PARSE_INT,(PARSE_ARG)&r,"first record"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0}
};

int
main(argc,argv)
	int argc;
	char **argv;
{
	int *ivr;
	float *ovr=NULL;
	short int head[256];
	FILE *ifd,*ofd;
	int no,i,iblki,iblko,nblki,nblko,type;

/*
	parse command line and set modes based on command line
*/
	parse_basic(argc,argv);

	rss = parse_opexist("start");
	res = parse_opexist("end");
	cs = parse_opexist("count") | parse_opexist("n");
/*
	prompt for parameters
*/
	if(prompt){
		ifn = parse_acs("input file",ifn,"da",NULL);
		ofn = parse_acs("output file",ofn,"da",NULL);
		parse_act("col mode",col,&col);
		parse_aci("first record #",r,&r);
		if(col){
			parse_act("modify default record count",cs,&cs);
			if(cs)parse_aci("record count",c,&c);
			parse_aci("col number",rs,&rs);
		}else{
			parse_aci("# records",c,&c);
			parse_act("modify start of record",rss,&rss);
			if(rss)parse_aci("record start",rs,&rs);
			parse_act("modify end of record",res,&res);
			if(res)parse_aci("record end",re,&re);
		}
	}
/*
	open files
*/
        ifn = parse_suffix_malloc(ifn,"da");
        ofn = parse_suffix_malloc(ofn,"da");

        ifd = parse_fopen(ifn,"r");
        ofd = parse_fopen(ofn,"w");
/*
	header information
*/
        parse_rblock(ifd,0,(char *)head,1);
        parse_from_can_shortv(head,256);
        n = head[0];
        m = head[1];
        type = head[2];
        if(type)parse_err("exda: can't handle integer da files:",ifn);
/*
	fix up record pointers
*/
	if(col){
		re = rs;
		if(!cs){
			c = m - r;
		}
	}else{
		if(!rss)rs = 0;
		if(!res)re = n-1;
	}
/*
	check parameter validity
*/
        if(rs > re)parse_er("record start > end");
	if(col){
		parse_pcheck(c,1,m,"too man rows requested");
	}else{
		parse_pcheck(r,0,m-1,"bad record");
		parse_pcheck(c,1,m-r,"bad record count");
		parse_pcheck(rs,0,n-1,"bad record start");
		parse_pcheck(re,0,n-1,"bad record end");
	}
/*
	get internal vectors
*/
	ivr = parse_malloc_int(BLOCK_FLOAT(n)*128*2);
	if(col)ovr=parse_malloc_float(BLOCK_FLOAT(c)*128);
/*
	print program status
*/
	if(print){
		parse_wtext("****************** exda *********************");
		parse_pdt("---> start time");
		parse_rcsid_print(rcsid);
		parse_pri("da-file record size:",n);
		parse_pri("da-file records:",m);
		parse_prs("input file:",ifn);
		parse_prs("output file:",ofn);
		parse_pri("first record:",r);
		parse_pri("records:",c);
		if(col){
			parse_wtext("col mode");
			parse_pri("col:",rs);
		}else{
			if(rss)parse_pri("record start:",rs);
			if(res)parse_pri("record end:",re);
		}
	}
/*
	compute record information
*/
	nblki = BLOCK_FLOAT(n);
        iblki = 1 + r*nblki;

	no = col? c : re - rs + 1;
	nblko = BLOCK_FLOAT(no);
        iblko = 1; 
/*
	write header to the output file
*/
        head[0] = no;
	head[1] = col? 1 : c;
        parse_to_can_shortv(head,256);
        parse_wblock(ofd,0,(char *)head,1);
/*
	read input records and send them to the output file
*/
	for(i=0;i<c;i++){
		parse_rblock(ifd,iblki,(char *)ivr,nblki);
		iblki += nblki;
		if(!col){
			parse_wblock(ofd,iblko,(char *)(ivr+rs),nblko);
			iblko += nblko;
		}else{
			ovr[i] = parse_from_can_float(ivr[rs]);
		}
	}
/*
	write col
*/
	if(col){
		parse_to_can_floatv(ovr,c);
		parse_wblock(ofd,iblko,(char *)ovr,nblko);
	}
/*
	clean up and exit
*/
        if(print)parse_pdt("---> stop time");
	exit(0);
}
