/*
minda - minify da files
*/

static char rcsid[] = "$Id: minda.c,v 1.6 2004/05/16 04:06:18 ccrawford Exp $";

#include <stdlib.h>
#include <math.h>
#include <parse.h>

static char *ofn="min.da";		/*output filename*/
static char *ifn="y.da";		/*input filename*/
static int xm=2;			/*minification in cols*/
static int ym=2;			/*minification in rows*/
static int transpose=0;			/*1 = transpose output*/
static int print=1;			/*1 = print program info*/

struct PARSE_OPTION_TABLE parse_table[] = {
	{"if",PARSE_CHAR,(PARSE_ARG)&ifn,"input file"},
	{"=",PARSE_CHAR,(PARSE_ARG)&ifn,"if"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,"output file"},
	{"ym",PARSE_INT,(PARSE_ARG)&ym,"row minification factor"},
	{"xm",PARSE_INT,(PARSE_ARG)&xm,"column minification factor"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0,0}
};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'p',&print,PARSE_FLAG_CLR,"don't print program information"},
	{'t',&transpose,PARSE_FLAG_SET,"transpose output "},
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

int main(int argc,char **argv)
{ 
	int v,s;		/* view/column loopers */
	int ind;		/* 2d pointer */
	int nblk,iblk;		/* block count and pointer */
	int iv,is;		/* loopers for boxcar integration */
	float fsum;		/* temp sum during boxcar integration */
	float mfact;		/* scale factor during boxcar integration */
	FILE *ou,*iu;		/* for reading/writing da-files */
	int mvout;		/* views out */
	int mdout;		/* detectors (columns) out */
	int insiz;		/* size of cor data */
	int ousiz;		/* size of output cor data */
	int mview;		/* input number of views */
	int ndet;		/* input number of detectors */
	float *indat;		/* input data */
	float *oudat;		/* output data */
	float *temp;		/* temp output data */
	short int head[256];	/* header for da-files */
/*
	parse command line and check for errors
*/
	parse_all_basic(".minda","MINDA",argc,argv);

	if(xm <= 0)parse_er("minda: xm must positive");
	if(ym <= 0)parse_er("minda: ym must be positive");
/*
	open files
*/
	ifn = parse_suffix_malloc(ifn,"da");
	ofn = parse_suffix_malloc(ofn,"da");
	iu = parse_fopen(ifn,"r");
	ou = parse_fopen(ofn,"w");
/*
	read input da header information
*/
        parse_rblock(iu,0,(char *)head,1);
        parse_from_can_shortv(head,256);

	ndet = head[0];
	mview = head[1];

	insiz = ndet * mview;
	mvout = mview / ym;
	mdout = ndet / xm;
	ousiz = mvout * mdout;

	if(mvout <= 0)parse_er("minda: output rows must be positive");
	if(mdout <= 0)parse_er("minda: output columns must be positive");
/*
	allocate space
*/	
	indat = parse_malloc_float(insiz);	
	oudat = parse_malloc_float(ousiz);	
	if(transpose){
		temp = parse_malloc_float(mvout);	
	}else{ 
		temp = parse_malloc_float(mdout);	
	}
/*
	print program status
*/
	if(print){
		parse_wtext("****************** minda *********************");
		parse_pdt("---> start time");
		parse_rcsid_print(rcsid);
		parse_prs("input file:",ifn);
		parse_prs("output file:",ofn);
		parse_pri("rows in =",mview);
		parse_pri("columns in =",ndet);
		parse_pri("rows out =",mvout);
		parse_pri("columns out=",mdout);
		parse_pri("xm =",xm);
		parse_pri("ym =",ym);
		parse_prt("transpose",transpose);
	}
/*
	read data
*/
	nblk = BLOCK_FLOAT(ndet);
	iblk = 1;
	for(v=0;v<mview;v++){
		parse_fread_block((char *)(indat+v*ndet),sizeof(float),
			ndet,iu,iblk);

		iblk += nblk;
	}
	parse_from_can_floatv(indat,insiz);
/*
	minify data
*/
	mfact = 1.0 / (xm * ym);
	for(v=0;v<mvout;v++){
		for(s=0;s<mdout;s++){
			ind = v*ndet*ym + s*xm;
			fsum = 0;
			for(iv=0;iv<ym;iv++){
			  for(is=0;is<xm;is++){
				fsum += indat[ind+ndet*iv+is];
			  }
			}
			oudat[v*mdout+s] = fsum * mfact;
		}
	}
/*
	write output files
*/
	if(!transpose){
		head[0] = mdout;	
		head[1] = mvout;	
		parse_to_can_shortv(head,256);
		parse_wblock(ou,0,(char *)head,1);

		nblk = BLOCK_FLOAT(mdout);
		iblk = 1;
	
		for(v=0;v<mvout;v++){
			for(s=0;s<mdout;s++)temp[s] = oudat[v*mdout+s];
			parse_to_can_floatv(temp,mdout);
			parse_wblock(ou,iblk,(char *)temp,nblk); 
			iblk += nblk;
		}
	}else{
		head[0] = mvout;	
		head[1] = mdout;	
		parse_to_can_shortv(head,256);
		parse_wblock(ou,0,(char *)head,1);

		nblk = BLOCK_FLOAT(mvout);
		iblk = 1;
	
		for(s=0;s<mdout;s++){
			for(v=0;v<mvout;v++)temp[v] = oudat[v*mdout+s];
			parse_to_can_floatv(temp,mvout);
			parse_wblock(ou,iblk,(char *)temp,nblk); 
			iblk += nblk;
		}
	}
/*
	clean up and exit
*/
	if(print)parse_pdt("---> finish time");
	exit(0);
}
