/*
sagcor - generate coronal and sagittal slices
*/

static char rcsid[] = "$Id: sagcor.c,v 1.11 2011/09/14 14:09:03 carl Exp $";

#include <parse.h>
#include <string.h>

static int print=1;		/* 1=print program outtatus */
static char *ifn="pic.mi";	/* input file name*/
static char *ofn="sc.mi";	/* output file name*/
static int row;			/* row in coronal mode */
static int col;			/* column in sagittal mode */
static int n=1;			/* number of output images */
static int fimage;		/* first image */
static int limage;		/* last image */
static int allslice=0;		/* 1=generate all slices */
static int cs=0;		/* 0=coronal 1=sagittal */
static int syn=0;		/* 1=synthetic view */
static int offset=0;		/* offset to input images */

struct PARSE_OPTION_TABLE parse_table[] = {
	{"if",PARSE_CHAR,(PARSE_ARG)&ifn,"input file"},
	{"=",PARSE_CHAR,(PARSE_ARG)&ifn,"if"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,"output filename"},
	{"row",PARSE_INT,(PARSE_ARG)&row,"row for coronal"},
	{"col",PARSE_INT,(PARSE_ARG)&col,"column for sagittal"},
	{"fimage",PARSE_INT,(PARSE_ARG)&fimage,"first image"},
	{"limage",PARSE_INT,(PARSE_ARG)&limage,"last image"},
	{"n",PARSE_INT,(PARSE_ARG)&n,"output slices"},
	{"offset",PARSE_INT,(PARSE_ARG)&offset,"image offset"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0}
};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'p',&print,PARSE_FLAG_CLR,"don't print program information"},
	{'c',&cs,PARSE_FLAG_CLR,"coronal slices"},
	{'s',&cs,PARSE_FLAG_SET,"sagittal slices"},
	{'a',&allslice,PARSE_FLAG_SET,"generate all slices"},
	{'v',&syn,PARSE_FLAG_SET,"synthetic view"},
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

int
main(int argc,char **argv)
{
	int r,c;		/* row/columns in images */
	int i,j,k;		/* loopers and pointers */
	short head[256];	/* mi-file header */
	FILE *fd;		/* for input */
	int imin;		/* number of input images */
	int imuse;		/* number of used input images */
	int or,oc;		/* output rows and columns */
	unsigned short *a,*b;	/* intput and output images */
	PARSE_MIHEAD *mihead;	/* for parse_omi */
	int nblk;		/* blocks per output image */
	int oblk;		/* output block pointer */
	int np;			/* pixels to pad to block boundary */
	int npic;		/* total pixels per output image */
	int pixel;		/* pixel in output images */
	int *proj=NULL;		// for generating projection image
/*
	parse command line 
*/
	parse_all_basic(".sagcor","SAGCOR",argc,argv);
/*
	fix parameters 
*/
	ifn = parse_suffix_malloc(ifn,"mi");
	ofn = parse_suffix_malloc(ofn,"mi");
/*
	read input file header 
*/
	ifn = parse_suffix_malloc(ifn,"mi");
	fd = parse_fopen(ifn,"r");
	mihead = parse_omi(fd,0);

	imin = mihead->n;
	if(!imin)parse_er("sagcor: mi-file empty");

	if(!parse_opexist("fimage")) fimage = 1;
	if(!parse_opexist("limage")) limage = imin;
	parse_pcheck(fimage,1,imin,"sagcor: bad first image");
	parse_pcheck(limage,1,imin,"sagcor: bad last image");
	if(fimage > limage) parse_er("sagcor: first > last image");
	imuse = limage - fimage + 1;

	c = mihead->ncol[fimage-1];
	r = mihead->nrow[fimage-1];
	for(i=1;i<imuse;i++){
		if(mihead->ncol[i+fimage-1] != c){
			parse_er("sagcor: images not same size");
		}
		if(mihead->nrow[i+fimage-1] != r){
			parse_er("sagcor: images not same size");
		}
	}
/*
	set up for sagittal and coronal
*/
	if(!parse_opexist("row") && !parse_opexist("col")){
		if(cs == 0){
			row = r/2 + 1;
		}else{
			col = c/2 + 1;
		}
	}else if(parse_opexist("row") && !parse_opexist("col")){
		cs = 0;
	}else if(!parse_opexist("row") && parse_opexist("col")){
		cs = 1;
	}else if(parse_opexist("row") && parse_opexist("col")){
		parse_er("sagcor: can't set row and column");
	}else{
		parse_er("sagcor: internal error 1");
	}
/*
	setup number of range of output slices
*/
	if(allslice){
		if(parse_opexist("row") || parse_opexist("col")){
			parse_er("sagcor: can't use -a and row or col");
		}
		if(cs == 0){
			row = 1;
			n = r;
		}else{
			col = 1;
			n = c;
		}
	}
/*
	setup input and output buffers
*/
	if(n <= 0) parse_er("sagcor: bad n");
	if(cs == 0){
		parse_pcheck(row,1,r,"sagcor: invalid row");
		if(row+n-1>r) parse_er("sagcor: bad n");
		or = c;
		oc = imuse;
	}else{
		parse_pcheck(col,1,c,"sagcor: invalid column");
		if(col+n-1>c) parse_er("sagcor: bad n");
		or = r;
		oc = imuse;
	}
	npic = oc * or;
	a = (unsigned short *) parse_malloc_short(r * c);
	if(syn == 0){
		b = (unsigned short *) parse_malloc_short(oc*or*n);
	}else{
		b = (unsigned short *) parse_malloc_short(oc*or);
		proj = parse_malloc_int(oc*or);
		parse_zibuf(proj,oc*or);
	}

/*
	print program info
*/
	if(print){
		parse_wtext("****************** sagcor *********************");
		parse_pdt("---> start time");
		parse_rcsid_print(rcsid);
		parse_prs("input file:",ifn);
		parse_prs("output file:",ofn);
		parse_pri("input columns:",c);
		parse_pri("input rows:",r);
		parse_pri("input images:",imin);
		parse_pri("used input images:",imuse);
		parse_pri("output columns:",oc);
		parse_pri("output rows:",or);
		parse_pri("first image:",fimage);
		parse_pri("last image:",limage);
		parse_prt("synthetic view",syn);
		parse_pri("offset =",offset);
		if(cs == 0){
			parse_wtext("mode: coronal");
			parse_pri("first row:",row);
		}else{
			parse_wtext("mode: sagittal");
			parse_pri("first col:",col);
		}
		if(syn == 0){
			parse_pri("output images:",n);
		}else{
			if(cs == 0){
				parse_pri("input rows:",n);
			}else{
				parse_pri("input columns:",n);
			}
			parse_pri("output images:",1);
		}
	}
/*
	do it
*/
	for(i=0;i<imuse;i++){ 	/* loop input images */
		parse_fread_block(a,sizeof(short),r*c,fd,
		  mihead->bptr[i+fimage-1]+1);
		parse_from_can_shortv((short *)a,r*c);

		for(k=0;k<n;k++){	/* loop output images */
			for(j=0;j<or;j++){	/* loop pixels in row/col */
				if(cs == 0){
					pixel = a[(row-1+k)*c+j];
				}else{
					pixel = a[c*j+col-1+k];
				} 
				pixel -= offset;
				if(pixel < 0) pixel = 0;
				if(syn == 0){
					b[k*npic+j*oc+i] = pixel;
				}else{
					proj[j*oc+i] += pixel;
				}
			}
		}
	}
	parse_fclose(fd);
/*
	synthetic view normalization
*/
	if(syn){
		for(j=0;j<or;j++){
			for(k=0;k<oc;k++){
				b[j*oc+k] = proj[j*oc+k] / n;
			}
		}
	}
/*
	write output image
*/
	if(syn == 0){
		ofn = parse_suffix_malloc(ofn,"mi");
		fd = parse_fopen(ofn,"w");

		nblk = PARSE_BLOCK_SHORT(npic);

		parse_zbuf((short *)head,256);

		head[PARSE_MI_LOC_MAGIC] = PARSE_MI_MAGIC_65K;
		head[PARSE_MI_LOC_N] = n;
		head[PARSE_MI_LOC_FIRST] = 1;

		parse_to_can_shortv((short *)head,256);
		parse_wblock(fd,0,(char *)head,1);
		oblk = 1;

		parse_zbuf(head,256);
		head[PARSE_MI_LOC_NCOL] = parse_to_can_short(oc);
		head[PARSE_MI_LOC_NROW] = parse_to_can_short(or);
		strncpy((char *) &head[PARSE_MI_LOC_LABEL],ifn,PARSE_MI_LEN_LABEL);
		parse_to_can_shortv((short *)b,npic*n);

		for(k=0;k<n;k++){	/* loop output images */
			parse_wblock(fd,oblk++,(char *)head,1);
			np = nblk * 512 / sizeof(short);
			parse_fwrite(b+k*npic,sizeof(short),npic,fd);
			parse_fwrite(head,sizeof(short),np-npic,fd);
			oblk += nblk;
		}
		parse_fclose(fd);
	}else{
		parse_wmi_l(ofn,b,oc,or,ifn);
	}
/*
	clean up and exit
*/
	if(print)parse_pdt("---> stop time");
	exit(0);
}
