/*
annotate - handle screen annotation
*/

#ident "$Id: annotate.c,v 1.8 2007/05/22 13:22:45 ccrawford Exp $"

#include "xpic.h"

static struct ANNO{
	int x;
	int y;
	char *s;
	struct ANNO *next;
} *aroot=NULL,*ap,*app;

static void
annotate_syntax(void)
{
	fputs("Synopsis: annotate display\n",ttyout);
	min_abbreviation();
	fputs("Related: clear, source\n",ttyout);
	fputs("Syntax:\n",ttyout);

	fputs("annotate text...         * display text when cursor deposited\n",ttyout);
	fputs("annotate -p x y text...  * display text at (x,y)\n",ttyout);
	fputs("annotate -c              * clear annotations\n",ttyout);
	fputs("annotate -w [!]file      * write annotation list to file\n",ttyout);
	fputs("                         * overwrite file with !\n",ttyout);
}

void
cmd_annotate(void)
{
	char **argp = argv;
	int x=0,y=0;

	if(argc <= 1){
		annotate_syntax();
		return;
	}
	if(argv[1][0] == '-' && argv[1][1] == 'w'){
		annotate_write();
		return;
	}
	if(argv[1][0] == '-' && argv[1][1] == 'c'){
		annotate_clear();
		load_pix(0);
		return;
	}
	if(argv[1][0] == '-' && argv[1][1] == 'p'){
		if(argc < 5){
			annotate_syntax();
			return;
		}
		x = parse_atof(argv[2]);
		if(parse_atof_error()) return;

		y = parse_atof(argv[3]);
		if(parse_atof_error()) return;

		note = 0;
		argc -= 3;
		argp += 3;
	}else{
		note = 1;
	}

	note_text[0] = 0;
	while(--argc){
		if(*note_text)strcat(note_text," ");
		strcat(note_text,*++argp);
	}
	if(!note){
		annotate_add(x,y,note_text);
	}
}

void
annotate_clear(void)
{
	app = aroot;
	while(app != NULL){
		ap = app;
		app = ap->next;
		free(ap);
	}
	aroot = NULL;

	box_table_clear();
	line_table_clear();
}

void
annotate_refresh(void)
{
	app = aroot;
	while(app != NULL){
		display_text(app->x,app->y,app->s);
		app = app->next;
	}

	box_refresh();
	line_refresh();
	if(box.vis && boxred == 0) draw_box(LINE_SOFT);
}

int
annotate_exist(void)
{
	return( (aroot != NULL) || box_exist() || line_exist() || box.vis);
}

void
annotate_add(int x,int y,char *s)
{
	if(NULL == (ap = (struct ANNO *)malloc(sizeof(struct ANNO)+strlen(s)+1))){
		fputs("Can't allocate space for annotation\n",ttyout);
		return;
	}

	ap->x = x;
	ap->y = y;
	ap->s = ((char *)ap) + sizeof(struct ANNO);
	strcpy(ap->s,s);
	ap->next = NULL;

	if(aroot == NULL){
		aroot = ap;
	}else{
		app = aroot;
		while(app->next != NULL) app = app->next;
		app->next = ap;
	}
	display_text(x,y,s);
}

void
annotate_write(void)
{
	int over;
	FILE *fd;
	char *ofn;
	struct stat sbuf;

	if(argc != 3){
		annotate_syntax();
		return;
	}

	ofn = argv[2];
	if(ofn[0] == '!'){
		over = 1;
		ofn++;
	}else{
		over = 0;
	}

	if(!stat(ofn,&sbuf) && !over){
		fputs("File exists. Use 'annotate !file' to overwrite\n",
			ttyout);
		return;
	}
	unlink(ofn);
	if((fd=fopen(ofn,"w")) == NULL){
		fprintf(ttyout,"Can't open: %s\n",ofn);
		return;
	}

	app = aroot;
	while(app != NULL){
		fprintf(fd,"annotate -p %d %d %s\n",app->x,app->y,app->s);
		app = app->next;
	}

	fclose(fd);
}
