/*
cmdlc - extract lines and columns from xpic (command line version of lc.c)
*/

#ident "$Id: cmdlc.c,v 1.7 2011/09/14 11:08:54 carl Exp $"

#include "xpic.h"
#include "crcplot.h"

static char *ofn = "xplot";

static int sxp=-1,syp=-1;	// (x,y) start of vector
static int xps,yps;	// x,y set on the command line
static int n=-1;	// number of points in vector
static int ns;		// n set on the command line
static int lc=0;	// 0=line (horizontal) 1=column (should be rc=row/column)
static int cur=0;	// 1=use cursor loc for x OR y; 2=use for x AND y

static void lc_syntax(void);

static struct PARSE_FLAG_TABLE flags[] = {
	{'l',&lc,PARSE_FLAG_CLR},
	{'c',&lc,PARSE_FLAG_SET},
	{'d',&cur,PARSE_FLAG_SET},
	{'D',&cur,PARSE_FLAG_VALUE(2)},
	{'?',(int *)lc_syntax,0,0},
	{0,0,0,0}
};

static struct PARSE_STRUCT options[] = {
	{"x",PARSE_INT,(PARSE_ARG)&sxp},
	{"x",PARSE_SET,(PARSE_ARG)&xps},
	{"y",PARSE_INT,(PARSE_ARG)&syp},
	{"y",PARSE_SET,(PARSE_ARG)&yps},
	{"n",PARSE_INT,(PARSE_ARG)&n},
	{"n",PARSE_SET,(PARSE_ARG)&ns},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn},
	{"?",PARSE_PROC,(PARSE_ARG)lc_syntax},
	{0,0,0,0}
};

static void 
werr(char *ofn)
{
	fprintf(ttyout,"Error writing to: %s\n",ofn);
}

static void
lc_syntax(void)
{
	fputs("Synopsis: plots lines and columns\n",ttyout);
	min_abbreviation();
	fputs("Related: write halftone\n",ttyout);
	fputs("Syntax: lc [options]\n",ttyout);
	fprintf(ttyout,"x=: x-coordinate of vector start\n");
	fprintf(ttyout,"y=: y-coordinate of vector start\n");
	fprintf(ttyout,"n=: number of vector points\n");
	fprintf(ttyout,"of=[%s]: output location xplot=send to xplot\n",ofn);
	fputs("-l: plot line (horizontal) (default)\n",ttyout);
	fputs("-c: plot column (vertical)\n",ttyout);
	fputs("-d: use cursor for x OR y\n",ttyout);
	fputs("-D: use cursor for x AND y\n",ttyout);
}

void
cmd_lc(void)
{
	int i;			// looper
	int sx,sy;		// size of canvas
	int type;		// type or cursor
	int cx=-1,cy=-1;	// cursor location
	float *fx=NULL,*fy=NULL;// output vectors
	int have_vectors=0;	// 1=vectors malloc'ed
	int have_file=0;	// 1=file open
	int nblk; 		// #blocks in output file
	FILE *ofd=NULL;		// for writing vector
	short head[256];	// header for da-file
	int outloc;		// 0=file 1=xplot
	int np;			// related to n

		// stuff for plotting
	float xxs=0.0;		/*x starting location of plot*/
	float yys=0.0;		/*y starting location of plot*/
	float xlen=4.0;		/*length of x axis*/
	float ylen=3.0;		/*length of y axis*/
	float scfac=1.0;	/*scale factor*/
	float xmin,xmax;	// min/max of x vector
	float ymin,ymax;	// min/max of y vector



	if(argc < 0){
		lc_syntax();
		return;
	}

	lc = cur = ns = xps = yps = 0;
	if(-1 == parse_general(options,flags,ttyout,argc,argv)) return;

	if(strcmp(ofn,"xplot")){
		ofn = parse_suffix_malloc(ofn,"da");
		outloc = 0;
	}else{
		outloc = 1;
	}

	// xpic_status(&sx,&sy,&w,&l,&m);
	sx = canvas_size_x;
	sy = canvas_size_y;

	if(cur){
		if(lc && xps){
			fputs("can't set column in cursor mode\n",ttyout);
			goto out;
		}
		if((!lc) && yps){
			fputs("can't set row in cursor mode\n",ttyout);
			goto out;
		}

		//xpic_cursor(&type,&cx,&cy,&cxw,&cyw);
		type = cursor_stat;
		cx = box.lx;
		cy = box.ly;

		if(!type){
			fputs("cursor not enabled\n",ttyout);
			goto out;
		}

		if(lc){
			sxp = cx;
			xps = 1;
			if(!yps){
				yps = 1;
				syp = (cur ==2)? cy : 0;
			}
		}else{
			syp = cy;
			yps = 1;
			if(!xps){
				xps = 1;
				sxp = (cur ==2)? cx : 0;
			}
		}
	}else{
		if(lc){
			if(!xps) sxp = sx/2;
			if(!yps) syp = 0;
		}else{
			if(!xps) sxp = 0;
			if(!yps) syp = sy/2;
		}
		xps = yps = 1;
	}

	if(!xps || !yps){
		fputs("internal error: sxp or syp not set\n",ttyout);
		goto out;
	}

	if(sxp < 0 || sxp >= sx || syp < 0 || syp >= sy){
		fputs("head of line not in canvas\n",ttyout);
		goto out;
	}

	if(!ns){
		if(lc){
			n = sy - syp;
		}else{
			n = sx - sxp;
		}
		ns = 1;
	}else{
		if((lc && syp + n > sy) || (!lc && sxp + n > sx)){
			fputs("tail of line not in canvas\n",ttyout);
			goto out;
		}
	}

	if(n <= 0){
		fputs("number of points < 1\n",ttyout);
		goto out;
	}
/*
	print program status
*/
	fprintf(ttyout,"%s: x: %d y: %d n %d",lc? "column":"line",sxp,syp,n);
	if(cur){
		fprintf(ttyout," cursor: %s (%d,%d)",cur==1? "OR" : "AND",cx,cy);
	}
	fputs("\n",ttyout);
	parse_uprs(ttyout,"output sent to:",ofn);
/*
	get data
*/
	np = n + outloc*2;
	nblk = PARSE_BLOCK_FLOAT(np);
	fx = (float *) malloc(nblk*512);
	fy = (float *) malloc(nblk*512);
	if(fx == NULL || fy == NULL){
		fputs("Can't allocate space for output vectors\n",ttyout);
		return;
	}else{
		have_vectors = 1;
	}

	if(lc == 0){
		for(i=0;i<n;i++){
			fy[i] = image16[syp*sx + sxp + i];
		}
	}else{
		for(i=0;i<n;i++){
			fy[i] = image16[sxp + (syp+i)*sx];
		}
	}
	for(i=0;i<n;i++){
		fx[i] = i;
	}
/*
	write data to file 
*/
	if(outloc == 0){
		unlink(ofn);

		if((ofd=fopen(ofn,"w")) == NULL){
			fprintf(ttyout,"Can't open: %s\n",ofn);
			goto out;
		}
		have_file = 1;

		parse_zbuf(head,256);
		head[0] = n;
		head[1] = 1;
		head[2] = 0;
		parse_to_can_shortv(head,256);

		if(256 != fwrite(head,sizeof(short int),256,ofd)){
			werr(ofn);
			goto out;
		}  

		parse_to_can_floatv(fy,n);
		if(nblk != fwrite(fy,512,nblk,ofd)){
			werr(ofn);
			goto out;
		}  

		goto out;
	}
/*
	send output to xplot
*/
	if(outloc == 1){
	xmax = xmin = fx[0];
	ymax = ymin = fy[0];

	for(i=0;i<n;i++){
		if(fy[i] > ymax)ymax = fy[i];
		if(fy[i] < ymin)ymin = fy[i];
		if(fx[i] > xmax)xmax = fx[i];
		if(fx[i] < xmin)xmin = fx[i];
	}

	if(ymin == ymax){
		fprintf(ttyout,"ymin = ymax = %f\n",ymin);
		goto out;
	}

	if(xmin == xmax){
		fprintf(ttyout,"xmin = xmax = %f\n",xmin);
		goto out;
	}

	fx[n] = xmin;
	fx[n+1] = (xmax - xmin) / xlen;
	fy[n] = ymin;
	fy[n+1] = (ymax - ymin) / ylen;

	if(plotss()){
		fprintf(ttyout,"Can't connect to: %s\n","xplot");
		goto out;
	}
	clear();

	plot(xxs,yys,-3);

	factor(scfac * 2.0);

	plot(.75,1.0,-3);


	axis(0.0,0.0,NULL,-999,xlen,0.0,fx[n],fx[n+1]);
	axis(0.0,0.0,NULL,999,ylen,90.0,fy[n],fy[n+1]);

	line(fx,fy,n,1,0,0);

	plot(0.0,0.0,999);
}
/*
clean up and return
*/
out:
if(have_vectors){
	free(fx);
	free(fy);
}

if(have_file){
	if(EOF == fclose(ofd)){
		fprintf(ttyout,"Error closing: %s\n",ofn);
	}
}

return;
}
