/*
images - handle command line of images and how to handle them
*/

#ident "$Id: images.c,v 1.36 2011/09/14 14:30:55 carl Exp $"

#include "xpic.h"
#include <pwd.h>

#include <dirent.h>
#include <fnmatch.h>
#include <sys/types.h>
#include <string.h>

#include <libgen.h>   //basename, dirname


static int my_scandir(const char *dir, struct dirent ***namelist,
            int (*select)(const struct dirent *, char *str), char *str,
            int (*compar)(const struct dirent **, const struct dirent **))
{
  DIR *d;
  struct dirent *entry;
  /*register*/ int i=0;
  size_t entrysize;

  if ((d=opendir(dir)) == NULL) return(-1);
	
  *namelist=NULL;
  while ((entry=readdir(d)) != NULL)
  {
    if (select == NULL || (select != NULL && (*select)(entry, str)))
    {
      *namelist=(struct dirent **)realloc((void *)(*namelist),
                 (size_t)((i+1)*sizeof(struct dirent *)));
	if (*namelist == NULL) return(-1);
	entrysize=sizeof(struct dirent)-sizeof(entry->d_name)+strlen(entry->d_name)+1;
	(*namelist)[i]=(struct dirent *)malloc(entrysize);
	if ((*namelist)[i] == NULL) return(-1);
	memcpy((*namelist)[i], entry, entrysize);
	i++;
    }
  }
  if (closedir(d)) return(-1);
  if (i == 0) return(0);
  if (compar != NULL)
    qsort((void *)(*namelist), (size_t)i, sizeof(struct dirent *), (int (*)(const void *,const void *)) compar);

  return(i);
}

static int my_alphasort(const struct dirent **a, const struct dirent **b)
{
  return(strcmp((*a)->d_name, (*b)->d_name));
}

void
close_file(void)
{
	if(itype == I_MI) mi_close();
#ifdef GZMI
	if(itype == I_GZMI) {
	  gzmi_close();
	  gzclose(gz_image_fd);
	}else{
#endif
	fclose(image_fd);
#ifdef GZMI
	}
#endif
	file_open = 0;
}

static short
tilde_expand(char *src, char *dest)
/* Expands a ~ or ~user directory shorthand into the full path.
   returns 0 on success and -1 on failure 
   
   *dest is MAXPATHLEN sized*/
{
	char *s;
	struct passwd *pw;
	char *sp;
	int uid;
	
	s = src;
	if(*src == '~'){
		if(isalpha(s[1])){
			sp = ++s;
			while(isalnum(*s)) s++;
			strncpy(dest,sp,s-sp);
			dest[s-sp] = 0;
			if(NULL == (pw = getpwnam(dest))){
				fprintf(ttyout,"Don't know user: %s\n",dest);
				return -1;
			}
		}else{
			uid = geteuid();
			if(NULL == (pw = getpwuid(uid))){
				fprintf(ttyout,"Don't know uid: %d\n",uid);
				return -1;
			}
			s++;
		}
		strcpy(dest,pw->pw_dir);
		strncat(dest,s,MAXPATHLEN-strlen(dest));
	}else{
		strncpy(dest,s,MAXPATHLEN);
	}
	return 0;
}

static char *
equal_format(char *fullpath, char *fmtpath)
/*fullpath and fmtpath are MAXPATHLEN long*/
{
	char *s = strrchr(fullpath, '=');
	if(s == NULL){
		strncpy(fmtpath, fullpath, MAXPATHLEN);
	}else{
		strncpy(fmtpath, fullpath, strlen(fullpath)-strlen(s));
		fmtpath[strlen(fullpath)-strlen(s)]='\0'; //why doesnt it copy the null 0?
	}
	return s;
}

void
open_file(void)
{
	char *s,t;
	static char path[MAXPATHLEN];

	if(file_open){
		close_file();
	}

	if((s=strrchr(ifn,'=')) != NULL){
		*s++ = 0;
		if(*s == 'i'){
			fprintf(ttyout,"applied type MI ");
			itype = I_MI;
		}else if(*s == 'l'){
			fprintf(ttyout,"applied type ");
			itype = I_LM;
			fprintf(ttyout,"LM-");
			t = *(s+1);
			switch (t){
			case 0:
			case 'h':
				isub = I_LM_HIGH;
 				fprintf(ttyout,"high ");
			break;
			case 'l': 
				isub = I_LM_LOW; 
 				fprintf(ttyout,"low ");
			break;
			case 'p': 
				isub = I_LM_PROJ; 
 				fprintf(ttyout,"proj ");
			break;
			default:
				fprintf(ttyout,"Illegal sub-image type\n");
				return;
			}
#ifdef GZMI
		}else if(*s == 'z'){
		  fprintf(ttyout, "applied type GZMI ");
		  itype = I_GZMI;
#endif
		}else{
			fprintf(ttyout,"Illegal image type\n");
			argc = -1;
			return;
		}
		if(s-1 == ifn){
			fprintf(ttyout,"\n");
			return;
		}
	}

	if(itype == I_MI){
		if((s=strrchr(ifn,'/')) == NULL) s = ifn;
		if(strchr(s,'.') == NULL)strcat(ifn,".mi");
	}

	if(itype == I_LM){
		if((s=strrchr(ifn,'/')) == NULL) s = ifn;
		if(strchr(s,'.') == NULL)strcat(ifn,".lm");
	}
#ifdef GZMI
	if(itype == I_GZMI){
		if((s=strrchr(ifn,'/')) == NULL) s = ifn;
		if(strchr(s,'.') == NULL)strcat(ifn,".gz.mi");
	}
#endif
	s = ifn;
	if (tilde_expand(ifn,path) < 0)
		return ;
	switch(itype){

	case I_MI:
	  if((image_fd=fopen(path,"r")) == NULL){
	    fprintf(ttyout,"open: Can't open %s: %s\n",path,strerror(errno));
	    return;
	  }
	  images_total = mi_total();
	  break;

	case I_LM:
	  if((image_fd=fopen(path,"r")) == NULL){
	    fprintf(ttyout,"open: Can't open %s: %s\n",path,strerror(errno));
	    return;
	  }
	  images_total = lm_total();
	  break;

#ifdef GZMI
	case I_GZMI:
	  if((gz_image_fd=gzopen(path,"r")) == NULL){
	    fprintf(ttyout,"open: Can't open %s: %s\n",path,strerror(errno));
	    return;
	  }
	  image_fd = NULL;
	  images_total = gzmi_total();
	  break;
#endif

	default:
		fputs("images (internal error): bad file type\n",ttyout);
		fclose(image_fd);
		return;
	}

	if(-1 == images_total){
	  if(itype == I_MI || itype == I_LM) {
		fclose(image_fd);
		return;
	  }
#ifdef GZMI
	  else {
	    gzclose(gz_image_fd);
	    return;
	  }
#endif
	}

	current_image = 0;
	file_open = 1;
	fprintf(ttyout,"Image(s): %d\n",images_total);
}

void
expand_open(int argcnt, char *arglist[])
{
  /*this function takes an argc/argv style pointer array and opens each
    element with open_file*/
  
  int opencount = ENV_N - current_env;
  int i;
  char *s;
  char current_dir[MAXPATHLEN];
  int cwdlen;
  char dname[MAXPATHLEN];
  char temp[MAXPATHLEN];
  char *bname;
  int  dnamelen;
  int status;
    
  if(opencount + 1 < argcnt)
    fputs("Warning: specified files exceed number of available environments! \
Excess will be truncated.\n",ttyout);
  /*run this loop from the second element of the array until one of the following:
    o we reach the end of the array
    o we run out of environments
    o the element is null
  */

  status = (int)getcwd(current_dir,MAXPATHLEN);
  if(status == 0) {
    fprintf(ttyout,"open: could not get current working directory! (%s)\n",strerror(errno));
    fflush(ttyout);
    return;
  }
  cwdlen = strlen(current_dir);

  for(i=1;(i<argcnt) && (i<=opencount) && (arglist[i] != NULL);i++)
    {
      //printf("processing arg %d of %d\n",i,argcnt);
      bname = strrchr(arglist[i],(int)'/') + sizeof(char);
      //printf("bname = %s\n",bname);
      dnamelen = bname - arglist[i];
      //printf("dnamelen = %d\n",dnamelen);
      if(dnamelen + 1 < MAXPATHLEN || dnamelen == 0) {
	strncpy(dname,arglist[i],dnamelen);
	dname[dnamelen] = '\0';
      }else{
	/* path name too long */
	fprintf(ttyout,"open: invalid path(too long, or null):<%s>\naborting.\n",arglist[i]);
	fflush(ttyout);
	return;
      }
      //printf("dname = %s (%d long)\n",dname,dnamelen);
      
      /* check if arglist[i] is an absolute path */
      if(arglist[i][0]=='/') {
	//printf("absolute path, changing to %s\n",dname);
	status = chdir(dname);
	if(status < 0){
	  fprintf(ttyout,"open: could not change directory to %s! (%s)\n",dname,strerror(errno));
	  fflush(ttyout);
	  return;
	}
      }else{ /* we'll have to assemble the path */
	strncpy(temp,current_dir,MAXPATHLEN);
	if(arglist[i][0] == '.' && arglist[i][1] == '/'){
	  s = arglist[i]+2*sizeof(char);
	}else{
	  s = arglist[i];
	}
	strncat(temp,"/",MAXPATHLEN-cwdlen);
	strncat(temp,dname,MAXPATHLEN-cwdlen-1);
	strncpy(dname,temp,MAXPATHLEN);
	//printf("relative path, changing to %s\n",dname);
	status = chdir(dname);
	if(status < 0){
	  fprintf(ttyout,"open: could not change directory to %s! (%s)\n",dname,strerror(errno));
	  fflush(ttyout);
	  return;
	}

      }

      strncpy(ifn,bname,KB_BUF);
      s = equal_format(ifn, temp); /*for display*/
      if (s != arglist[i]){
	fprintf(ttyout,"open: %s Env: %d ",basename(temp),current_env+1);
	fflush(ttyout);
      }

      errno=0; /*zero errno to see error reporting in open_file*/

      /* we shouldnt be writing into ifn, but this function doesnt take arguemnts.. */
      //printf("opening file %s\n",ifn);
      open_file();
      if(autoreset) reset(); /*mimic cmd_open old behavior*/

      if(i<opencount && !errno && images_total != -1 && file_open == 1){ /*dont inc env on: last open, error, out of envs, image error */
	if(i<argcnt-1)
	  env_inc(0); /* dont be verbose */
      }else{
	opencount++; //we skipped one, so the capacity increases
      }
      /* change back to starting directory */
      //printf("changing back to %s\n",current_dir);
      status = chdir(current_dir);
      if(status < 0){
	fprintf(ttyout,"open: could not change directory to %s! (%s)\n",current_dir,strerror(errno));
	fflush(ttyout);
	return;
      }
    }
  cwd2kb=1;
  write_env_panel(current_env);
}

static int
file_selector(const struct dirent *entry, char *str)
{
	int errcode;
	char temp[MAXPATHLEN];
	char plainname[MAXPATHLEN];
	char *s;
	char typechar='\0';
	
	/*select files from wildcard exp*/
	/*called from my_scandir*/
	
	strncpy(temp, str,MAXPATHLEN);
	s=equal_format(basename(temp),plainname);
	/*we've got to see if the user specified an image type with the current file
	  that hasnt been applied yet for this open cycle.  only then will we know 
	  the right extension to append*/
	if(s != NULL){
		if(strlen(s) >= 2)
			typechar=s[1];
		else //only an = is illegal, try to fix this for the user
			s[0]='\0';
	}  
	if((errcode=fnmatch(plainname,entry->d_name,FNM_PATHNAME))!=0)
	{
		if(errcode!=FNM_NOMATCH){
			fputs("Unknown Error in fnmatch",ttyout);
		}else{	//try name with default extension
			if(itype == I_MI || typechar == 'i')
				strncat(plainname,".mi",MAXPATHLEN-strlen(plainname));
			else if (itype == I_LM || typechar == 'l')
#ifdef MERLIN
				strncat(plainname,".l3f",MAXPATHLEN-strlen(plainname));
#else
				strncat(plainname,".lm",MAXPATHLEN-strlen(plainname));
#endif
#ifdef GZMI
			else if (itype == I_GZMI || typechar == 'z')
             			  strncat(plainname,".mi.gz",strlen(plainname)-MAXPATHLEN);
#endif
			if((errcode=fnmatch(plainname,entry->d_name,FNM_PATHNAME))==0)
			{
				return 1;	//file matched string
			}else if(errcode!=FNM_NOMATCH){
				fputs("Unknown Error in fnmatch",ttyout);
			}

		}
		return 0;	//file didnt match string
	}
	return 1;		//file matched string
}

void
parse_expand_open(int argcnt, char *arglist[])
{

	/*
		generate an argc/argv style array for each argument to open
		and pass it to expand_open
	
	*/
	
	char dir[MAXPATHLEN];
	char temp[MAXPATHLEN];
	struct dirent **namelist=NULL;
	char **ar, *s;
	int entries,i;
	short dots; //=0,1 or 2
	short tagonly=0;
		
	if(argcnt < 2){
		fputs("parse_expand_open: internal error\n",ttyout);
		return;
	}


	for(i=1;i<argcnt;i++)
	{
		dots=0;
		tagonly=0;
		strncpy(dir,arglist[i],MAXPATHLEN);
	
		//expand ~ and ~user
		if(tilde_expand(dir,temp)<0)
			return ;
	
		//copy expanded string to dir
		strncpy(dir,dirname(temp),MAXPATHLEN);

		//scan directories for 
		entries = my_scandir (dir, &namelist, file_selector, arglist[i], my_alphasort);
		
		if(entries >= 1)
			dots += !strcmp(namelist[0]->d_name,"." );
		if(entries >= 2)
			dots += !strcmp(namelist[1]->d_name,"..");
		
		s = equal_format(arglist[i],temp); //set up temp for use if found, get s
		tagonly = (s == arglist[i]);/*behave differently if its only a tag*/
		if(entries < 0) /*error check my_scandir*/
		{
			fprintf(ttyout,"open: %s: %s\n",dir,strerror(errno));
			fflush(ttyout);
		}
		else if(((entries == 0) || entries == dots) && (!tagonly) ){
			if(strpbrk(arglist[i],"*?[]^~")){ 	//file_selector matches to basename( arglist[i] )
				fprintf(ttyout,"open: %s: No match.\n",basename(arglist[i]));
			}else{
				fprintf(ttyout,"open: %s: No such file.\n",basename(arglist[i]));
			}
			fflush(ttyout);
		}else{	//entries found
			int cnt;
			int allocate_size;
			
			if(!tagonly){
				if(file_open && i!=1){	//if theres a file open change to the next available env
        	                        if(NULL == getcwd(temp,MAXPATHLEN)) perror("xpic: parse_expand_open: getcwd error");
					env_upd_dir(current_env+1,temp);
					env_inc(0);
				}
		
				/*make the array and error checks*/
				
				strcat(dir,"/");
				if(strpbrk(arglist[i],"*?[]^~")){			
					fprintf(ttyout, "open: %d file(s) matched to \"%s\"\n", CRC_MIN(entries-dots,ENV_N-current_env),basename(temp));
					fflush(ttyout);
				}

				ar = malloc((entries+1-dots)*sizeof(char*));
				if(ar==NULL){
					fputs("parse_expand_open: Error: insufficient memory\n",ttyout);
					fflush(ttyout);
					return;
				}
			
				ar[0]=strdup("");
				if(ar[0]==NULL){
					fputs("parse_expand_open: Error: insufficient memory\n",ttyout);			
					fflush(ttyout);
					free(ar);
					return;				
				}
				for(cnt=1;cnt<=entries-dots;cnt++)
				{
					/*add size of directory, size of file name, and size of trailing =, and size for \0*/
					allocate_size = strlen(namelist[cnt-1+dots]->d_name)+strlen(dir)+((s==NULL)?0:strlen(s))+1;
					/*make the array*/
					if((ar[cnt] = malloc(allocate_size))==NULL){
						fputs("parse_expand_open: Error: insufficient memory\n",ttyout);
						fflush(ttyout);
						/*free allocated memory before error return*/
						for(entries=0;entries<cnt;entries++)
							free(ar[entries]);
						free(ar);
						return;
					}else{
						strcpy(ar[cnt],dir);
						strcat(ar[cnt],basename(namelist[cnt-1+dots]->d_name));
						if(s != NULL) /*if there is an =param, append it*/
							strcat(ar[cnt],s);
#ifdef DEBUG
						if(strlen(ar[cnt])+1!=allocate_size){ /*+1 for null 0*/
							fprintf(stdout,"WARNING: Memory allocation error: cnt=%d overflow data=%s\n",cnt,ar[cnt]);
							fflush(stdout);
						}
#endif /*DEBUG*/						
					}
				}
#ifdef DEBUG
				fprintf(stderr,"\nDEBUG MODE: Passing expanded argument vector to expand_open:");
				fprintf(stderr,"\nentries: %d:\n These came from open command argument %d: %s",entries, i, arglist[i]);
				for(cnt=1;cnt<=entries-dots;cnt++)
				{
					fprintf(stderr,"Entry %d: %s\n",i,ar[cnt]);
				}
#endif /*DEBUG*/
				expand_open(entries+1-dots, ar);
				for(cnt=0;cnt<=entries-dots;cnt++){
					free(ar[cnt]);
				}
				free(ar);
				free(namelist);
			}else{ /*tagonly*/
				expand_open(2, arglist+i-1);
			}

		}
	}
}

void
images(char *s)
{
	int gr=0,gc=0,grid,i,sec,row,col,n,num[MIMAGES+1],nn,sgn,prev;
	char *sp;

	if(s[strlen(s)-1] == '\n')s[strlen(s)-1] = '\0';

	for(grid=row=col=0,sp=s;*sp;sp++){
		if(*sp == '/'){
			row++;
		}else if(*sp == '|'){
			col++;
		}else if(*sp == '@'){
			grid = 1;
			if(*(sp+1)){
				fputs("@ not last character on line\n",ttyout);
				return;
			}
		}
	}

	if(row && col){
		fputs("Can't specify columns and rows\n",ttyout);
		return;
	}

	if(grid && (row || col)){
		fputs("Can'use @ and columns or rows\n",ttyout);
		return;
	}

	for(sp=s,prev=sec=n=0;;) if(isdigit(*sp)){
		nn = 0;
		while(isdigit(*sp)){
			nn = nn*10 + *sp - '0';
			sp++;
		}
		num[n++] = nn;
		if(n > MIMAGES){
			fputs("Too many images\n",ttyout);
			return;
		}
		if(*sp == '-'){
			sp++;
			nn = 0;
			while(isdigit(*sp)){
				nn = nn*10 + *sp - '0';
				sp++;
			}
			if(nn == 0){
				fputs("invalid number after -\n",ttyout);
				return;
			}
			if((nn-num[n-1]) > 0)
				sgn = 1;
			else
				sgn = -1;
			
			for(i=abs(nn-num[n-1]);i>0;i--){
				num[n] = num[n-1] + sgn;
				n++;
				if(n > MIMAGES){
					fputs("Too many images\n",ttyout);
					return;
				}
			}
		}
		prev = 1;
	}else if(*sp == ',' || *sp == ' '){
		if(!grid && (prev == 0 || prev == 2)){
			num[n++] = -1;
			if(n > MIMAGES){
				fputs("Too many images\n",ttyout);
				return;
			}
		}
		sp++;
		prev = 2;
	}else if(*sp == '/' || *sp == '|'  || *sp == '@' || !*sp){
		if(!grid && prev == 2){
			num[n++] = -1;
			if(n > MIMAGES){
				fputs("Too many images\n",ttyout);
				return;
			}
		}

		if(grid){
			gc = ceil(sqrt((double)n));
			gr = ceil(((double) n) / gc);

			da_x = 0;
			da_y = 0;
			da_xw = canvas_size_x;
			da_yw = canvas_size_y;

			do_clear(1);
		}

		for(i=0;i<n;i++){
			if(num[i] < 0){
				continue;
			}else if(!row && !col){
				if(n == 1){
					load16(num[i],0);
				}else{
					if(!grid){
						da_yw = canvas_size_y;
						da_xw = canvas_size_x / n;
						da_y = 0;
						da_x = da_xw * i;
					}else{
						da_yw = canvas_size_y / gr;
						da_xw = canvas_size_x / gc;
						da_x = (i%gc) * da_xw;
						da_y = (i/gc) * da_yw;
					}
					load16(num[i],0);
				}
			}else if(row){
				da_yw = canvas_size_y / (row + 1);
				da_xw = canvas_size_x / n;
				da_y = sec * da_yw;
				da_x = da_xw * i;
				load16(num[i],0);
			}else if(col){
				da_xw = canvas_size_x / (col + 1);
				da_yw = canvas_size_y / n;
				da_x = sec * da_xw;
				da_y = da_yw * i;
				load16(num[i],0);
			}
		}
		sec++;
		if(grid) return;
		if(!*sp++) return;
		prev = n = 0;
	}else{
		fprintf(ttyout,"Illegal character: '%c'\n",*sp);
		return;
	}
	fflush(ttyout);
}

static void
images_syntax(void)
{
	fputs("Synopsis: load images\n",ttyout);
	min_abbreviation();
	fputs("Related: file, corner, center, cine, +,-, open\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("images list  * load images in list into display\n",ttyout);
	fputs("images list@ * load images in list into display\n",ttyout);
	fputs("             * list: r1,r2,[|/],r3,r4... \n",ttyout);
	fputs("             * range (ri): n-m or n or nil\n",ttyout);
	fputs("             * /: break into rows\n",ttyout);
	fputs("             * |: break into columns\n",ttyout);
	fputs("             * @: break into grids, / & | not allowed\n",ttyout);
	fputs("             * 'images' optional\n",ttyout);
}

void
cmd_images(void)
{
	if(argc < 2){
		images_syntax();
		return;
	}
	images(argv[1]);
}

static void
corner_syntax(void)
{
	fputs("Synopsis: set corner of image\n",ttyout);
	min_abbreviation();
	fputs("Related: center, reset\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("corner x y     * set corner pixel to (x,y)\n",ttyout);
	fputs("corner         * reverts to center of image\n",ttyout);
}

static void
center_syntax(void)
{
	fputs("Synopsis: set center of image\n",ttyout);
	min_abbreviation();
	fputs("Related: corner, reset\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("center x y     * set center pixel to (x,y)\n",ttyout);
	fputs("center         * center reverts to center of image\n",ttyout);
}

void
cmd_center(void)
{
	int i;

	if(argc != 3 && argc != 1){
		center_syntax();
		return;
	}

	if(argc == 1){
		cen_set = 0;
		cor_set = 0;
		fputs("Center and corner off\n",ttyout);
	}else{
		i = parse_atof(argv[1]);
		if(parse_atof_error()) return;
		cen_x = i;

		i = parse_atof(argv[2]);
		if(parse_atof_error()) return;
		cen_y = i;

		cen_set = 1;
		cor_set = 0;
	}
}

void
cmd_corner(void)
{
	int i;

	if(argc != 3 && argc != 1){
		corner_syntax();
		return;
	}

	if(argc == 1){
		cor_set = 0;
		cen_set = 0;
		fputs("Corner and center off\n",ttyout);
	}else{
		i = parse_atof(argv[1]);
		if(parse_atof_error()) return;
		cor_x = i;

		i = parse_atof(argv[2]);
		if(parse_atof_error()) return;
		cor_y = i;

		cen_set = 0;
		cor_set = 1;
	}
}

static void
print_suffixes(void)
{
	fputs("Types: i      * MI-files \n",ttyout);
	fputs("       z      * MIGZ-files\n",ttyout);
	fputs("       l      * LM high energy\n",ttyout);
	fputs("       lh     * LM high energy\n",ttyout);
	fputs("       ll     * LM low energy\n",ttyout);
	fputs("       lp     * LM projections\n",ttyout);

}

static void
open_syntax(void)
{
	fputs("Synopsis: opens files or set file types\n",ttyout);
	min_abbreviation();
	fputs("Related: file, reopen, images, status, header\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("open file1[=t] [file 2][=t] ... [file n][=t]\n",ttyout);
	fputs("\t\t* open n file(s) and set type to t\n",ttyout);
	fputs("open =t\t\t* set type to t\n",ttyout);
	fputs("open [shell wildcard expression][=t]\n\t\t* open all files matching [expression], set type to t\n",ttyout);
	print_suffixes();
}

void
cmd_open(void)
{
	if(argc < 2)
	{
		open_syntax();
		return;
	}
	parse_expand_open(argc,argv);
}

static void
reopen_syntax(void)
{
	fputs("Synopsis: reopens files\n",ttyout);
	min_abbreviation();
	fputs("Related: file, open, images\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("reopen        * reopen file\n",ttyout);
	fputs("reopen =t     * reopens file as type t\n",ttyout);
	print_suffixes();
}

void
cmd_reopen(void)
{
	if(argc < 1 || argc > 2){
		reopen_syntax();
		return;
	}
	if(file_open){
		if(argc == 2){
			strncat(ifn,"=",KB_BUF-strlen(ifn));
			strncat(ifn,argv[1],KB_BUF-strlen(ifn));
		}
		fprintf(ttyout,"reopening: %s\n",ifn);
		open_file();
	}else{
		fputs("need to open a file first\n",ttyout);
	}
}

static void
difference_syntax(void)
{
	fputs("Synopsis: subtract image from current display\n",ttyout);
	min_abbreviation();
	fputs("Related: set\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("difference #         * subtract image # from display\n",ttyout);
	fputs("difference # offset  * add offset to result\n",ttyout);
	fputs("                     * offset saved for next difference\n",ttyout);
	fputs("                     * offset also available with set command\n",ttyout);
}


void
cmd_difference(void)
{
	int i;

	if(argc == 1){
		fprintf(ttyout,"offset: %d\n",diff_offset);
		return;
	}

	if(argc < 1 || argc > 3){
		difference_syntax();
		return;
	}


	if(argc == 3){
		i = parse_atof(argv[2]);
		if(parse_atof_error()) return;
		diff_offset = i;
	}

	i = parse_atof(argv[1]);
	if(parse_atof_error()) return;

	load16(i,1);
}

void
next_prev(int mode)
{
	if(!file_open){
		fputs("Open a file first\n",ttyout);
		return;
	}
	if(current_image == 0){
		current_image = 1;
	}else if(mode == 1){
		current_image += pm_skip;
	}else{
		current_image -= pm_skip;
	}

	if(current_image > images_total) current_image = 1;
	if(current_image <= 0) current_image = images_total;

	load16(current_image,0);
	fflush(ttyout);
}

static void
next_syntax(void)
{
	fputs("Synopsis: displays next image in file\n",ttyout);
	min_abbreviation();
	fputs("Related: -, cine, panel, set\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("+          * bumps to next image\n",ttyout);
	fputs("           * '+' on panel is the same\n",ttyout);
	fputs("*options\n",ttyout);
	fputs("*set skip  * number of images to skip between\n",ttyout);
}

static void
previous_syntax(void)
{
	fputs("Synopsis: displays previous image in file\n",ttyout);
	min_abbreviation();
	fputs("Related: +, cine, panel, set\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("-          * bumps to previous image\n",ttyout);
	fputs("           * '-' on panel is the same\n",ttyout);
	fputs("*options\n",ttyout);
	fputs("*set skip  * number of images to skip between\n",ttyout);
}

void
cmd_next(void)
{
	if(argc != 1){
		next_syntax();
		return;
	}

	next_prev(1);

}

void
cmd_previous(void)
{
	if(argc != 1){
		previous_syntax();
		return;
	}

	next_prev(-1);
}

void
func_scroll(int pm,int mask)
/* pm: 1= postive -1=negative scroll */
/* mask: 0=both directions, 1=no x, 2=no y */
{
	int *px,*py;
	int nx,ny;
	char *mode;
	int sxp,syp;

	if(!file_open){
		fputs("Open a file first\n",ttyout);
		return;
	}

	if(current_nx <= 0 || current_ny <= 0){
		fputs("Zero or negative image size. Can't scroll\n",ttyout);
		return;
	}

	if(cor_set){
		mode = "corner";
		px = &cor_x;
		py = &cor_y;
	}else{
		if(!cen_set){
			fprintf(ttyout,"center or corner not set, using default center\n");
			cen_x = current_nx / 2;
			cen_y = current_ny / 2;
			cen_set = 1;
		}
		mode = "center";
		px = &cen_x;
		py = &cen_y;
	}

	sxp = scroll_x * pm;
	syp = scroll_y * pm;
	if(mask == 1) sxp = 0;
	if(mask == 2) syp = 0;

	nx = *px + sxp;
	while(nx < 0) nx += current_nx;
	nx %= current_nx;

	ny = *py + syp; 
	while(ny < 0) ny += current_ny;
	ny %= current_ny;

	fprintf(ttyout,"%s scrolled from (%d,%d), by (%d,%d), to (%d,%d)\n",
		mode, *px, *py, sxp, syp, nx, ny);

	*px = nx;
	*py = ny;

	load16(current_image,0);
}

static void
scroll_syntax(void)
{
	fputs("Synopsis: scroll images\n",ttyout);
	min_abbreviation();
	fputs("Related: panel, set\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("scroll      * move right & down by size of display\n",ttyout);
	fputs("scroll x y  * move image x right and y down\n",ttyout);
	fputs("            * x & y saved for next call\n",ttyout);
}

void
cmd_scroll(void)
{
	int ix,iy;

	if(argc < 1 || argc > 3){
		scroll_syntax();
		return;
	}

	if(!file_open){
		fputs("Open a file first\n",ttyout);
		return;
	}

	if(argc > 1){
		ix = parse_atof(argv[1]);
		if(parse_atof_error()) return;

		iy = parse_atof(argv[2]);
		if(parse_atof_error()) return;

		scroll_x = ix;
		scroll_y = iy;
	}

	func_scroll(1,0);
}

void
scroll_init(void)
{
	scroll_x = canvas_size_x;
	scroll_y = canvas_size_y;
}
