/*
lc - extract lines and columns from xpic 
*/

static char rcsid[] = "$Id: lc.c,v 1.4 2011/09/14 10:45:27 carl Exp $";

#include <parse.h>
#include <string.h>
#include <xpic_port.h>

static int xp=-1,yp=-1,xps,yps;
static char *ofn = "line.da";
static int n=-1,ns;
static int lc;
static int cur;
static int debug=0;
static int print=1;
static int port = PORT_SOCKET;
static char *host = "";

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'l',&lc,PARSE_FLAG_CLR,"line mode (default)"},
	{'c',&lc,PARSE_FLAG_SET,"column mode"},
	{'d',&cur,PARSE_FLAG_SET,"use cursor for x OR y"},
	{'D',&cur,PARSE_FLAG_VALUE(2),"use cursor for x AND y"},
	{'p',&print,PARSE_FLAG_CLR,"don't print program status"},
	{'?',(int *)parse_help_basic,0,0},
	{0,0,0,0}
};

struct PARSE_STRUCT parse_table[] = {
	{"x",PARSE_INT,(PARSE_ARG)&xp,"column"},
	{"y",PARSE_INT,(PARSE_ARG)&yp,"row (default 1/2 nrows)"},
	{"n",PARSE_INT,(PARSE_ARG)&n,"number of points"},
	{"of",PARSE_CHAR,(PARSE_ARG)&ofn,"output filename (float da-file)"},
	{"host",PARSE_CHAR,(PARSE_ARG)&host,"xpic host"},
	{"port",PARSE_INT,(PARSE_ARG)&port,"xpic port"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0,0}
};

int
main(argc,argv)
int argc;
char **argv;
{
	int i,np,sx,sy,w,l,m;
	int hl,pl;
	int type,cx,cy,cxw,cyw;
	unsigned short *d;
	float *f;
	char *xhost,*pnum;
	char xpm[132];			/* xpic's message */

	parse_basic(argc,argv);
	ofn = parse_suffix_malloc(ofn,"da");

	if(!parse_opexist("host") && !parse_opexist("port")){
		xhost = NULL;
	}else{
		hl = parse_opexist("host") * strlen(host);
		pnum = parse_itoa(port,NULL);
		pl = strlen(pnum) + 1;
		xhost = parse_malloc_char(hl+pl+1);
		xhost[0] = 0;
		if(hl) strcat(xhost,host);
		strcat(xhost,":");
		strcat(xhost,pnum);
	}
	xpic_open(xhost,0);
	xpic_status(&sx,&sy,&w,&l,&m);

	ns = parse_opexist("n");
	xps = parse_opexist("x");
	yps = parse_opexist("y");

	if(cur){
		if(lc && xps){
			fputs("can't set column in cursor mode\n",stderr);
			exit(1);
		}
		if((!lc) && yps){
			fputs("can't set row in cursor mode\n",stderr);
			exit(1);
		}
		xpic_cursor(&type,&cx,&cy,&cxw,&cyw);
		if(!type){
			fputs("cursor not enabled\n",stderr);
			exit(1);
		}
		if(lc){
			xp = cx;
			xps = 1;
			if(!yps){
				yps = 1;
				yp = (cur ==2)? cy : 0;
			}
		}else{
			yp = cy;
			yps = 1;
			if(!xps){
				xps = 1;
				xp = (cur ==2)? cx : 0;
			}
		}
	}else{
		if(lc){
			if(!xps) xp = sx/2;
			if(!yps) yp = 0;
		}else{
			if(!xps) xp = 0;
			if(!yps) yp = sy/2;
		}
		xps = yps = 1;
	}

	if(!xps || !yps){
		fputs("internal error: xp or yp not set\n",stderr);
		exit(1);
	}

	if(xp < 0 || xp >= sx || yp < 0 || yp >= sy){
		fputs("head of line not in canvas\n",stderr);
		exit(1);
	}

	if(!ns){
		if(lc){
			n = sy - yp;
		}else{
			n = sx - xp;
		}
		ns = 1;
	}else{
		if((lc && yp + n > sy) || (!lc && xp + n > sx)){
			fputs("tail of line not in canvas\n",stderr);
			exit(1);
		}
	}
/*
	print program status
*/
	if(print){
		parse_wtext("****************** lc *********************");
		parse_pdt("---> start time");
		parse_rcsid_print(rcsid);
		parse_prs("output file:",ofn);
		parse_prs("mode:",lc? "column" : "line");
		parse_prm("cursor used (No, Single, Double)","nsd",cur);
		parse_pri("x:",xp);
		parse_pri("y:",yp);
		parse_pri("n:",n);
	}
/*
	get data, convert to real, and write it out
*/
	if(lc) xpic_transpose(1);
	if(debug)printf("x,y,n: %d %d %d\n",xp,yp,n);

	d = (unsigned short *) parse_malloc(n*sizeof(short));
	if(n != (np = xpic_read(xp,yp,d,n))){
		fprintf(stderr,"xpic_read error, %d points read\n",np);
		exit(1);
	}
	f = (float *) parse_malloc(n*sizeof(float));
	for(i=0;i<n;f[i++] = *d++);
	parse_vwrite(f,n,0,ofn,"");
/*
	send message to xpic's window
*/
	sprintf(xpm,"lc: head=(%d,%d) n=%d %s\n",
		xp,yp,n,
		lc? "column" : "line");

	xpic_message(xpm);
/*
	clean up and exit
*/
        if(print)parse_pdt("---> stop time");
	exit(0);
}
