/*
load16 - transfer contents of file into 16 bit storage
*/

#ident "$Id: load16.c,v 1.14 2007/05/22 13:22:45 ccrawford Exp $"

#include "xpic.h"

void
load16(int image_num,int diff_mode)
{
	register int i=0,j;
	CPIXEL *mip,*line=NULL;
	int nx=0,ny=0,first_da_x,first_da_y;
	int image_x,image_y,nx_da,ny_da,jump_x;
	char label[PARSE_MI_LEN_LABEL+1];
	int before,after,bhalf,ahalf;

	if(!file_open){
		fputs("Open a file first\n",ttyout);
		return;
	}

	if(image_num < 1 || image_num > images_total){
		fprintf(ttyout,"Bad image number\n");
		return;
	}
				
	if(itype == I_LM){
		if(-1 == lm_image(&nx,&ny,image_num,&label[0])) return;
	}else if(itype == I_MI){
		if(-1 == mi_image(&nx,&ny,image_num,&label[0])) return;
	}
#ifdef GZMI
	else if(itype == I_GZMI){
		if(-1 == gzmi_image(&nx,&ny,image_num,&label[0])) return;
	}
#endif

	current_image = image_num;
	current_nx = nx;
	current_ny = ny;

	fprintf(ttyout,"Image: %d (%d,%d)",image_num,nx,ny);
	if(strlen(label)) fprintf(ttyout," \"%s\"",label);
	fputc('\n',ttyout);

	if(cen_set && (cen_x < 0 || cen_x >= nx || cen_y < 0 || cen_y >= ny)){
		fputs("Bad center. Reset center\n",ttyout);
		return;
	}

	if(cor_set && (cor_x < 0 || cor_x >= nx || cor_y < 0 || cor_y >= ny)){
		fputs("Bad corner. Reset corner\n",ttyout);
		return;
	}

	if(cor_set){
		image_x = cor_x;
		first_da_x = da_x;
		if(nx - cor_x <= da_xw){
			jump_x = cor_x;
			nx_da = nx - cor_x;
		}else{
			nx_da = da_xw;
			jump_x = nx - da_xw;
		}

		image_y = cor_y;
		first_da_y = da_y;
		if(ny - cor_y <= da_yw){
			ny_da = ny - cor_y;
		}else{
			ny_da = da_yw;
		}
	}else{
		if(!cen_set){
			cen_x = nx / 2;
			cen_y = ny / 2;
		}

		bhalf = da_xw/2;
		ahalf = da_xw - bhalf;
		if(cen_x < bhalf){
			first_da_x = bhalf - cen_x;
			image_x = 0;
			before = cen_x ;
		}else{
			first_da_x = 0;
			image_x = cen_x - bhalf;
			before = bhalf;
		}
		if(nx-cen_x > ahalf){
			after = ahalf;
		}else{
			after = nx - cen_x;
		}
		nx_da = before + after;
		jump_x = nx - nx_da;
		first_da_x += da_x;

		bhalf = da_yw/2;
		ahalf = da_yw - bhalf;
		if(cen_y < bhalf){
			first_da_y = bhalf - cen_y;
			image_y = 0;
			before = cen_y ;
		}else{
			first_da_y = 0;
			image_y = cen_y - bhalf;
			before = bhalf ;
		}
		if(ny-cen_y > ahalf){
			after = ahalf;
		}else{
			after = ny - cen_y;
		}
		ny_da = before + after;
		first_da_y += da_y;
	}

	if(itype == I_MI || itype == I_LM) {
	  if(-1 == fseek(image_fd,(long) (image_y*nx + image_x) * sizeof(CPIXEL),SEEK_CUR)){
	    fputs("Error in seeking to beginning of image\n",ttyout);
	    return;
	  }
	}
#ifdef GZMI
	else {
	  if(-1 == gzseek(gz_image_fd,(long) (image_y*nx + image_x) * sizeof(CPIXEL),SEEK_CUR)){
	    fputs("Error in seeking to beginning of image\n",ttyout);
	    return;
	  }
	}
#endif
	mip = image16 + first_da_y*canvas_size_x + first_da_x;

	if(autoclear) do_clear(0); 

	if(diff_mode){
		if(NULL == (line = (CPIXEL *)malloc(nx_da*sizeof(CPIXEL)))){
			fputs("Can't get temp memory for difference\n",ttyout);
			return;
		}
	}

	for(i=0;i<ny_da;i++){
	  if(itype == I_MI || itype == I_LM) {
	    if(nx_da != 
	       fread(diff_mode? line: mip,sizeof(CPIXEL),nx_da,image_fd)){
	      fputs("Error in reading image\n",ttyout);
	      load_pix(1);
	      return;
	    }
	  }
#ifdef GZMI
	  else {
	    /* gzread returns number of bytes, not number of elements */
	    if(sizeof(CPIXEL)*nx_da != 
	       (j=gzread(gz_image_fd,diff_mode? line: mip,sizeof(CPIXEL)*nx_da))){
	      fputs("Error in reading gz-image\n",ttyout);
	      load_pix(1);
	      return;
	    }
	  }
#endif
	   	if(itype != I_LM){
			parse_from_can_shortv((short *)(diff_mode? line: mip),
				nx_da);
		}else{
			ican_shortv((short *)(diff_mode? line: mip),nx_da);
		}

		if(diff_mode){
			for(j=0;j<nx_da;j++){
				if((int)line[j] > mi_max)line[j] = mi_max;
				mip[j] -= line[j] - diff_offset;
				if(mip[j] > mi_max)mip[j] = 0;
			}
		}else{
			for(j=0;j<nx_da;j++){
				if((int)mip[j] > mi_max)mip[j] = mi_max;
			}
		}
		if(i<ny_da-1) { /* no need to seek to very end of file */
		  if(itype == I_MI || itype == I_LM) {
		    if(-1 == fseek(image_fd,(long) jump_x*sizeof(CPIXEL),SEEK_CUR)){
		      fputs("Error in seeking to requested image\n",ttyout);
		      load_pix(1);
		      return;
		    }
		  }
#ifdef GZMI
		  else {
		    if(-1 == gzseek(gz_image_fd,(long) jump_x*sizeof(CPIXEL),SEEK_CUR)){
		      fputs("Error in seeking to requested image\n",ttyout);
		      load_pix(1);
		      return;
		    }
		  }
#endif
		}
		mip += canvas_size_x;
	}
	if(diff_mode){
		free(line);
	}

	da2_x = first_da_x;
	da2_y = first_da_y;
	da2_xw = nx_da;
	da2_yw = ny_da;;

	if(automagnify) do_mag(0,0,0);
	if(autowl) do_awl();
	if(autoclear | automagnify) 
		load_pix(1);
	else
		load_pix(2);
	fflush(ttyout);
}
