/*
lpr - send image to lpr using postscript halftoning
*/

#ident "$Id: lpr.c,v 1.5 2007/05/22 13:22:45 ccrawford Exp $"

#include "xpic.h"

static int aspect_s,width_s,height_s;
static int status;
static char ofn[80] = "img.eps";
static int oplace =0;		/* output place: 0=pipe 1=file */
static int otype =0;		/* output type: 0=ps 1=eps */

static void halftone_syntax(void);

static struct PARSE_OPTION_TABLE options[] = {
	{"width",PARSE_FLOAT,(PARSE_ARG)&half_width},
	{"width",PARSE_SET,(PARSE_ARG)&width_s},
	{"height",PARSE_FLOAT,(PARSE_ARG)&half_height},
	{"height",PARSE_SET,(PARSE_ARG)&height_s},
	{"xo",PARSE_FLOAT,(PARSE_ARG)&half_xo},
	{"yo",PARSE_FLOAT,(PARSE_ARG)&half_yo},
	{"gamma",PARSE_FLOAT,(PARSE_ARG)&gamma_halftone},
	{"printer",PARSE_STRING,(PARSE_ARG)half_printer},
	{"printer",PARSE_CLR,(PARSE_ARG)&oplace},
	{"printer",PARSE_CLR,(PARSE_ARG)&otype},
	{"=",PARSE_STRING,(PARSE_ARG)ofn},
	{"=",PARSE_SET,(PARSE_ARG)&oplace},
	{"of",PARSE_STRING,(PARSE_ARG)ofn},
	{"of",PARSE_SET,(PARSE_ARG)&oplace},
	{"?",PARSE_PROC,(PARSE_ARG)halftone_syntax},
	{0,0,0,0}
};

static struct PARSE_FLAG_TABLE flags[] = {
	{'a',&aspect_s,PARSE_FLAG_SET},
	{'e',&otype,PARSE_FLAG_SET},
	{'e',&oplace,PARSE_FLAG_SET},
	{'f',&oplace,PARSE_FLAG_SET},
	{'l',&half_orientation,PARSE_FLAG_SET},
	{'n',&oplace,PARSE_FLAG_CLR},
	{'n',&otype,PARSE_FLAG_CLR},
	{'p',&half_orientation,PARSE_FLAG_CLR},
	{'s',&status,PARSE_FLAG_SET},
	{'?',(int *)halftone_syntax,0},
	{0,0,0,0}
};

static void
lpr_send(void)
{
	FILE *pout;
	static int my_client_object;
	static int *me = &my_client_object;
	DPIXEL8 *image;
	int i,j,k;
	char c[100];
	int ppmin,ppmax,pp,pid;

	image = pr2_data;


	if(oplace == 0){

#ifdef HAS_LP
		strcpy(c,"lp -o nobanner -d ");
#else
		strcpy(c,"lpr -P");
#endif
		strcat(c,half_printer);

		fprintf(ttyout,"Output piped to: %s\n",c);
		fflush(ttyout);

		if((pid = fork()) == -1){
			pperror("Can't fork copy of xpic");
			exit(1);
		}
		if(pid != 0){
			display_wait3((int)me,pid);
			return;
		}

		fclose(ttyout);
		fclose(stderr);
		fclose(stdout);
		ttyout = fopen("/dev/console","w");

		display_destroy();

		if(NULL == (pout = popen(c,"w"))){
			fprintf(ttyout,"Can't pipe to: %s\n",c);
			return;
		}
	}else{
		parse_suffix(ofn,otype? "eps":"ps");

		if(NULL == (pout = fopen(ofn,"w"))){
			fprintf(ttyout,"Can't open: %s\n",ofn);
			return;
		}

		fprintf(ttyout,"Output saved in: %s\n",ofn);
		fflush(ttyout);
	}
/* 
	SEND PostScript DEFINES 
*/
	if(otype == 0){
		fprintf(pout,"%%!PS-Adobe-2.0\n");
	}else{
		fputs("%!PS-Adobe-2.0 EPSF-2.0\n",pout);
		fprintf(pout,"%%%%BoundingBox: %d %d %d %d\n",
			0,0,(int)half_width *72,(int)half_height*72);
	}
	fprintf(pout,"save\n");
	fprintf(pout,"/inch { 72 mul }def\n");

	fprintf(pout,"/picstr %d string def\n",canvas_size_x);

	fprintf(pout,"gsave\n");
/* 
	translate to lower left of image 
*/
	if(!otype){
		fprintf(pout,"%3.1f inch %3.1f inch translate\n",
			half_xo,half_yo);
		if(half_orientation){
			fprintf(pout,"%3.1f inch %3.1f inch translate\n", 
				half_height,0.0);
			fprintf(pout,"90 rotate\n");
		}
	}

	/* THIS MAKES THE IMAGE width INCHES by height INCHES */
	fprintf(pout,"%3.1f inch %3.1f inch scale\n",half_width,half_height);

	/* THIS READYS THE PRINTER TO RECEIVE DATA - IT TELLS THE IMAGE OPERATOR
       HOW TO INTERPRET THE IMAGE DATA                                      */
	fprintf(pout," %d %d %d [%d 0 0 %d 0 %d]\n",canvas_size_x,
	    canvas_size_y,
	    8,
	    canvas_size_x,
	    (-1) * canvas_size_y,
	    canvas_size_y);
	fprintf(pout,"{ currentfile picstr readhexstring pop } image\n");

/*
	send out image itself
*/
	k = 0;
	ppmin = 1000000;
	ppmax = -ppmin;

	for(j=0;j<canvas_size_y;j++){
		for(i=0;i<canvas_size_x;i++){ 
			pp = image[dsp_bytes_pixel*i];
			ppmin = CRC_MIN(ppmin,pp);
			ppmax = CRC_MAX(ppmax,pp);
			fprintf(pout,"%02x",pp);
			if(++k == 30){
				fputc('\n',pout);
				k = 0;
			}
		}
		image += pr2_pixels;
	}

	if(ppmin < 0 || ppmax > 255){
		fprintf(ttyout,"halftone error: data out of range (%d,%d)\n",
			ppmin,ppmax);
		return;
	}
	fprintf(pout,"\n");
	fprintf(pout,"grestore\n");

	fprintf(pout,"showpage\n");
	fprintf(pout,"restore\n");

	if(oplace == 0){
		pclose(pout);
		exit(0);
	}else{
		fclose(pout);
		return;
	}
}

static void
printer_syntax(void)
{
	fputs("Synopsis: manages printer for halftone command\n",ttyout);
	min_abbreviation();
	fputs("Related: halftone\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("printer          * prints printer name\n",ttyout);
	fputs("printer p        * sets printer name to p\n",ttyout);
}

void
cmd_printer(void)
{
	if(argc < 1){
		printer_syntax();
		return;
	}
	if(argc == 1){
		fprintf(ttyout,"%s\n",half_printer);
	} else{
		strcpy(half_printer,argv[1]);
	}
}

void
cmd_halftone(void)
{
	if(argc < 0){
		halftone_syntax();
		return;
	}

	status = aspect_s = width_s = height_s = 0;

	if(-1 == parse_general(options,flags,ttyout,argc,argv)) return;

	if(aspect_s){
		if(!width_s && !height_s){
			fputs("Need to set width or height with -a\n",ttyout);
			return;
		} else if(aspect_s && width_s && height_s){
			fputs("Can't set width and height with -a\n",ttyout);
			return;
		} else if(width_s){
			half_height = half_width* canvas_size_y / canvas_size_x;
		} else{
			half_width = half_height* canvas_size_x / canvas_size_y;
		}
	}

	if(half_xo <= 0.0 || half_yo <= 0.0 || half_width <= 0.0 ||
	    half_height <= 0.0){
		fputs("size/position specification error\n",ttyout);
		return;
	}

	if(fabs(half_height/half_width - ((float) canvas_size_y) / 
	    canvas_size_x) > .1){
		fputs("warning: aspect ratio mismatch\n",ttyout);
	}

	fprintf(ttyout,"printer: %s orientation: %s gamma: %3.1f\n",
	    half_printer, half_orientation? "landscape" : "portrait",
	    gamma_halftone);
	fprintf(ttyout,
	    "width: %3.1f height: %3.1f x-origin: %3.1f y-origin: %3.1f\n",
	    half_width,half_height,half_xo,half_yo);
	fprintf(ttyout,"file: %s location: %s type: %s\n",
		ofn,
		oplace? "file" : "printer",
		otype? "EPS" : "PS");

	if(status) return;
	gamma_set_device(gamma_halftone);
	lpr_send();
	gamma_restore();
}

static void
halftone_syntax(void)
{
	fputs("Synopsis: prints images or generate postscript\n",ttyout);
	min_abbreviation();
	fputs("Related: write\n",ttyout);
	fputs("Syntax: halftone [options]\n",ttyout);
	fprintf(ttyout,"xo=[%f]: x-coordinate lower left (inches)\n",half_xo);
	fprintf(ttyout,"yo=[%f]: y-coordinate lower left (inches)\n",half_yo);
	fprintf(ttyout,"width=[%f]: width (inches)\n",half_width);
	fprintf(ttyout,"height=[%f]: height (inches)\n",half_height);
	fprintf(ttyout,"gamma=[%f]: gamma for printer\n",gamma_halftone);
	fprintf(ttyout,"of=[%s][.eps|.ps]: file\n",ofn);
	fprintf(ttyout,"printer=[%s]: printer\n",half_printer);
	fputs("-a: set other dimension using proper aspect ratio\n",ttyout);
	fputs("-e: EPS to file\n",ttyout);
	fputs("-f: put output to file\n",ttyout);
	fputs("-l: landscape orientation\n",ttyout);
	fputs("-n: PS output to printer\n",ttyout);
	fputs("-p: portrait orientation\n",ttyout);
	fputs("-s: print status of command\n",ttyout);
}
