/*
op - perform operations on display
*/

#ident "$Id: op.c,v 1.6 2007/05/22 13:22:45 ccrawford Exp $"

#include "xpic.h"

static void
transpose_syntax(void)
{
	fputs("Synopsis: transposes images\n",ttyout);
	min_abbreviation();
	fputs("Related: copyquad, flip\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("transpose          * transposes image\n",ttyout);
}

void
cmd_transpose(void)
{
	int i,j;
	CPIXEL tmp,*fl,*ll;

	if(argc != 1){
		transpose_syntax();
		return;
	}

	if(da_yw != da_xw){
		fputs("Subcanvas must be square\n",ttyout);
		return;
	}
	for(i=0;i<da_yw-1;i++){
		fl = (i+da_y)*canvas_size_x + da_x+i+1 + image16;
		ll = (i+1+da_y)*canvas_size_x + da_x+i + image16;
		for(j=i+1;j<da_xw;j++){
			tmp = *ll;
			*ll = *fl;
			*fl++ = tmp;
			ll += canvas_size_x;
		}
	}
	load_pix(1);
}

static void
flip_syntax(void)
{
	fputs("Synopsis: flip images about horizontal/vertical axes\n",ttyout);
	min_abbreviation();
	fputs("Related: copyquad, transpose\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("flip          * flip about horizontal axis\n",ttyout);
	fputs("flip x        * flip about horizontal axis\n",ttyout);
	fputs("flip y        * flip about vertical axis\n",ttyout);
}

void
cmd_flip(void)
{
	int axis=0;
	int i,j;
	CPIXEL tmp,*fl,*ll,*flp,*llp;

	if(argc < 1 || argc > 2){
		flip_syntax();
		return;
	}

	if(argc == 2){
		if(!strcmp(argv[1],"x")) axis = 0;
		else if(!strcmp(argv[1],"y")) axis = 1;
		else{
			fputs("Illegal flip axis\n",ttyout);
		}
	}else{
		axis = 0; 
	}

	if(!axis){
		if(da_yw <= 1) return;
		fl = da_y * canvas_size_x + da_x + image16;
		ll = fl + (da_yw-1) * canvas_size_x;
		for(i=0;i<da_yw/2;i++){
			for(flp=fl,llp=ll,j=0;j<da_xw;j++){
				tmp = *flp;
				*flp++ = *llp;
				*llp++ = tmp;
			}
			fl += canvas_size_x;
			ll -= canvas_size_x;
		}
	}else{
		if(da_xw <= 1) return;
		fl = da_y * canvas_size_x + da_x + image16;
		ll = fl + da_xw - 1;
		for(i=0;i<da_xw/2;i++){
			for(flp=fl,llp=ll,j=0;j<da_yw;j++){
				tmp = *flp;
				*flp = *llp;
				*llp = tmp;
				flp += canvas_size_x;
				llp += canvas_size_x;
			}
			fl += 1;
			ll -= 1;
		}
	}
	load_pix(1);
}

static void
do_scale(double s,int c)
{
	int tmp;
	int i,j;
	CPIXEL *fl,*flp;

	fl = da_y * canvas_size_x + da_x + image16;
	for(i=0;i<da_yw;i++){
		for(flp=fl,j=0;j<da_xw;j++){
			tmp = s * ((int) *flp - c) + c;
			if(tmp < 0) tmp = 0;
			if(tmp > mi_max) tmp = mi_max;
			*flp++ = tmp;
		}
		fl += canvas_size_x;
	}
	load_pix(1);
}

static void
scale_syntax(void)
{
	fputs("Synopsis: scales images about center value\n",ttyout);
	min_abbreviation();
	fputs("Related: inverse, offset, zero, reverse\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("scale s          * scales images by s about 0\n",ttyout);
	fputs("scale s cen      * scales images by s about center\n",ttyout);
}


void
cmd_scale(void)
{
	int c;
	float s;

	if(argc != 2 && argc != 3){
		scale_syntax();
		return;
	}

	s = parse_atof(argv[1]);
	if(parse_atof_error()) return;

	if(argc == 3){ 
		c = parse_atof(argv[2]) + zero_offset;
		if(parse_atof_error()) return;
	}else{
		c = 0;
	}

	do_scale(s,c);
}

static void
invert_syntax(void)
{
	fputs("Synopsis: inverts gray scale about level\n",ttyout);
	min_abbreviation();
	fputs("Related: scale, offset, reverse, level, wl\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("invert        * multiplies pixels by -1 centered at level\n",ttyout);
}

void
cmd_invert(void)
{
	if(argc != 1){
		invert_syntax();
		return;
	}
	do_scale(-1.0,level+zero_offset);
}

static void
offset_syntax(void)
{
	fputs("Synopsis: adds offset to display image\n",ttyout);
	min_abbreviation();
	fputs("Related: scale, inverse, difference\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("offset o          * adds o to pixels \n",ttyout);
}

void
cmd_offset(void)
{
	int o,tmp;
	int i,j;
	CPIXEL *fl,*flp;

	if(argc != 2){
		offset_syntax();
		return;
	}
	o = parse_atof(argv[1]);
	if(parse_atof_error()) return;

	fl = da_y * canvas_size_x + da_x + image16;
	for(i=0;i<da_yw;i++){
		for(flp=fl,j=0;j<da_xw;j++){
			tmp = *flp + o;
			if(tmp < 0) tmp = 0;
			if(tmp > mi_max) tmp = mi_max;
			*flp++ = tmp;
		}
		fl += canvas_size_x;
	}
	load_pix(1);
}
