/*
port - subroutines to connect to xpic's port
*/

#ident "$Id: port.c,v 1.6 2007/05/22 13:22:45 ccrawford Exp $"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <sys/utsname.h>
#include "xpic_port.h"

#ifndef MAXHOSTNAME
#define MAXHOSTNAME 64
#endif

typedef unsigned short CPIXEL;	/* intermediate storage (16-bits) */

static int sock_fd;
static int cstat;		/* communication status, 0=closed, 1=open */
static int xpic_local;		/* 1=xpic on local host */
#ifdef MITSHM
static CPIXEL *image16;		/* canvas for shared memory access */
#endif
static int canvas_attached;	/* 1=canvas attached */
static int canvas_id;		/* id for shared memory canvas */

static void
check_open(void)
{
	if(cstat) return;
	fprintf(stderr,"attempt to communicate before call to xpic_open\n");
	exit(1);
}

static void
xpic_put(void *p,int n)
{
	if(write(sock_fd,p,n) != n){
		perror("xpic_put write error");
		exit(1);
	}
}

static void
xpic_get(p,n)
void *p;
int n;
{
	if(read(sock_fd,p,n) != n){
		perror("xpic_get read error");
		exit(1);
	}
}

static void
detach_cond(void)
{
#ifdef MITSHM
	if(canvas_attached == 1){
		shmdt((char *)image16);
	}
#endif
}

static void
close_shop()
{
	shutdown(sock_fd,0);
	close(sock_fd);
	cstat = 0;
}

static void 
broke(int dummy)
{
	fputs("pipe to xpic broke\n",stderr);
	exit(1);
}

int
xpic_open(char *host,int err)
{
	struct sockaddr_in sa;
	struct hostent *hp;
	int sock;
	char *c,local_host[MAXHOSTNAME+1];
	char xpic_host[MAXHOSTNAME+1];
	char xpic_host_full[MAXHOSTNAME+1];
	int port=PORT_SOCKET;
	static struct utsname buf;

	if(cstat == 1){
		fprintf(stderr,"xpic_open: connection already open\n");
		exit(1);
	}
	canvas_attached = cstat = 0;

	if(host == NULL){
		xpic_host[0] = 0;
	}else{
		strcpy(xpic_host,host);
		if(NULL != (c = strchr(xpic_host,':'))){
			*c = 0;
			port = atoi(c+1);
		}
	}

	if(!strlen(xpic_host)){
		uname(&buf);
		strncpy(xpic_host,&buf.nodename[0],MAXHOSTNAME);
	}

	if(NULL == (hp = gethostbyname(xpic_host)))goto out;
	strcpy(xpic_host_full,hp->h_name);

	memset(&sa,0,sizeof(sa));
	memcpy( (char *)&sa.sin_addr, hp->h_addr, hp->h_length);
	sa.sin_family= hp->h_addrtype;
	sa.sin_port= htons((u_short)port);

	if ((sock= socket(hp->h_addrtype,SOCK_STREAM,0)) < 0) goto out;
	if (connect(sock,(struct sockaddr *)&sa,sizeof sa) < 0) goto out;
	signal(SIGPIPE,broke); 
	sock_fd = sock;
	canvas_attached = 0;
	cstat = 1;

	strncpy(local_host,&buf.nodename[0],MAXHOSTNAME);
	if(NULL == (hp = gethostbyname(local_host)))goto out;
	strcpy(local_host,hp->h_name);
	xpic_local = !strcmp(local_host,xpic_host_full);

	return(0);

out:
	if(err) return(1);
	fprintf(stderr,"can't connect to xpic on %s via port %d\n",
		xpic_host,port);
	exit(1);
}

int
xpic_test(void)
{
	int stat,i;
	TEST t;

	check_open();
	t.size = sizeof(t);
	t.code = P_TEST;
	for(i=0;i<P_TEST_WORDS;i++) t.buf[i] = i;
	xpic_put((char *)&t,sizeof(t));
	xpic_get(&stat,sizeof(stat));
	return(stat);
}

void
xpic_close(void)
{
	CLOSE t;
	if(cstat == 0){
		fprintf(stderr,"xpic_close: no connection to close\n");
		exit(1);
	}

	check_open();
	detach_cond();
	t.size = sizeof(t);
	t.code = P_CLOSE;
	xpic_put((char *)&t,sizeof(t));

	close_shop();
}

void
xpic_message(char *s)
{
	MESSAGE m;

	check_open();
	m.size = sizeof(m);
	m.code = P_MESSAGE;
	strncpy(m.message,s,MAX_MESSAGE);
	m.message[MAX_MESSAGE] = 0;
	xpic_put((char *)&m,sizeof(m));
}

void
xpic_annotate(int x,int y,char *s)
{
	ANNOTATE m;

	check_open();
	m.size = sizeof(m);
	m.code = P_ANNOTATE;
	m.x = x;
	m.y = y;
	strncpy(m.message,s,MAX_MESSAGE);
	m.message[MAX_MESSAGE] = 0;
	xpic_put((char *)&m,sizeof(m));
}

void
xpic_status(int *x,int *y,int *w,int *l,int *mi_max)
{
	STATUS_INFO m;
	STATUS s;

	check_open();
	s.size = sizeof(s);
	s.code = P_STATUS;
	xpic_put((char *)&s,sizeof(s));

	xpic_get(&m,sizeof(m));

	*x = m.x;
	*y = m.y;
	*w = m.window;
	*l = m.level;
	*mi_max = m.mi_max;
	canvas_id = m.canvas_id;
}

void
xpic_wl(int w,int l)
{
	WL m;

	check_open();
	m.size = sizeof(m);
	m.code = P_WL;
	m.window = w;
	m.level = l;
	xpic_put((char *)&m,sizeof(m));
}

void
xpic_load(void)
{
	LOAD m;

	check_open();
	m.size = sizeof(m);
	m.code = P_LOAD;
	xpic_put((char *)&m,sizeof(m));
}

void
xpic_debug(int d)
{
	DEBUG_PORT t;

	check_open();
	t.size = sizeof(t);
	t.code = P_DEBUG_PORT;
	t.debug = d;
	xpic_put((char *)&t,sizeof(t));
}

void
xpic_clear(void)
{
	CLEAR t;

	check_open();
	t.size = sizeof(t);
	t.code = P_CLEAR;
	xpic_put((char *)&t,sizeof(t));
}

void
xpic_write(int x,int y,unsigned short *data,int n)
{
	WRITE t;
	int i;

	check_open();

	if(n <= 0)return;
	if(n > XP_LINE_MAX)n = XP_LINE_MAX;

	t.size = sizeof(t) - (XP_LINE_MAX-n)*sizeof(short);
	t.code = P_WRITE;
	t.x = x;
	t.y = y;
	t.n = n;
	for(i=0;i<n; t.data[i++] = *data++);
	xpic_put((char *)&t,t.size);
}

int
xpic_read(int x,int y,unsigned short *data,int n)
{
	READ t;
	READ_INFO rd;
	int i;

	check_open();

	if(n <= 0)return(0);
	if(n > XP_LINE_MAX)n = XP_LINE_MAX;

	t.size = sizeof(t);
	t.code = P_READ;
	t.x = x;
	t.y = y;
	t.n = n;
	xpic_put((char *)&t,t.size);
	xpic_get(&rd,sizeof(rd)-(XP_LINE_MAX-n)*sizeof(short));
	for(i=0;i<rd.n;*data++ = rd.data[i++]);
	return(rd.n);
}

void
xpic_transpose(int d)
{
	TRANSPOSE t;

	check_open();
	t.size = sizeof(t);
	t.code = P_TRANSPOSE;
	t.transpose = d;
	xpic_put((char *)&t,sizeof(t));
}

void
xpic_cursor(int *type,int *x,int *y,int *xw,int *yw)
{
	CURSOR_INFO m;

	CURSOR_I s;

	check_open();
	s.size = sizeof(s);
	s.code = P_CURSOR;
	xpic_put((char *)&s,sizeof(s));

	xpic_get(&m,sizeof(m));

	*type = m.type;
	*x = m.x;
	*y = m.y;
	*xw = m.xw;
	*yw = m.yw;
}

void
xpic_command(char *s)
{
	COMMAND m;

	check_open();
	m.size = sizeof(m);
	m.code = P_COMMAND;
	strncpy(m.cmd,s,MAX_MESSAGE);
	m.cmd[MAX_MESSAGE] = 0;
	xpic_put((char *)&m,sizeof(m));
}

void
xpic_move(int x,int y)
{
	MOVE m;

	check_open();
	m.size = sizeof(m);
	m.code = P_MOVE;
	m.x = x;
	m.y = y;
	xpic_put((char *)&m,sizeof(m));
}

void
xpic_line(int x,int y)
{
	LINE l;

	check_open();
	l.size = sizeof(l);
	l.code = P_LINE;
	l.x = x;
	l.y = y;
	xpic_put((char *)&l,sizeof(l));
}

void
xpic_exit(void)
{
	EXIT m;

	check_open();
	m.size = sizeof(m);
	m.code = P_EXIT;
	detach_cond();
	xpic_put((char *)&m,sizeof(m));
	close_shop();
}

	/* button (b): 0=select 1=adjust 2=menu (usually left, middle, right) */
int
xpic_mouse(int b)
{
	MOUSE m;
	int stat;

	if(b < 0 || b > 3){
		fputs("xpic_mouse: button out of range\n",stderr);
		exit(1);
	}
	check_open();
	m.size = sizeof(m);
	m.code = P_MOUSE;
	m.b = b;
	xpic_put((char *)&m,sizeof(m));
	xpic_get(&stat,sizeof(stat));
	return(stat);
}

CPIXEL *
xpic_attach(void)
{
#ifdef MITSHM
	int x,y,w,l,mi_max;
#endif

	check_open();

#ifdef MITSHM
	if(!xpic_local){
		fputs("xpic_attach: xpic not running locally\n",stderr);
		exit(1);
	}

	xpic_status(&x,&y,&w,&l,&mi_max);

	if(canvas_id == -1){
		fputs("xpic_attach: xpic not using shared memory\n",stderr);
		exit(1);
	}

	image16 = (CPIXEL *)shmat(canvas_id, 0, 0 );
	if((int) image16 == -1){
		perror("xpic_attach: can't attach canvas memory");
		exit(1);
	}
	canvas_attached = 1;
	return(image16);
#else
	fputs("xpic_attach: shared memory not available\n",stderr);
	exit(1);
#endif
}

void 
xpic_detach(void)
{
	check_open();

	if(canvas_attached == 0){
		fputs("xpic_detach: no canvas to detach\n",stderr);
		exit(1);
	}
	canvas_attached = 0;
	detach_cond();
}
