/*
smpte - insert smpte-like patttern into display
*/

#ident "$Id: smpte.c,v 1.3 2007/05/22 13:22:45 ccrawford Exp $"

#include "xpic.h"

#define NX	512
#define NY	512

typedef struct	pair	{
	int	x;
	int	y;
} pair;

struct	drawing_parameters {
	unsigned short *canvas_origin;
	unsigned short *image_origin;
	pair	image_extent;
	pair	canvas_extent;
	pair	min_draw;
	pair	max_draw;
};

static	struct	drawing_parameters params;

static	void	blit(unsigned short *p, 
		int x, int y, int width, int height, 
		unsigned short value) {

	unsigned short *row_p;
	int	i,j;
	int	x0, y0;
	int	y1, x1;

	y0 = y;
	y1 = y0 + height-1;
	if(y0 < params.min_draw.y) {
		y0 = params.min_draw.y; }
	if(y1 > params.max_draw.y) {
		y1 = params.max_draw.y; }

	x0 = x;
	x1 = x0 + width-1;
	if(x0 < params.min_draw.x) {
		x0 = params.min_draw.x; }
	if(x1 > params.max_draw.x) {
		x1 = params.max_draw.x; }

	for(j=y0;j<=y1;j++) {
		row_p = params.image_origin + params.canvas_extent.x * j;
		for(i=x0;i<=x1;i++) {
			row_p[i] = value;
		}
	}
}

static	void	hcomb(unsigned short	*p,
		int x0, int y0, int width, int height, 
		int	stripe, int direction,
		unsigned short on_value,
		unsigned short off_value) {
	int	j,k;

	blit(p, x0, y0, width, height, off_value);
	if(direction == 1) {
		for(j=y0;j<(y0+height);j+=(2*stripe)) {
			k = stripe;
			if((k+j) > (y0+height)) k = y0 + height - j;
			blit(p, x0, j, width, k, on_value);
		}
	} else {
		for(j=y0+height-stripe;(j+stripe)>=y0;j-=(2*stripe)) {
			k = j;
			if(k<y0)k=y0;
			blit(p, x0, k, width, stripe-(k-j), on_value);
		}
	}
}

static	void	vcomb(unsigned short	*p,
		int x0, int y0, int width, int height, 
		int	stripe, int direction,
		unsigned short on_value,
		unsigned short off_value) {
	int	i,k;

	blit(p, x0, y0, width, height, off_value);
	if(direction == 1) {
		for(i=x0;i<(x0+width);i+=(2*stripe)) {
			k = stripe;
			if((k+i) > (x0+width)) k = x0 + width - i;
			blit(p, i, y0, k, height, on_value);
		}
	} else {
		for(i=x0+width-stripe;(i+stripe)>=x0;i-=(2*stripe)) {
			k = i;
			if(k<x0)k=x0;
			blit(p, k, y0, stripe-(k-i), height, on_value);
		}
	}
}

static	void	res_comb(unsigned short	*p,
		int x0, int y0, int	dx, int dy,
		unsigned short on, unsigned short off)
{
	int	left, right, up, down;

	if(dx < 0) {
		left = x0 - 32 + 1;
		right = x0;
	} else {
		left = x0;
		right = x0 + 32 - 1;
	}
	if(dy < 0) {
		up = y0 - 32 + 1;
		down = y0;
	} else {
		up = y0;
		down = y0 + 32 + 1;
	}
		
	vcomb(p, left, up, 32, 32, 1, dx, on, off);
	left += dx * 32;
	right += dx * 32;
	hcomb(p, left, up, 32, 32, 1, dy, on, off);
	up += dy * 32;
	down = dy * 32;
	hcomb(p, left, up, 32, 32, 2, dy, on, off);
	left -= dx * 32;
	right -= dx * 32;
	vcomb(p, left, up, 32, 32, 2, dx, on, off);
	up += dy * 32;
	down = dy * 32;
	vcomb(p, left, up, 32, 32, 3, dx, on, off);
	left += dx * 32;
	right += dx * 32;
	hcomb(p, left, up, 32, 32, 3, dy, on, off);
}

static	void	gray_comb(unsigned short	*p,
		int x0, int y0, int	dx, int dy,
		unsigned short off, unsigned short on)
{
	int	left, right, up, down;
	unsigned short mid;
	short	delta;

	mid = (on + off)/2;
	delta = (on - off)/100;


	if(dx < 0) {
		left = x0 - 32 + 1;
		right = x0;
	} else {
		left = x0;
		right = x0 + 32 - 1;
	}
	if(dy < 0) {
		up = y0 - 32 + 1;
		down = y0;
	} else {
		up = y0;
		down = y0 + 32 + 1;
	}
		
	hcomb(p, left, up, 32, 32, 2, dy, mid, mid+delta);
	left += dx * 32;
	right += dx * 32;
	vcomb(p, left, up, 32, 32, 2, dx, mid+delta, mid);
	up += dy * 32;
	down = dy * 32;
	vcomb(p, left, up, 32, 32, 2, dx, mid+delta, mid-2*delta);
	left -= dx * 32;
	right -= dx * 32;
	hcomb(p, left, up, 32, 32, 2, dy, mid-2*delta, mid+delta);
	up += dy * 32;
	down = dy * 32;
	hcomb(p, left, up, 32, 32, 2, dy, mid-2*delta, mid+3*delta);
	left += dx * 32;
	right += dx * 32;
	vcomb(p, left, up, 32, 32, 2, dx, mid+3*delta, mid-2*delta);
}

static	void	grid(unsigned short *p,
		int	*x_array, int *y_array, int n_xvals, int n_yvals,
		unsigned short value)
{
	int	i;


	for(i=0;i<n_xvals;i++) {
		blit(p, x_array[i], 0, 1, NY, value);
	}

	for(i=0;i<n_yvals;i++) {
		blit(p, 0, y_array[i], NX, 1, value);
	}
}

static	void	frame( unsigned short *p,
	int	margin, int thick, unsigned short value)
{
	blit(p, margin, margin, thick, NY - 2 * margin, value);
	blit(p, NX - margin - thick, margin, thick, NY - 2 * margin, value);

	blit(p, margin, margin, NX - 2 * margin, thick, value);
	blit(p, margin, NX - margin - thick, NY - 2 * margin, thick, value);
}

static
int	simple_t(
	unsigned short *data_p, 
	int	n_columns, 
	int	n_rows
	)
{
	int	i,k;
	unsigned short floor;
	unsigned short ceiling ;

	int	x_nodes[11], y_nodes[11];

	floor = 974;
	ceiling = 1074;

	params.canvas_origin = data_p;
	params.image_origin = data_p;
	params.min_draw.x = 0;
	params.min_draw.y = 0;
	params.max_draw.x = n_columns-1;
	params.max_draw.y = n_rows - 1;

	params.canvas_extent.x = n_columns;
	params.canvas_extent.y = n_rows;
	params.image_extent = params.canvas_extent;	

	/*
	 *	Blit out any canvas not covered by the image.
	 *		(Just do the whole canvas. Paint the image later.)
	 */
	blit(data_p, 0, 0, params.image_extent.x, params.image_extent.y, floor);

	/*
	 *	Adjust drawing params to account for real image in canvas.
	 */
	params.image_extent.x = NX;
	params.image_extent.y = NY;

	if(params.image_extent.x < params.canvas_extent.x) {
		params.image_origin += (params.canvas_extent.x - params.image_extent.x)/2;
	} else {
		i = ( params.image_extent.x - params.canvas_extent.x)/2;
		params.image_origin -= i;
		params.min_draw.x = i;
		params.max_draw.x = params.min_draw.x + params.canvas_extent.x - 1;
	}

	if(params.image_extent.y < params.canvas_extent.y) {
		params.image_origin += ((params.canvas_extent.y - params.image_extent.y)/2)
				* params.canvas_extent.x;
	} else {
		i = ( params.image_extent.y - params.canvas_extent.y)/2;
		k = i*params.canvas_extent.x;
		params.image_origin -= k;
		params.min_draw.y = i;
		params.max_draw.y = params.min_draw.y + params.canvas_extent.y - 1;
	}

		
	/*
	 *	Now blit out the image to half scale.
	 */
	blit(data_p, 0, 0, params.image_extent.x, 
		params.image_extent.y, (floor+ceiling)/2);

	
	for(i=0;i<=10;i++) {
		x_nodes[i] = i * (NX-1)/10 ;
		y_nodes[i] = i * (NY-1)/10 ;
	}
	
	/*
	 *	Set up the grid at 75%
	 */
	grid(data_p, x_nodes, y_nodes, 11, 11, (floor + 3 * ceiling) / 4);

	/*
	 *	Box it
	 */
	frame(data_p, 5, 3, (1 * floor + 3 * ceiling)/4);
	/*
	 *	Draw in the long contrasting rectangles
	 */
	k = y_nodes[3] - y_nodes[2];
	blit(data_p, x_nodes[1] , y_nodes[2] + k/6, x_nodes[9] - x_nodes[1] + 2,
			y_nodes[3] - y_nodes[2] - k/3+7, .05 * floor + 0.95 * ceiling);
	blit(data_p, x_nodes[1] , y_nodes[7] + 1, x_nodes[9] - x_nodes[1] + 2,
			y_nodes[8] - y_nodes[7] - k/3+8, .95 * floor + 0.05 * ceiling);
	blit(data_p, x_nodes[3] + 1, y_nodes[2] + k/3 + 1, x_nodes[7] - x_nodes[3] + 1,
			k/3 + 4, .95 * floor + 0.05 * ceiling);
	blit(data_p, x_nodes[3] + 1, y_nodes[7] + k/6+4, x_nodes[7] - x_nodes[3] + 1,
			k/3 + 4, .05 * floor + 0.95 * ceiling);

	/*
	 *	Now do the arch of 10% gradient blocks
	 */
	blit(data_p, x_nodes[2]+1, y_nodes[6]+1, 
		x_nodes[3] - x_nodes[2]-1,
		y_nodes[7] - y_nodes[6] - 1,
		(10 * floor + 0 * ceiling)/10);
	blit(data_p, x_nodes[2]+1, y_nodes[5]+1, 
		x_nodes[3] - x_nodes[2]-1,
		y_nodes[6] - y_nodes[5] - 1,
		(9 * floor + 1 * ceiling)/10);
	blit(data_p, x_nodes[2]+1, y_nodes[4]+1, 
		x_nodes[3] - x_nodes[2]-1,
		y_nodes[5] - y_nodes[4] - 1,
		(8 * floor + 2 * ceiling)/10);
	blit(data_p, x_nodes[2]+1, y_nodes[3]+1, 
		x_nodes[3] - x_nodes[2]-1,
		y_nodes[4] - y_nodes[3] - 1,
		(7 * floor + 3 * ceiling)/10);
	blit(data_p, x_nodes[3]+1, y_nodes[3]+1, 
		x_nodes[4] - x_nodes[3]-1,
		y_nodes[4] - y_nodes[3] - 1,
		(6 * floor + 4 * ceiling)/10);

	blit(data_p, x_nodes[7]+1, y_nodes[6]+1, 
		x_nodes[8] - x_nodes[7]-1,
		y_nodes[7] - y_nodes[6] - 1,
		(0 * floor + 10 * ceiling)/10);
	blit(data_p, x_nodes[7]+1, y_nodes[5]+1, 
		x_nodes[8] - x_nodes[7]-1,
		y_nodes[6] - y_nodes[5] - 1,
		(1 * floor + 9 * ceiling)/10);
	blit(data_p, x_nodes[7]+1, y_nodes[4]+1, 
		x_nodes[8] - x_nodes[7]-1,
		y_nodes[5] - y_nodes[4] - 1,
		(2 * floor + 8 * ceiling)/10);
	blit(data_p, x_nodes[7]+1, y_nodes[3]+1, 
		x_nodes[8] - x_nodes[7]-1,
		y_nodes[4] - y_nodes[3] - 1,
		(3 * floor + 7 * ceiling)/10);
	blit(data_p, x_nodes[6]+1, y_nodes[3]+1, 
		x_nodes[7] - x_nodes[6]-1,
		y_nodes[4] - y_nodes[3] - 1,
		(4 * floor + 6 * ceiling)/10);


	/*
	 *	Here are the 5% contrast blocks
	 */
	blit(data_p, x_nodes[3]+1, y_nodes[6]+1, 
		x_nodes[4] - x_nodes[3]-1,
		y_nodes[7] - y_nodes[6] - 1,
		(10 * floor + 0 * ceiling)/10);
	blit(data_p, x_nodes[6]+1, y_nodes[6]+1, 
		x_nodes[7] - x_nodes[6]-1,
		y_nodes[7] - y_nodes[6] - 1,
		(0 * floor + 10 * ceiling)/10);
	k = (x_nodes[4] - x_nodes[3])/4;
	blit(data_p, x_nodes[3]+k+2, y_nodes[6]+k+2, 
		x_nodes[4] - x_nodes[3]-1-2*k,
		y_nodes[7] - y_nodes[6] - 1-2*k,
		(19 * floor + 1 * ceiling)/20);
	blit(data_p, x_nodes[6]+k+2, y_nodes[6]+k+2, 
		x_nodes[7] - x_nodes[6]-1-2*k,
		y_nodes[7] - y_nodes[6] - 1-2*k,
		(1 * floor + 19 * ceiling)/20);

	/*
	 *	Now put in the combs, center and corners.
	 */
	res_comb(data_p, x_nodes[5]+1, y_nodes[5] - 47, 1, 1, floor, ceiling);
	res_comb(data_p, (x_nodes[1] + x_nodes[2])/2+1, y_nodes[ 0] + 12, -1, +1, floor, ceiling);
	res_comb(data_p, (x_nodes[8] + x_nodes[9])/2+1, y_nodes[ 0] + 12, +1, +1, floor, ceiling);
	res_comb(data_p, (x_nodes[1] + x_nodes[2])/2+1, y_nodes[10] - 13, -1, -1, floor, ceiling);
	res_comb(data_p, (x_nodes[8] + x_nodes[9])/2+1, y_nodes[10] - 13, +1, -1, floor, ceiling);
	gray_comb(data_p, x_nodes[5]-0, y_nodes[5] - 47, -1, 1, floor, ceiling);
	gray_comb(data_p, (x_nodes[1] + x_nodes[2])/2+2, y_nodes[ 0] + 12, +1, +1, floor, ceiling);
	gray_comb(data_p, (x_nodes[8] + x_nodes[9])/2-0, y_nodes[ 0] + 12, -1, +1, floor, ceiling);
	gray_comb(data_p, (x_nodes[1] + x_nodes[2])/2+2, y_nodes[10] - 13, +1, -1, floor, ceiling);
	gray_comb(data_p, (x_nodes[8] + x_nodes[9])/2-0, y_nodes[10] - 13, -1, -1, floor, ceiling);

	return(0);
}

static void
smpte_syntax(void)
{
	fputs("Synopsis: insert smpte-like patttern into display\n",ttyout);
	min_abbreviation();
	fputs("Related: gamma\n",ttyout);
	fputs("Syntax:\n",ttyout);
	fputs("smpte       * insert pattern\n",ttyout);
}

void
cmd_smpte(void)
{
	if(argc != 1){
		smpte_syntax();
		return;
	}

	simple_t(image16,canvas_size_x,canvas_size_y);
	do_awl();
	load_pix(0);
}
