/*
xpic.h - include file for xpic
*/

static const char rcsid_xpic_h[] = "$Id: xpic.h,v 1.44 2007/05/17 09:58:09 ccrawford Exp $";

#ifdef GZMI
#include<zlib.h>
#endif
/*
	configuration defines
*/
#ifndef NOMITSHM		/* change sense to simplify makefiles */
#ifndef MITSHM
#define MITSHM
#endif
#endif

#define MERLIN			// set if MERLIN mode for LM-files = L3F files

/*
	general constants
*/
#define LUT_SIZE_P 512			/* size of internal lut before gamma */
#define MAX_CMD_WORDS 20		/* max words on prompt line */
#define MIMAGES	64			/* max images on prompt line */
#define CTEXT_LEN 20			/* length of for cursor display */
#define PACKET_LENGTH 2100		/* for port communication */
#define KB_BUF 256			/* for kb - xpic communication */
#define RING_LENGTH 2048		/* for port communication */
#define BUFL 128			/* length intern buffers for strings */
#define START_UP_FILE ".xpic"		/* file with initial arguments */
#define OSL 128				/* length of option strings */
#define MAIN_KB "XPIC_KB"		/* token between main and kb */
#define ENV_N 10			/* number of env's */
#define SPECIALMODE "11jo3rb99v2b5"	/* special mode key for dir transfer */
#define OUTLINE_LENGTH 15		// number of outline pixels
/*
	macros
*/
#define CRC_MAX(x,y) ( ((x) > (y))? (x) : (y) )
#define CRC_MIN(x,y) ( ((x) < (y))? (x) : (y) )
#define MAX3(x,y,z) (CRC_MAX(CRC_MAX(x,y),z))

#ifdef INIT_GLOBALS
#define EXTERN
#define INIT(x) =x
#else
#define EXTERN extern
#define INIT(X)
#endif

#ifdef DEBUG
#define BUG(x)	parse_uwtext(ttyout,x);
#else
#define BUG(x)
#endif
/*
	include files
*/
#include	<stdio.h>
#include	<unistd.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<math.h>
#include 	<string.h>
#include 	<errno.h>

#include	<sys/stat.h>
#include 	<sys/param.h>

#include 	"parse.h"
/*
	defines
*/
#define I_MI 0			/* image type for mi-files */
#define I_LM 1			/* image type for lm */
#ifdef GZMI
#define I_GZMI 2                /* image type for gzmi-files */
#endif

#define I_LM_HIGH 0		/* image sub-type lm-high axial */
#define I_LM_LOW 1		/* image sub-type lm-low axial */
#define I_LM_PROJ 2		/* image sub-type lm-projections */

#define EOC 1			/* flag sent by keyboard handler */

#define LINE_HARD 0		/* draw hard line */
#define LINE_SOFT 1		/* draw soft line */
#define LINE_ERASE 2		/* erase soft line */

#ifndef MAXHOSTNAME
#define MAXHOSTNAME 64
#endif
/*
	typedefs
*/
typedef unsigned char DPIXEL8;		/* pixel for display (8-bits) */
typedef unsigned int DPIXEL32;		/* pixel for display (32-bits) */
typedef unsigned short CPIXEL;		/* canvas storage (16-bits) */
typedef void (*PROC)(void);		/* for command table */
/*
	global variables w/o initialization
*/
EXTERN int cmd_num;			// command number in cmd_help
EXTERN FILE *ttyin, *ttyout;		/* FD's for tty */
EXTERN int da_x;			/* x-position start of subcanvas */
EXTERN int da_y;			/* y-position start of subcanvas */
EXTERN int da_xw;			/* x-width of subcanvas */
EXTERN int da_yw;			/* y-width of subcanvas */
EXTERN int da2_x, da2_y, da2_xw, da2_yw; /* subcanvas for load(2) */
EXTERN DPIXEL32 *cms_map;		/* mapping to cms for X */
EXTERN int lut_scale;			/* scale factor for small lut */
EXTERN unsigned int *mi_lut;		/* lut 16 bit to display */
EXTERN CPIXEL *image16;			/* pointer to canvas */
EXTERN DPIXEL8 *pr;			/* for 8 bit image */
EXTERN DPIXEL8 *pr2_data;		/* data for 8 bit image to send out*/
EXTERN int pr2_pixels;			/* pixels per line */
EXTERN int images_total;		/* number of images in file */
EXTERN int current_image;		/* current image; 0= open but not load */
EXTERN int current_nx;			/* nx of current image */
EXTERN int current_ny;			/* ny of current image */
EXTERN FILE *image_fd;			/* FD of open file */
#ifdef GZMI
EXTERN gzFile *gz_image_fd;               /* FD for gz files */
#endif
EXTERN int cen_set, cen_x, cen_y;	/* set & center of image */
EXTERN int cor_set, cor_x, cor_y;	/* set & corner of image */
EXTERN int mi_max;			/* max value in mi-image (16 bit) */
EXTERN int file_open;			/* 1=there is an open file */
EXTERN int argc; 			/* # words on prompt line */
EXTERN char *argv[MAX_CMD_WORDS];	/* pointers to words on commandline */
EXTERN int command_src;			/* origin of command 0=prompt 1=port */
EXTERN char ifn[KB_BUF];		/* file name for open_file() */
EXTERN int note;			/* 1=annotation active for cursor */
EXTERN char note_text[KB_BUF];		/* annotation */
EXTERN char host_name[MAXHOSTNAME];	/* name of host cpu */
EXTERN int annotate_loc;		/* 0=pixrect 1=pixwin */
EXTERN int cursor_in_win;		/* 1=cursor in window */
EXTERN int font_height,font_width;	/* size of characters in annot font */
EXTERN int scroll_x,scroll_y; 		/* movement for scroll command */
EXTERN int wltrack; 			/* 1=track wl w/ sliders */
EXTERN int dsp_byte_order;		/* server's byte order */
EXTERN int host_byte_order;		/* host's byte order */
EXTERN int dsp_depth;			/* display's bit depth: 8 or 24 */
EXTERN int dsp_class;			/* display's class */
EXTERN int dsp_bits_pixel;		/* display's bits per pixel */
EXTERN int dsp_bytes_pixel;		/* display's bytes per pixel */
EXTERN int cwd2kb;			/* send cwd to kb */
EXTERN int current_env;
EXTERN int lut_red;			/* location of color red in lut */
EXTERN int lut_blue;			/* location of color blue in lut */
EXTERN int lut_green;			/* location of color green in lut */
EXTERN int lut_outline;			/* location of colors for outline  in lut */
/*
	globals with initialization
*/
EXTERN int cursor_stat INIT(1);		/* 1=cursor is active */
EXTERN int diff_offset INIT(1024);	/* offset in difference mode */
EXTERN int port_access INIT(0);		/* 1=allow port access */
EXTERN int kludge INIT(0);		/* kludge for debugging */
EXTERN int canvas_size_x INIT(512);	/* width of canvas */
EXTERN int canvas_size_y INIT(512);	/* height of canvas */
EXTERN int xp INIT(0);			/* upper left corner of canvas -x */
EXTERN int yp INIT(0);			/* upper left corner of canvas -y */
EXTERN int rows INIT(8);		/* # rows of text in tty window */
EXTERN int window INIT(256);		/* window for lut */
EXTERN int level INIT(0);		/* level for lut */
EXTERN int image_bits INIT(14);		/* number of active bits in input */
EXTERN int do_fork INIT(1);		/* 1=fork to detach xpic */
EXTERN int itype INIT(I_MI);		/* image type */
EXTERN int isub INIT(0);		/* image sub-type */
EXTERN int lut_size INIT(64);		/* size of look-up-table */
EXTERN int zero_offset INIT(1024);	/*zero ct style offset */
EXTERN int autowl INIT(0);		/* auto window/level per image */
EXTERN int autoclear INIT(0);		/* auto clear per image */
EXTERN int automagnify INIT(0);		/* auto maginify per image */
EXTERN int autoreset INIT (1);		/* reset display on open */
EXTERN int port_status INIT(0);		/* 1=waiting 2=active 0=closed */
EXTERN int rvideo INIT(0);		/* 1 = reverse video */
EXTERN float gamma_halftone INIT(1.0);	/* gamma of postscript printer */
EXTERN float gamma_display INIT(1.0);	/* gamma of display itself */
EXTERN float half_width INIT(6.0);	/* width of halftone output inches */
EXTERN float half_height INIT(6.0);	/* height of halftone output inches */
EXTERN float half_xo INIT(1.0);		/* x origin of halftone output inches */
EXTERN float half_yo INIT(1.0);		/* y origin of halftone output inches */
EXTERN int half_orientation INIT(0);	/* halftone 0=portrait 1=landscape */
EXTERN char half_printer[80] INIT("lp"); /* name of lineprinter for halftone */
EXTERN int last_hex INIT(1);		/* last hexant used */
EXTERN int last_quad INIT(1);		/* last quadrant used */
EXTERN int last_nan INIT(1);		/* last nanant used */
EXTERN int opt_panel INIT(1);		/* 1 = options panel */
EXTERN int sep_ctr INIT(0);		/* 1 = separate controls w/ baseframe */
EXTERN int no_attach INIT(1);		/* 1 = don't allow xpic_attach */
EXTERN float pixel_dx INIT(1.0);	/* x pixel size */
EXTERN float pixel_dydx INIT(1.0);	/* ratio y to x pixel size */
EXTERN int mag_replicate INIT(0);	/* 1=replicate (pixel) in do_mag */
EXTERN int boxred INIT(1);		/* 1=red boxes versus XORing */
EXTERN int roiellipse INIT(0);		/* 1=roi() uses ellipse versus box */
EXTERN int roimin INIT(0);		/* exclude pixels < this value */
EXTERN int mitshm_override INIT(0);	/* 1=don't use mit shm */
EXTERN int pm_skip INIT(1);		/*  skip in +/- commands */
EXTERN int outline_start INIT(9025);	// pixel value of first outline
/*
	global structures
*/
EXTERN struct {				/* status of box cursor */
	int x;				/* x location center */
	int y;				/* y location center */
	int xw;				/* width */
	int yw;				/* height */
	int lx;				/* last x value when left button */
	int ly;				/* last y value when left button */
	int status;			/* 1=box active */
	int vis;			/* 1=box presenty on screen */
} box;

struct PROC_TAB{			/* table of commands */
	char *name;			/* name (string) for command */
	PROC func;			/* procedure to call */
}; 
#ifndef INIT_GLOBALS
extern struct PROC_TAB clist[];
extern int ncom;			/* number of comamnds in clist */
#endif
/*
	prototypes
*/
void	cmd_annotate(void),cmd_box(void),cmd_cd(void),cmd_center(void),cmd_clear(void),
	cmd_corner(void),cmd_cursor(void),cmd_difference(void),cmd_exit(void),
	cmd_file(void),cmd_flip(void),cmd_gamma(void),cmd_halftone(void),cmd_help(void),
	cmd_hexant(void),cmd_horizontal(void),cmd_invert(void),
	cmd_level(void),cmd_log(void),cmd_nanant(void),cmd_quadrant(void),
	cmd_offset(void),cmd_open(void),cmd_pause(void),cmd_port(void),cmd_pwd(void),
	cmd_printer(void),cmd_reset(void),cmd_reverse(void),cmd_roi(void),
	cmd_scale(void),cmd_source(void),cmd_status(void),
	cmd_transpose(void),cmd_vertical(void),cmd_window(void),
	cmd_expression(void),cmd_wl(void),cmd_write(void),cmd_shell(void),
	cmd_refresh(void),cmd_source(void),cmd_pause(void),cmd_reopen(void),
	cmd_copyquad(void),cmd_next(void),cmd_previous(void),
	cmd_cine(void),cmd_zero(void),cmd_set(void),cmd_magnify(void),
	cmd_scroll(void),cmd_measure(void),cmd_header(void),cmd_env(void),
	cmd_images(void),cmd_panel(void),cmd_smpte(void),cmd_mouse(void),
	cmd_lc(void),cmd_browse(void);

void annotate_add(int x,int y,char *s);
void annotate_clear(void);
int annotate_exist(void);
void annotate_refresh(void);
void annotate_write(void);
int check_source(char *s);
void command(char *s,int src);
void cur_on_off(void);
void display_attach(int *gargc,char **gargv);
int display_canvas_id(void);
void display_clean_up(void);
void display_destroy(void);
void display_get_image(void);
void display_init(char *prog_name,int init,int pcolor,char *title);
void display_release_image(void);
void display_run(void);
void display_set_sliders(void);
void display_status(void);
int display_tty_width(void);
void display_wait3(int me,int pid);
void do_clear(int load);
void gamma_insert(char *host,double gamma);
void gamma_match(char *host);
void gamma_restore(void);
void gamma_set_device(double gamma_device);
void handle_port(void);
void images(char *s);
void kb(int c_argc,char **c_argv);
void line_add(int x,int y,int type);
int line_exist(void);
void line_refresh(void);
void line_table_clear(void);
void load16(int image_num,int diff_mode);
void load_pix(int area);
int lookup(char *s,int e);
void lut_init(void);
void lut_init(void);
void open_file(void);
int port_button_proc(int b);
int port_get_port(void);
void port_set_port(int p);
void pperror(char *s);
void quit_guts(void);
void reset(void);
void ring_get(char *c,int n);
void ring_init(void);
int ring_length(void);
void ring_put(char *c,int n);
void set_gamma_tab(double value,int map);
void window_level_lut(int window,int level);
void words(char *buf,int *w_argc,char **w_argv,int wmax);
int box_exist(void);
void draw_box(int gc);
void box_refresh(void);
void display_text(int x,int y,char *s);
void box_table_clear(void);
void display_tty_out(char *line);
void display_drawline(int x1,int y1,int x2,int y2,int type);
void draw_cursor(int x,int y);
void display_set_mouse(int x,int y);
void box_add(int x,int y,int xw,int yw,int type);
void ctext(char *s);
void refresh(int left,int top,int width,int height);
void display_clear_annotate_clip(void);
void display_set_annotate_clip(int left,int top,int width,int height);
void display_copy_area(int rx,int ry,int rdx,int rdy);
void display_put_image(int rx,int ry,int rdx,int rdy);
void display_text_raw(int x,int y,char *s);
void display_cross_on(void);
void display_cross_off(void);
void display_panel_message(char *s1);
void display_get_mouse(int *x,int *y);
void roi(int mode);
void deposit_cursor(void);
char *rev_date(void);
void next_prev(int mode);
void wl_status(void);
void cine_syntax(void);
int ican_int(int n);
short ican_short(short n);
void ican_shortv(short *v,int n);
int lm_image(int *w,int *h,int i,char *lbl);
int lm_total(void);
void do_awl(void);
int mi_image(int *nx,int *ny,int image_num,char *label);
int mi_total(void);
#ifdef GZMI
int gzmi_image(int *nx,int *ny,int image_num,char *label);
int gzmi_total(void);
#endif
void do_mag(int cursor,int x_only,int y_only);
void mi_close(void);
#ifdef GZMI
void gzmi_close(void);
#endif
int write_tiff_head(int height, int width, FILE *fp,char *ofn);
void scroll_init(void), func_scroll(int pm,int mask);
void push_cursor(int x,int y);
void func_measure(void);
void env_init(void);
void mi_push(int e),mi_pop(int e);
#ifdef GZMI
void gzmi_push(int e),gzmi_pop(int e);
#endif
void close_file(void);
void min_abbreviation(void);
void box_init(void);
void env_inc(short verbose);
void env_dec(short verbose);
void  env_push(int e);
void  env_pop(int e, short verbose);
void  env_upd_dir(int envdest, char *src);
void  expand_open(int argcnt, char *arglist[]);
void  write_env_panel(int env);
void  cmd_ls(void);
void  parse_expand_open(int argcnt, char *arglist[]);
#ifdef GZMI
PARSE_MIHEAD *parse_ogzmi_fe(gzFile *iu,int writeonly,FILE *ou,int err);
#endif
