/*
xpic_port.h - include file for ports to xpic
*/

static const char rcsid_port_h[] = "$Id: xpic_port.h,v 1.4 2003/08/01 13:58:10 mhirsch Exp $";

/*
	configuration defines
*/
#define PORT_SOCKET 8125

#ifndef NOMITSHM		/* change sense to simplify makefiles */
#ifndef MITSHM
#define MITSHM
#endif
#endif
/*
	packet types
*/
#define P_TEST 1
#define P_MESSAGE 2
#define P_STATUS 3
#define P_ANNOTATE 4
#define P_WL 5
#define P_LOAD 6
#define P_READ 7
#define P_WRITE 8
#define P_DEBUG_PORT 9
#define P_CLEAR 10
#define P_TRANSPOSE 11
#define P_CURSOR 12
#define P_COMMAND 13
#define P_CLOSE 14
#define P_MOVE 15
#define P_LINE 16
#define P_EXIT 17
#define P_MOUSE 18
/*
	structure for test packet
*/
#define P_TEST_WORDS 20
typedef struct{
	int size;
	int code;
	int buf[P_TEST_WORDS];
} TEST;
/*
	structure for message packet 
*/
#define MAX_MESSAGE 132
typedef struct{
	int size;
	int code;
	char message[MAX_MESSAGE+1];
} MESSAGE;
/*
	structure for annotate packet 
*/
typedef struct{
	int size;
	int code;
	int x;
	int y;
	char message[MAX_MESSAGE+1];
} ANNOTATE;
/*
	structures for status
*/
typedef struct{
	int size;
	int code;
} STATUS;

typedef struct{
	int x;
	int y;
	int window;
	int level;
	int mi_max;
	int canvas_id;
} STATUS_INFO;
/*
	structure for window/level packet 
*/
typedef struct{
	int size;
	int code;
	int window;
	int level;
} WL;
/*
	structures for load
*/
typedef struct{
	int size;
	int code;
} LOAD;
/*
	structures for debug
*/
typedef struct{
	int size;
	int code;
	int debug;
} DEBUG_PORT;
/*
	structures for clear
*/
typedef struct{
	int size;
	int code;
} CLEAR;
/*
	structures for write
*/
#define XP_LINE_MAX 1024
typedef struct{
	int size;
	int code;
	int x;
	int y;
	int n;	
	unsigned short data[XP_LINE_MAX];
} WRITE;
/*
	structures for write
*/
typedef struct{
	int size;
	int code;
	int x;
	int y;
	int n;	
} READ;

typedef struct{
	int n;	
	unsigned short data[XP_LINE_MAX];
} READ_INFO;
/*
	structures for transpose
*/
typedef struct{
	int size;
	int code;
	int transpose;
} TRANSPOSE;
/*
	structures for cursor
*/
typedef struct{
	int size;
	int code;
} CURSOR_I;

typedef struct{
	int type;	/* 0=off, 1=cursor, 2=box */
	int x;
	int y;
	int xw;
	int yw;
} CURSOR_INFO;

/*
	structure for command
*/
typedef struct{
	int size;
	int code;
	char cmd[MAX_MESSAGE+1];
} COMMAND;
/*
	structure for close
*/
typedef struct{
	int size;
	int code;
} CLOSE;
/*
	structure for move
*/
typedef struct{
	int size;
	int code;
	int x;
	int y;
} MOVE;
/*
	structure for line
*/
typedef struct{
	int size;
	int code;
	int x;
	int y;
} LINE;
/*
	structure for exit
*/
typedef struct{
	int size;
	int code;
} EXIT;
/*
	structure for mouse
*/
typedef struct{
	int size;
	int code;
	int b;
} MOUSE;

/*
	prototypes
*/
int xpic_open(char *host,int err);
int xpic_test(void);
void xpic_close(void);
void xpic_message(char *s);
void xpic_annotate(int x,int y,char *s);
void xpic_status(int *x,int *y,int *w,int *l,int *mi_max);
void xpic_wl(int w,int l);
void xpic_load(void);
void xpic_debug(int d);
void xpic_clear(void);
void xpic_write(int x,int y,unsigned short *data,int n);
int xpic_read(int x,int y,unsigned short *data,int n);
void xpic_transpose(int d);
void xpic_cursor(int *type,int *x,int *y,int *xw,int *yw);
void xpic_command(char *s);
void xpic_move(int x,int y);
void xpic_line(int x,int y);
void xpic_exit(void);
int xpic_mouse(int b);
short unsigned int *xpic_attach(void);
void xpic_detach(void);
