/*
xplot (main) - display plot(5) files in the X environment
*/

#ident "$Id: main.c,v 1.11 2011/09/14 10:09:37 carl Exp $"

#define OWTOOLKIT_WARNING_DISABLED
#include <X11/Xlib.h>

#include <xview/xview.h>
#include <xview/canvas.h>
#include <xview/panel.h>
#include <xview/font.h>
#include <xview/svrimage.h>
#include <xview/icon.h>
#include <xview/font.h>

/*
#include <sys/ioctl.h>
*/
#include <sys/types.h>
#include <sys/stat.h>

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

#include "parse.h"
#include "crcplot.h"
#include "crcplot_config.h"
#include "xplot.h"

/*
	program operational modes and parameters
*/
static int plot_rot;			/* 0=portrait 1=landscape */
static int nofork;			/* 1= don't fork from tty */
static int no_panel=1;			/* 1= don't use panel */
static int plot_scale=SCALE_D;		/* % of max size for hardcopy */
static float plot_ho=HOR,plot_vo=VER;	/* offsets (inches) for for hardcopy */
static char printer_name[80];		/* name of hardcopy printer */

static int plot_file_status;	/* 1= graphics exist in main window */
extern FILE *plot_file;		/* see notify.c */

/*
	window objects
*/
static Frame BaseFrame;			/* top level frame */
static Frame prop_frame;		/* frame for hardcopy properties*/
static Frame prop_panel;		/* panel hardcopy properties*/
static Panel panel;			/* main options panel */
static Display *dpy;			/* display's handle */
static Server_image frame_image_copy;	/* image of icon */
static Menu pop_menu;			/* menu for main windows */

/*
	popertry panel objects
*/
static Panel_item message_item;		
static Panel_item message_item2;
static Panel_item panel_printer;
static Panel_item panel_scale;
static Panel_item panel_rot;
static Panel_item panel_vo;
static Panel_item panel_ho;

		/* strings for horizontal and vertical offsets */
static char ho_string[80];
static char vo_string[80];

/*
	buffered command line
*/
#define MAX_GARGC 100		/* max # of args in file, env, command line */
int gargc;
char *gargv[MAX_GARGC];

/*
	stuff for buffering lines
*/
static int LineBufMax;
/*
	icons
*/
static unsigned short icon_bits[]  = {
#include "xplot.icon"
       };

static unsigned short icon_copy_bits[]  = {
#include "xplot_copy.icon"
       };
/*
	scaling macros - plot space to canvas
*/
#define	ScaleX(x)	(floor(pc->FactorX*((x)-pc->BottomX)+0.5))
#define	ScaleY(y)	(pc->CanvasY-floor(pc->FactorY*((y)-pc->BottomY)+0.5))
/*
	stuff for different line types and graphics contexes
*/
#define DOTTED 0
#define DOT_DASHED 1
#define SHORT_DASHED 2
#define LONG_DASHED 3
#define SOLID 4
#define LINE_TYPES 5

#define DOTTED_LIST_LENGTH 2
#define DOT_DASHED_LIST_LENGTH 4
#define SHORT_DASHED_LIST_LENGTH 2
#define LONG_DASHED_LIST_LENGTH 2

static int dash_list_length[] = {
	DOTTED_LIST_LENGTH,
	DOT_DASHED_LIST_LENGTH,
	SHORT_DASHED_LIST_LENGTH,
	LONG_DASHED_LIST_LENGTH,
};

	/* must be at least one element in each list */
static char dotted[DOTTED_LIST_LENGTH] = {3, 1};
static char dot_dashed[DOT_DASHED_LIST_LENGTH] = {3, 4, 3, 1};
static char short_dashed[SHORT_DASHED_LIST_LENGTH] = {4, 4};
static char long_dashed[LONG_DASHED_LIST_LENGTH] = {4, 7};

static char *dash_list[] = {
	dotted,
	dot_dashed,
	short_dashed,
	long_dashed
};

static GC gc[LINE_TYPES];	/* graphics context of for drawing lines */

/*
	inside-out-call stuff
*/
static int pd[2];			/* pipe */
static int iso_id;			/* noitifier id */

#define ISO_HARDCOPY	1
#define ISO_REDRAW	2

/*
	linked list and other information for copy windows
*/
typedef struct COPYLLs {
	Frame frame;			/* frame of copied graphics */
	Menu menu;			/* menu for frame */
	struct COPYLLs *prev;		/* pointer to previous copy frame */
	struct COPYLLs *next;		/* pointer to next copy frame */
	char *plotdata;			/* plot data */
	int n;				/* size of plot data (bytes) */
	int closed;			/* 0/1=frame open/closed */
	PC pc;				/* plot context */
} COPYLL;

COPYLL cll;

static int copyx=-1,copyy=-1;		/* default size of copy windows */
/*
	stuff for command line parser
*/
struct PARSE_STRUCT parse_table[] = {
	{"hor",PARSE_FLOAT,(PARSE_ARG)&plot_ho,
		"hardcopy horizontal offset (inches)"},
	{"ver",PARSE_FLOAT,(PARSE_ARG)&plot_vo,
		"hardcopy vertical offset (inches)"},
	{"port",PARSE_INT,(PARSE_ARG)&port_num,"IPC port address"},
	{"printer",PARSE_STRING,(PARSE_ARG)printer_name,
		"hardcopy printer name"},
	{"scale",PARSE_INT,(PARSE_ARG)&plot_scale,
		"hardcopy scale (percent)"},
	{"?",PARSE_PROC,(PARSE_ARG)parse_help_basic},
	{0,0,0,0}
	};

struct PARSE_FLAG_TABLE parse_flag_table[] = {
	{'F',&nofork,PARSE_FLAG_SET,"don't fork"},
	{'p',&no_panel,PARSE_FLAG_CLR,"no options panel"},
	{'?',(int *)(parse_help_basic),0,0},
	{0,0,0,0}
};

/*
	debugging macros
*/
#ifdef DEBUG
#define MESSAGE(x) puts(x)
#else
#define MESSAGE(x)
#endif	

#ifdef NEED_PRINTLL
static void
printll(s)
char *s;
{
	COPYLL *ll = &cll;
	
	puts(s);
	do{
		printf("\tframe: %d\n",ll->frame);
		ll = ll->next;
	}while(ll != NULL);
}
#endif

static COPYLL *
framell(frame)
Frame frame;
{
	COPYLL *ll = &cll;

	while(ll->frame != frame){
		if(ll->next == NULL) parse_err("xplot error: framell:",
					"missing frame");
		ll = ll->next;
	}
	return(ll);
}

static void
flush_lines(pc)
PC *pc;
{
	if(pc->LineBufCount){
		XDrawLines(dpy,pc->canvas_win,gc[pc->gcn],
			pc->LineBuf,pc->LineBufCount,CoordModeOrigin);
		pc->LineBufCount = 0;
	}
}

void
openpl(pc)
PC *pc;
{
	pc->gcn  = SOLID;
	pc->cmderr = 0;
	pc->mouse = 0;
}

void
closepl(pc)
PC *pc;
{
	MESSAGE("closepl");
	flush_lines(pc);
	pc->mouse = 0;
}


static XPoint *
LineBufInit()
{
	if(LineBufMax > 2){
		return((XPoint *) parse_malloc(sizeof(XPoint)*LineBufMax));
	}else{
		return(NULL);
	}
}

static void
LineBufRelease(p)
XPoint *p;
{
	if(LineBufMax > 2) parse_free((char *)p);
}

static void 
tell_user2(s1,s2)
char *s1,*s2;
{
	char cm[80];
	if(no_panel){
		strcpy(cm,s1);
		strcat(cm," ");
		strcat(cm,s2);
		xv_set(message_item,PANEL_LABEL_STRING,cm,NULL);
	}else{
		xv_set(message_item,PANEL_LABEL_STRING,s1,NULL);
		xv_set(message_item2,PANEL_LABEL_STRING,s2,NULL);
	}
}

void 
tell_user(s)
char *s;
{
	tell_user2(s,"");
}

void 
erase(pc)
PC *pc;
{
	MESSAGE("erase");
	flush_lines(pc);
	XClearWindow(dpy,pc->canvas_win);

	if(pc->type){
		fclose(plot_file);
		if((plot_file = fopen(plot_name,"w")) == NULL){
			tell_user("Can't reopen plot file");
		}
		plot_file_status = 0;
	}
}

static void 
erase_proc()
{
	tell_user("Erasing graphics");
	erase(&pc_main);
}


static void
redraw()
{
	char c[512];
	int n;
	FILE *fd;

	MESSAGE("redraw");
	if(!plot_file_status) return;
	fflush(plot_file);
	if((fd = fopen(plot_name,"r")) == NULL){
		tell_user("Can't open plot file");
		return;
	}
	XClearWindow(dpy,pc_main.canvas_win);
	openpl(&pc_main);
	while((n = fread(c,1,512,fd)) != 0){
		driver(&pc_main,c,n,0);
	}
	closepl(&pc_main);
	fclose(fd);
	tell_user("Refresh completed");
}

static void 
repaint_canvas(c,p,r)
Canvas c;
Xv_Window *p;
Rectlist *r;
{
	/* puts("repaint canvas"); */
	
	redraw();
}

static void
redraw_proc()
{
	char n;

	MESSAGE("redraw_proc");
	tell_user("Refresh begun");
	n = ISO_REDRAW;
	if(-1 == write(pd[1],&n,sizeof(n))) tell_user("redraw_proc: write error");
}

static void
copy_refresh(ll)
COPYLL *ll;
{
	/* printf("copy refresh: frame %d\n",ll->frame); */

	if(ll->n == 0)return;
	XClearWindow(dpy,ll->pc.canvas_win);
	openpl(&(ll->pc));
	driver_reset(&(ll->pc));
	driver(&(ll->pc),ll->plotdata,ll->n,0);
	closepl(&(ll->pc));
}

static void
copy_repaint_proc(canvas,paint_window,repaint_area)
Canvas canvas;
Xv_Window paint_window;
Rectlist *repaint_area;
{
	Frame frame = (Frame) xv_get(canvas,WIN_CLIENT_DATA);
	COPYLL *ll = framell(frame);

	/* printf("copy repaint: frame %d\n",frame); */
	copy_refresh(ll);
}

static Notify_value 
copy_event_proc(canvas,event)
Xv_Window canvas;
Event *event;
{
	Frame frame;
	COPYLL *ll;

	frame = (Frame) xv_get(canvas,WIN_CLIENT_DATA);

	ll = framell(frame);

	/*
	printf("copy event: frame %d event %d\n",frame,event_action(event));
	*/

	if(event_action(event) == ACTION_MENU && event_is_down(event)){
		menu_show(ll->menu,canvas,event,NULL);
	}else if(event_action(event) == WIN_RESIZE){
		copyx = (int) xv_get(ll->pc.canvas, CANVAS_WIDTH);
		copyy = (int) xv_get(ll->pc.canvas, CANVAS_HEIGHT);
	}
	return(NOTIFY_DONE);
}

static void
copy_frame_done(frame)
Frame frame;
{
	COPYLL *ll = framell(frame);

	/* printf("copy frame done, frame %d\n",frame); */

	LineBufRelease(ll->pc.LineBuf);
	if(ll->n && ll->plotdata != NULL) parse_free(ll->plotdata);
	ll->prev->next = ll->next;
	if(ll->next != NULL) ll->next->prev = ll->prev;
	parse_free((char *)ll);
	xv_destroy_safe(frame);
}

static void
frame_done()
{
	xv_destroy_safe(BaseFrame);
}

static void 
copyplot(ll)
COPYLL *ll;
{
	FILE *fd;
	struct stat buf;

	MESSAGE("copy update");

	if(ll->plotdata != NULL){
		parse_free(ll->plotdata);
		ll->n = 0;
		ll->plotdata = NULL;
	}

	if(!plot_file_status) return;
	fflush(plot_file);
	if((fd = fopen(plot_name,"r")) == NULL){
		tell_user("Can't open plot file");
		return;
	}
	if(stat(plot_name,&buf) == -1){
		tell_user("Can't stat plot file");
		return;
	}
	if(buf.st_size == 0){
		tell_user("Nothing to copy!");
		return;
	}

	ll->plotdata = (char *) parse_malloc(buf.st_size);
	ll->n = buf.st_size;

	if(fread(ll->plotdata,1,buf.st_size,fd) != buf.st_size){
		tell_user("copyplot read error");
		parse_free(ll->plotdata);
		ll->n = 0;
		return;
	}
	tell_user("copy completed");
}

static void
pop_up_proc()
{
	xv_set(prop_frame,FRAME_CMD_PUSHPIN_IN,TRUE,NULL);
	xv_set(prop_frame, WIN_SHOW, TRUE, NULL);
}

static void
hardcopy_prop()
{
	plot_scale = (int) xv_get(panel_scale,PANEL_VALUE);
	strcpy(printer_name,(char *) xv_get(panel_printer,PANEL_VALUE));
	plot_ho = atof((char *) xv_get(panel_ho,PANEL_VALUE));
	plot_vo = atof((char *) xv_get(panel_vo,PANEL_VALUE));
	plot_rot = (int) xv_get(panel_rot,PANEL_VALUE);
}

static void
lprpipe_close(pfd)
FILE *pfd;
{
#ifndef DEBUG_HARD
	char s[512];
	pclose(pfd);
	strcpy(s,"Printer: ");
	strcat(s,printer_name);
	tell_user2("Hardcopy queued:",s);
#else
	parse_close(pfd);
#endif
}

static FILE *
lprpipe_open()
{
	char s[512];
	FILE *pfd;

#ifndef DEBUG_HARD
#ifdef HAS_LP
	strcpy(s,"lp -o nobanner -d ");
#else
	strcpy(s,"lpr -P");
#endif
	strcat(s,printer_name); 
	if(NULL == (pfd = popen(s,"w"))){
		tell_user2("Can't pipe to:",s);
		return(NULL);
	}else{
		return(pfd);
	}
#else
	pfd = parse_open("1.ls","w");
	fputs("debug output sent to 1.ls\n",stdout);
	return(pfd);
#endif
}


static void
hardcopy_copy(ll)
COPYLL *ll;
{
	FILE *pfd;

	if(ll->n == 0){
		tell_user("Plot buffer empty");
		return;
	}

	if((pfd = lprpipe_open()) == NULL) return;

	hardcopy_prop();
	plot_tops_buf(ll->plotdata,ll->n,pfd,plot_scale,
		0,plot_rot,plot_ho,plot_vo);

	lprpipe_close(pfd);
}


static void
copy_menu_notify_proc(menu,item)
Menu menu;
Menu_item item;
{
	char *choice = (char *)xv_get(item,MENU_STRING);
	Frame frame = (Frame) xv_get(menu,MENU_CLIENT_DATA);
	COPYLL *ll = framell(frame);

	if(!strcmp(choice,"Hardcopy")){
		hardcopy_copy(ll);
	}else if(!strcmp(choice,"Copy")){
		copyplot(ll);
		copy_refresh(ll);
	}else if(!strcmp(choice,"Refresh")){
		copy_refresh(ll);
	}else if(!strcmp(choice,"Properties")){
		pop_up_proc();
	}else if(!strcmp(choice,"Quit")){
		copy_frame_done(frame);
	}
}


static void
copy_proc()
{
	Icon icon;		
	Frame copy_frame;		
	Canvas copy_canvas;		
	COPYLL *new;
	COPYLL *ll = &cll;

	MESSAGE("copy_proc");
	tell_user("Copying graphics");

	while(ll->next != NULL) ll = ll->next;
	new = (COPYLL *) parse_malloc(sizeof(COPYLL));
	ll->next = new;
	new->prev = ll;
	new->next = NULL;

	if(copyx == -1 || copyy == -1){
		new->pc.CanvasX = (int) xv_get(pc_main.canvas, CANVAS_WIDTH);
		new->pc.CanvasY = (int) xv_get(pc_main.canvas, CANVAS_HEIGHT);
	}else{
		new->pc.CanvasX = copyx;
		new->pc.CanvasY = copyy;
	}

	new->frame = copy_frame = xv_create(BaseFrame, FRAME,
		FRAME_LABEL, "xplot copied graphics",
		FRAME_DONE_PROC, copy_frame_done,
		NULL);

	icon = (Icon) xv_create(XV_NULL,ICON,
		ICON_IMAGE, frame_image_copy,
		ICON_TRANSPARENT, TRUE,
		NULL);

	xv_set(copy_frame,FRAME_ICON,icon,NULL);

	new->pc.canvas = copy_canvas = xv_create(copy_frame, CANVAS,
		CANVAS_WIDTH,new->pc.CanvasX,
		CANVAS_HEIGHT,new->pc.CanvasY,
		XV_WIDTH,new->pc.CanvasX,
		XV_HEIGHT,new->pc.CanvasY,
		CANVAS_FIXED_IMAGE, FALSE,
		CANVAS_REPAINT_PROC, copy_repaint_proc,
		WIN_CLIENT_DATA, copy_frame,
		OPENWIN_SHOW_BORDERS, FALSE,
		NULL);

	window_fit(copy_frame);

	new->menu = (Menu) xv_create((int)NULL,MENU,
		MENU_TITLE_ITEM, 	"xplot copy pane",
		MENU_STRINGS,		"Hardcopy", 
					"Copy",
					"Properties",
					"Refresh",
					"Quit",
					NULL,
		MENU_NOTIFY_PROC,	copy_menu_notify_proc,
		MENU_CLIENT_DATA,	copy_frame,
		NULL);

	xv_set(canvas_paint_window(copy_canvas),
		WIN_EVENT_PROC,	copy_event_proc,
		WIN_CONSUME_EVENTS,
			WIN_NO_EVENTS,
			WIN_MOUSE_BUTTONS,
			WIN_REPAINT,
			WIN_RESIZE,
			NULL,
		WIN_CLIENT_DATA, copy_frame,
		NULL);

	new->plotdata = NULL;
	new->n = 0;
	new->pc.type = 0;
	copyplot(new);

	new->pc.canvas_win = 
		(Window)xv_get(canvas_paint_window(copy_canvas), XV_XID);
	new->pc.gcn = SOLID;
	new->pc.LineBuf = (XPoint *) LineBufInit();

	xv_set(copy_frame, WIN_SHOW, TRUE, NULL);
}

static void 
xplot_hardcopy()
{
	FILE *fd;
	FILE *pfd;

	if(plot_file_status == 0){
		tell_user("Plot buffer empty");
		return;
	}

	if((pfd = lprpipe_open()) == NULL) return;

	fflush(plot_file);
	if(NULL == (fd = fopen(plot_name,"r"))){
		tell_user2("Can't open:",plot_name);
		pclose(pfd);
		return;
	}

	plot_tops(fd,pfd,plot_scale,0,plot_rot,plot_ho,plot_vo);

	fclose(fd);
	lprpipe_close(pfd);
}

void
start_hard_copy(iso)
int iso;
{
	char n;

	tell_user("Starting hardcopy");
	hardcopy_prop();
	XFlush(dpy);
	if(iso){
		n = ISO_HARDCOPY;
		if(-1 == write(pd[1],&n,sizeof(n))) tell_user("start_hard_copy: write error");
	}else{
		xplot_hardcopy();
	}
}

static void
pop_done(item,event)
Panel_item item;
Event *event;
{
	xv_set(prop_frame,FRAME_CMD_PUSHPIN_IN,FALSE,NULL);
	xv_set(prop_frame, WIN_SHOW, FALSE, NULL);
}

static void 
pop_hard(item,event)
Panel_item item;
Event *event;
{
	start_hard_copy(1);
}

static void 
pop_up_create()
{
	int i=0;
	prop_frame = xv_create(BaseFrame,FRAME_CMD,
		  FRAME_LABEL, (no_panel)?
			"xplot Properties" :
			"xplot Hardcopy Properties",
		  NULL);
	prop_panel = (Panel) xv_get(prop_frame,FRAME_CMD_PANEL);

	panel_printer = xv_create(prop_panel,PANEL_TEXT,
		PANEL_LABEL_X, xv_col(prop_panel,0),
		PANEL_LABEL_Y, xv_row(prop_panel,i++),
		PANEL_VALUE, printer_name,
		PANEL_VALUE_DISPLAY_LENGTH, 30,
		PANEL_LABEL_STRING, "Printer:",
		NULL);

	panel_scale = xv_create(prop_panel,PANEL_SLIDER,
		PANEL_LABEL_X, xv_col(prop_panel,0),
		PANEL_LABEL_Y, xv_rows(prop_panel,i++),
		PANEL_VALUE, plot_scale,
		PANEL_MIN_VALUE, 0,
		PANEL_MAX_VALUE, 100,
		PANEL_SLIDER_WIDTH, 150,
		PANEL_LABEL_STRING,"Scale(%):",
		NULL);

	panel_ho = xv_create(prop_panel,PANEL_TEXT,
		PANEL_LABEL_X, xv_col(prop_panel,0),
		PANEL_LABEL_Y, xv_rows(prop_panel,i++),
		PANEL_VALUE_DISPLAY_LENGTH, 30,
		PANEL_VALUE, ho_string,
		PANEL_LABEL_STRING, "Horizontal offset:",
		NULL);

	panel_vo = xv_create(prop_panel,PANEL_TEXT,
		PANEL_LABEL_X, xv_col(prop_panel,0),
		PANEL_LABEL_Y, xv_rows(prop_panel,i++),
		PANEL_VALUE_DISPLAY_LENGTH, 30,
		PANEL_VALUE, vo_string,
		PANEL_LABEL_STRING, "Vertical offset:",
		NULL);

	panel_rot = xv_create(prop_panel,PANEL_CHOICE,
		PANEL_LABEL_X, xv_col(prop_panel,0),
		PANEL_LABEL_Y, xv_rows(prop_panel,i++),
		PANEL_LABEL_STRING, "Orientation:",
		PANEL_CHOICE_STRINGS, "Portrait", "Landscape", NULL,
		NULL);

	if(no_panel){
		message_item = xv_create(prop_panel,PANEL_MESSAGE,
			PANEL_LABEL_X, xv_col(prop_panel,0),
			PANEL_LABEL_Y, xv_rows(prop_panel,i++),
			PANEL_LABEL_STRING, "   ",
			0);
	}

	xv_create(prop_panel,PANEL_BUTTON,
		PANEL_LABEL_X, xv_col(prop_panel,0),
		PANEL_LABEL_Y, xv_rows(prop_panel,++i),
		PANEL_LABEL_STRING, "Dismiss",
		PANEL_NOTIFY_PROC, pop_done, 
		NULL);

	xv_create(prop_panel, PANEL_BUTTON,
		PANEL_LABEL_STRING, "Hard",
		PANEL_NOTIFY_PROC, pop_hard, 
		NULL);
	
	xv_create(prop_panel,PANEL_BUTTON,
		PANEL_LABEL_STRING,"Erase",
		PANEL_NOTIFY_PROC, erase_proc,
		NULL);

	xv_create(prop_panel,PANEL_BUTTON,
		PANEL_LABEL_STRING,"Refresh",
		PANEL_NOTIFY_PROC, redraw_proc,
		NULL);

	xv_create(prop_panel,PANEL_BUTTON,
		PANEL_LABEL_STRING,"Copy",
		PANEL_NOTIFY_PROC, copy_proc,
		NULL);

	xv_create(prop_panel,PANEL_BUTTON,
		PANEL_LABEL_STRING,"Quit",
		PANEL_NOTIFY_PROC, frame_done,
		NULL);

	window_fit(prop_panel);
	
	window_fit(prop_frame);
}

static void 
hard_panel_event(item,event)
Panel_item item;
Event *event;
{
	if(event_action(event) == ACTION_MENU && event_is_down(event)){
		pop_up_proc();
	}else if(event_action(event) == ACTION_SELECT && event_is_down(event)){
		start_hard_copy(1);
	}else{
		panel_default_handle_event (item, event);
	}
}

static Notify_value
canvas_event_proc(canvas_obj,event)
Xv_Window canvas_obj;
Event *event;
{
	int xp,yp;

	/* printf("canvas_event_proc, event %d\n", event_action(event)); */

	if(event_action(event) == ACTION_MENU && event_is_down(event)){
		menu_show(pop_menu,canvas_obj,event,NULL);
	}else if(event_action(event) == ACTION_SELECT && event_is_down(event)
	  && pc_main.mouse){
		xp = floor((event_x(event) -0.5)/pc_main.FactorX +
			pc_main.BottomX);
		yp = floor((pc_main.CanvasY-event_y(event)+0.5)/pc_main.FactorY 
			+pc_main.BottomY);

		/*
		fprintf(stderr,"mouse: (%d %d) = (%d %d)\n",
			event_x(event),event_y(event),xp,yp);
		*/
		mouse_send(xp,yp);
		pc_main.mouse = 0;
	}
	return(NOTIFY_DONE);
}

static void
menu_notify_proc(menu_obj,menu_item)
Menu menu_obj;
Menu_item menu_item;
{
	char *menu_choice = (char *)xv_get(menu_item,MENU_STRING);

	if(!strcmp(menu_choice,"Hardcopy")){
		start_hard_copy(1);
	}else if(!strcmp(menu_choice,"Erase")){
		erase(&pc_main);
	}else if(!strcmp(menu_choice,"Refresh")){
		redraw();
	}else if(!strcmp(menu_choice,"Properties")){
		pop_up_proc();
	}else if(!strcmp(menu_choice,"Copy")){
		copy_proc();
	}else if(!strcmp(menu_choice,"Quit")){
		frame_done();
	}
}

static void
menu_pop_create()
{
	pop_menu = (Menu) xv_create((int) NULL,MENU,
		MENU_TITLE_ITEM, 	"xplot pane",
		MENU_STRINGS,		"Hardcopy", 
					"Erase", 
					"Refresh",
					"Copy",
					"Properties",
					"Quit",
					NULL,
		MENU_NOTIFY_PROC,	menu_notify_proc,
		NULL);
}

void
extension(pc,s)
PC *pc;
char *s;
{
	flush_lines(pc);

	if(!strcmp(s,"mouse")){
		pc->mouse = 1;
	}else if(!strcmp(s,"flush")){
		;
	}else{
		tell_user2("bad extension command",s);
		pc->cmderr = 1;
	}
}

void
space(pc,LowX,LowY,HighX,HighY)
PC *pc;
int LowX,LowY,HighX,HighY;
{
	float	CanvasSize;
	
	MESSAGE("space");
	pc->CanvasX = (int) xv_get(pc->canvas, CANVAS_WIDTH);
	pc->CanvasY = (int) xv_get(pc->canvas, CANVAS_HEIGHT);
	
	if (pc->CanvasX < pc->CanvasY){
		CanvasSize = pc->CanvasX;
	}else{
		CanvasSize = pc->CanvasY;
	}

	pc->FactorX = CanvasSize/(float)(HighX-LowX);
	pc->FactorY = CanvasSize/(float)(HighY-LowY);
	pc->BottomX = LowX;
	pc->BottomY = LowY;
	if(pc->type)plot_file_status = 1;
	flush_lines(pc);
}

void
move(pc,x,y)
PC *pc;
int x,y;
{
	flush_lines(pc);
	pc->CurrentX = ScaleX(x);
	pc->CurrentY = ScaleY(y);
	if(pc->type)plot_file_status = 1;
}

void
cont(pc,x,y)
PC *pc;
int x,y;
{
	register int	NewX, NewY;

	NewX = ScaleX(x);
	NewY = ScaleY(y);
	if(LineBufMax >2){
		if(pc->LineBufCount == LineBufMax) flush_lines(pc);
		if(!pc->LineBufCount){
			pc->LineBuf->x = pc->CurrentX;
			pc->LineBuf->y = pc->CurrentY;
			pc->LineBufCount = 1;
		}
			(pc->LineBuf+pc->LineBufCount)->x = NewX;
			(pc->LineBuf+pc->LineBufCount)->y = NewY;
			pc->LineBufCount++; 
	}else{
		XDrawLine(dpy,pc->canvas_win,gc[pc->gcn],
			pc->CurrentX,pc->CurrentY,NewX,NewY);
	}

	pc->CurrentX = NewX;
	pc->CurrentY = NewY;
	if(pc->type)plot_file_status = 1;
}

void
draw_line(pc,x0, y0, x1, y1)
PC *pc;
int x0, y0, x1, y1;
{
	register int	OldX, OldY;

	flush_lines(pc);
	OldX = ScaleX(x0);
	OldY = ScaleY(y0);
	pc->CurrentX = ScaleX(x1);
	pc->CurrentY = ScaleY(y1);
	XDrawLine(dpy,pc->canvas_win,gc[pc->gcn], 
		OldX, OldY, pc->CurrentX, pc->CurrentY);
	if(pc->type)plot_file_status = 1;
}

void
point(pc,x,y)
PC *pc;
int x,y;
{
	flush_lines(pc);
	pc->CurrentX = ScaleX(x);
	pc->CurrentY = ScaleY(y);
	XDrawPoint(dpy,pc->canvas_win,gc[pc->gcn], pc->CurrentX, pc->CurrentY);
	if(pc->type)plot_file_status = 1;
}

void
label(pc,s)
PC *pc;
char	*s;
{
	flush_lines(pc);
	XDrawString(dpy,pc->canvas_win,gc[pc->gcn], 
		pc->CurrentX, pc->CurrentY, s, strlen(s));
	if(pc->type)plot_file_status = 1;
}

void
linemod(pc,mode)
PC *pc;
char *mode;
{
	flush_lines(pc);
	if(!strcmp(mode,"dotted"))
		pc->gcn = DOTTED;
	else if(!strcmp(mode,"longdashed"))
		pc->gcn =  LONG_DASHED;
	else if(!strcmp(mode,"shortdashed"))
		pc->gcn =  SHORT_DASHED;
	else if(!strcmp(mode,"dotdashed"))
		pc->gcn =  DOT_DASHED;
	else
		pc->gcn = SOLID;
}

void
arc(pc,xc,yc,xbeg,ybeg,xend,yend)
PC *pc;
int xc,yc,xbeg,ybeg,xend,yend;
{
	double fa1,fa2;
	int r, a1,a2;
	int screen_xc, screen_yc, screen_xbeg, screen_ybeg; 
	int screen_xend, screen_yend;

	flush_lines(pc);

	screen_xc = ScaleX(xc);
	screen_yc = ScaleY(yc);
	screen_xbeg = ScaleX(xbeg) - screen_xc;
	screen_ybeg = ScaleY(ybeg) - screen_yc;
	screen_xend = ScaleX(xend) - screen_xc;
	screen_yend = ScaleY(yend) - screen_yc;

	r = sqrt((double)(screen_xbeg*screen_xbeg+screen_ybeg*screen_ybeg));

	fa1 = atan2((double)screen_ybeg,(double)screen_xbeg);
	fa2 = atan2((double)screen_yend,(double)screen_xend);

	pc->CurrentX = r * cos(fa1) + screen_xc;
	pc->CurrentY = r * sin(fa1) + screen_yc;

	a1 = fa1 / M_PI * 180 * 64;
	a2 = fa2 / M_PI * 180 * 64;
	if(a2 > 0) a2 = -360*64 + a2;

	XDrawArc(
		dpy,
		pc->canvas_win,
		gc[pc->gcn],
		screen_xc-r,
		screen_yc-r,
		2*r,
		2*r,
		a1,
		a2);

	if(pc->type)plot_file_status = 1;
}

void
circle(pc,xc,yc,r)
PC *pc;
int xc,yc,r;
{
	flush_lines(pc);

	arc(pc,xc,yc, xc+r,yc, xc-r,yc);
	arc(pc,xc,yc, xc-r,yc, xc+r,yc);
	if(pc->type)plot_file_status = 1;
}

void
parse(cargc,cargv)
int cargc;
char **cargv;
{
	int i;

	for(i=1;i<cargc;i++){
		if(gargc == MAX_GARGC) 
			parse_err("too many command line arguments",NULL);

		gargv[gargc++] = cargv[i];
	}
}

static void
parse_xplot(argc,argv)
	int	argc;
	char	**argv;
{
	while(argv++ , --argc>0){
		if(argv[0][0] == '-'){
			parse_flags(*argv);
		}else{
			parse_comm_aa(&argc,&argv);
		}
	}
}

static Notify_value 
read_iso_pipe(client,fd)
Notify_client client;
int fd;
{
	char c;

	if(-1 == read(fd, &c, sizeof c)) tell_user("read_iso_pipe: read error");

	switch(c){

	case ISO_HARDCOPY:
		xplot_hardcopy();
		break;
	case ISO_REDRAW:
		redraw();
		break;
	default:
		parse_pri("xplot: bad internal pipe op code:",c);
	}

	return(NOTIFY_DONE);
}


static void
cleanup()
{
	clean_socket();
	fclose(plot_file);
	unlink(plot_name);
}

static Notify_value 
frame_interposer(frame,event,arg,type)
Frame frame;
Event *event;
Notify_arg arg;
Notify_event_type type;
{
	int closed_initial,closed_current;
	Notify_value value;
	COPYLL *ll = &cll;

	/* puts("in frame interposer"); */

	closed_initial = (int) xv_get(frame,FRAME_CLOSED);
	value = notify_next_event_func(frame,(int)event,arg,type);
	closed_current = (int) xv_get(frame,FRAME_CLOSED);
	if(closed_initial != closed_current){
		if(closed_current){
		  while(1){
			if(ll->frame != (Frame) NULL){
				ll->closed = xv_get(ll->frame,FRAME_CLOSED);
				if(!ll->closed){
					xv_set(ll->frame,
						FRAME_CLOSED,TRUE,NULL);
				}
			}
			if(ll->next == NULL) break;
			ll = ll->next;
		  }
		}else{
		  while(1){
			if(ll->frame != (Frame) NULL){
				if(!ll->closed){
					xv_set(ll->frame,
						FRAME_CLOSED,FALSE,NULL);
				}
			}
			if(ll->next == NULL) break;
			ll = ll->next;
		  }
		  xv_set(frame,XV_SHOW,TRUE,NULL);
		}
	}
	return(value);
}

int
main(argc, argv)
int	argc;
char	**argv;
{
	int RepaintCanvas();
	int init,i,pid;
	char *s,*getenv();
	char pt[30],header[50];
	XGCValues gcvalues;
	Xv_Font     _font;
	XFontStruct *font;
	Icon icon;		
	Xv_Font messagef;
	Canvas DrawingCanvas;
	Server_image frame_image;	/* image of icon */

/*
	general initialization
*/
	port_num = PORT;

	strcpy(printer_name,"lp"); 
	if((s = getenv("PRINTER")) != NULL) strcpy(printer_name,s); 

	cll.frame = (Frame) NULL;
	cll.next = NULL;
/*
	parse command line, files and environement
*/
	gargc = 1;
	gargv[0] = argv[0];

					/* _file and _args stuff their args 
					into gargv because of parse() */

	for(init=i=1;i<argc;i++){
		if(argv[i][0] == '-' && (s = strchr(argv[i],'@')) != NULL){
			init = 0;
			*s = ' ';
		}
	}

	parse_file_default(".xplot");
	parse_arg_default("XPLOTARGS");

	if(init){
		parse_file(".xplot");
		parse_args("XPLOTARGS");
	}

					/* copy command line into gargv */
	for(i=1;i<argc;i++){
		if(gargc == MAX_GARGC) 
			parse_err("too many command line arguments",NULL);

		gargv[gargc++] = argv[i];
	}

#ifndef NOXVINIT
#ifdef DEBUG
	for(i=0;i<gargc;i++)printf("before: %d %s\n",i,gargv[i]);
#endif

					/* xv_init takes what it needs */
	xv_init(XV_INIT_ARGC_PTR_ARGV,&gargc,gargv,NULL);

#ifdef DEBUG
	for(i=0;i<gargc;i++)printf("after: %d %s\n",i,gargv[i]);
#endif
#endif

					/* now parse what is left */
	parse_xplot(gargc,gargv);
/*
	set up plotter name
*/
	if(port_num == PORT){
		pt[0] = 0;
	}else{
		sprintf(pt,".%d",port_num);
	}
	plot_name = parse_malloc_char(strlen(PLOTNAME)+strlen(pt)+1);
	strcpy(plot_name,PLOTNAME);
	strcat(plot_name,pt);
/*
	set up more globals and do some checking
*/
	sprintf(ho_string,"%3.1f",plot_ho);
	sprintf(vo_string,"%3.1f",plot_vo);

	if(plot_scale < 1 || plot_scale > 100){
	   fprintf(stderr,"Scale must be in range [1,100]\n");
	   exit(1);
	}
/*
	detach from keyboard
*/
	if(!nofork){
		if((pid = fork()) == -1){
			perror("Can't fork copy of xplot");
			exit(1);
		}
		if(pid != 0)exit(0);
	}

	strcpy(header,"xplot");

	BaseFrame = xv_create(XV_NULL,FRAME,
		  FRAME_INHERIT_COLORS, 1,
		  XV_LABEL,header,
		  FRAME_NO_CONFIRM, FALSE,
		  NULL);

#ifndef __alpha
	(void) notify_interpose_event_func(BaseFrame,
		frame_interposer,NOTIFY_SAFE);
#endif

	dpy = (Display *)xv_get(BaseFrame, XV_DISPLAY);

    	if(nofork){
		XSynchronize(dpy,True);
		XSync(dpy,FALSE);
	}

	frame_image  = (Server_image) xv_create((int)NULL, 
		SERVER_IMAGE,
		XV_WIDTH, 64,
		XV_HEIGHT, 64,
		SERVER_IMAGE_BITS, icon_bits,
		NULL);

	frame_image_copy  = (Server_image) xv_create((int)NULL, 
		SERVER_IMAGE,
		XV_WIDTH, 64,
		XV_HEIGHT, 64,
		SERVER_IMAGE_BITS, icon_copy_bits,
		NULL);

	icon = (Icon) xv_create(XV_NULL,ICON,
		ICON_IMAGE, frame_image,
		ICON_TRANSPARENT, TRUE,
		NULL);

	xv_set(BaseFrame, FRAME_ICON,icon,NULL);

	if(!no_panel){
		messagef = xv_create(BaseFrame,FONT,NULL);
		panel = xv_create(BaseFrame,PANEL,NULL);
		xv_create(panel,PANEL_BUTTON,
			PANEL_LABEL_STRING,"Hard",
			PANEL_EVENT_PROC, hard_panel_event, 
			NULL);

		xv_create(panel,PANEL_BUTTON,
			PANEL_LABEL_STRING,"Erase",
			PANEL_NOTIFY_PROC, erase_proc,
			NULL);

		xv_create(panel,PANEL_BUTTON,
			PANEL_LABEL_STRING,"Refresh",
			PANEL_NOTIFY_PROC, redraw_proc,
			NULL);

		xv_create(panel,PANEL_BUTTON,
			PANEL_LABEL_STRING,"Copy",
			PANEL_NOTIFY_PROC, copy_proc,
			NULL);

		xv_create(panel,PANEL_BUTTON,
			PANEL_LABEL_STRING,"Quit",
			PANEL_NOTIFY_PROC, frame_done,
			NULL);

		message_item = xv_create(panel,PANEL_MESSAGE,
			PANEL_LABEL_STRING, "  ",
			0);

		message_item2 = xv_create(panel,PANEL_MESSAGE,
			PANEL_LABEL_STRING, "  ",
			XV_X,xv_get(message_item,XV_X),
			XV_Y,xv_get(message_item,XV_Y) + 
				xv_get(messagef,FONT_DEFAULT_CHAR_HEIGHT),
			NULL);

		window_fit_height(panel);
	}

	pc_main.canvas = DrawingCanvas = xv_create(BaseFrame,CANVAS,
		CANVAS_REPAINT_PROC, repaint_canvas,
		CANVAS_FIXED_IMAGE, FALSE,
		NULL);

	xv_set(canvas_paint_window(DrawingCanvas),
		WIN_EVENT_PROC,	canvas_event_proc,
		WIN_CONSUME_EVENTS,
			WIN_NO_EVENTS,
			WIN_MOUSE_BUTTONS,
			WIN_REPAINT,
			ACTION_CLOSE,
			NULL,
		NULL);

	pc_main.canvas_win = 
		(Window)xv_get(canvas_paint_window(DrawingCanvas), XV_XID);

	pc_main.type = 1;
	pc_main.gcn = SOLID;
	pc_main.mouse = 0;

	_font = (Xv_Font)xv_find(BaseFrame, FONT, 
		FONT_FAMILY,FONT_FAMILY_DEFAULT_FIXEDWIDTH,
		FONT_STYLE, FONT_STYLE_NORMAL,
		FONT_SIZE, 14,
		NULL);
	font = (XFontStruct *)xv_get(_font, FONT_INFO);

	gcvalues.font = font->fid;
	gcvalues.foreground = BlackPixel(dpy, DefaultScreen(dpy));
	gcvalues.background = WhitePixel(dpy, DefaultScreen(dpy));
	gcvalues.line_style = LineOnOffDash;

	for(i=0;i<LINE_TYPES;i++){
		if(i != SOLID){
			gc[i] = XCreateGC(dpy, 
				RootWindow(dpy, DefaultScreen(dpy)),
				 GCLineStyle | GCFont | GCForeground 
					| GCBackground, 
				 &gcvalues);
			XSetDashes(dpy,gc[i],0,dash_list[i],dash_list_length[i]);
		}else{
			gc[i] = XCreateGC(dpy, 
				RootWindow(dpy, DefaultScreen(dpy)),
				 GCFont | GCForeground | GCBackground, 
				 &gcvalues);
		}
	}

	pop_up_create();
	menu_pop_create();

	if(pipe(pd) == -1) parse_err("can't open internal pipe",NULL);
	notify_set_input_func((Notify_client) &iso_id,read_iso_pipe,pd[0]);

	notify();

	LineBufMax = (XMaxRequestSize(dpy)-3)/2 -1;	/* see XDrawLines */
	pc_main.LineBuf = (XPoint *) LineBufInit();

	xv_main_loop(BaseFrame);
	cleanup();
	exit(0);
}
